package mstparser;

import gnu.trove.map.hash.TLongIntHashMap;

import java.io.*;

import java.util.*;
import java.util.zip.GZIPInputStream;

import static mstparser.FeatureTemplate.*;
import static mstparser.FeatureTemplate.TemplateType.*;
import static mstparser.FeatureVector.end;

public class DependencyPipe2O {

	public FeatureAlphabet dataAlphabet;
	
	// dictionary
	public Alphabet typeAlphabet;
	public Alphabet cposAlphabet;		// cpos, need index separately
	public Alphabet posAlphabet;		// pos
	public Alphabet morphologyIDAlphabet;	// morphology val
	public Alphabet morphologyValueAlphabet;	// morphology val
	public Alphabet lexAlphabet;		// lemma, word
	
	public static final int START = 1;
	public static final int MID = 2;
	public static final int END = 3;
	//public static final int NONE = 4;
	
	public static final int QUOTE = 4;
	public static final int LRB = 5;
	public static final int RRB = 6;
	
	public static final int NOTYPE = 1;

	public static final int SBJ = 2;
	public static final int IDF = 3;
	
	// special value for spmrl
	public static final int NA = 1;
	public static final int B = 2;
	public static final int M = 3;
	public static final int F = 4;
	public static final int S = 5;
	public static final int P = 6;
	public static final int I = 7;

	public String[] types;
	public String[] posName;

	HashSet<String> possibleLab;

	public boolean labeled = false;
	public boolean hasLemma = true;
	public boolean hasCpos = true;

	public HashMap<String, String> coarseMap;
	public String lang;
	public String path;
	public int ccDepType;
	
	public boolean[] usedAsDev = null;

	public DependencyPipe2O() throws IOException {
		dataAlphabet = new FeatureAlphabet();
		typeAlphabet = new Alphabet(100);
		posAlphabet = new Alphabet(100);
		cposAlphabet = new Alphabet(100);
		morphologyIDAlphabet = new Alphabet(100);
		morphologyValueAlphabet = new Alphabet(100);
		lexAlphabet = new Alphabet(10000);
		possibleLab = new HashSet<String>();
		
		initAlphabet();
	}
	
	public void initAlphabet() {
		// fill alphabet with special value
		cposAlphabet.lookupIndex("#START#");
		cposAlphabet.lookupIndex("#MID#");
		cposAlphabet.lookupIndex("#END#");
		//cposAlphabet.lookupIndex("#NONE#");
		
		posAlphabet.lookupIndex("#START#");
		posAlphabet.lookupIndex("#MID#");
		posAlphabet.lookupIndex("#END#");

		lexAlphabet.lookupIndex("#START#");
		lexAlphabet.lookupIndex("#MID#");
		lexAlphabet.lookupIndex("#END#");
		lexAlphabet.lookupIndex("\"");
		lexAlphabet.lookupIndex("(");
		lexAlphabet.lookupIndex(")");
		
		typeAlphabet.lookupIndex("<no-type>");
		
		if (DependencyParser.isSpmrl) {
			if (DependencyParser.labeled) {
				typeAlphabet.lookupIndex("SBJ");
				typeAlphabet.lookupIndex("IDF");
			}
		
			morphologyValueAlphabet.lookupIndex("na");
			morphologyValueAlphabet.lookupIndex("B");
			morphologyValueAlphabet.lookupIndex("M");
			morphologyValueAlphabet.lookupIndex("F");
			morphologyValueAlphabet.lookupIndex("S");
			morphologyValueAlphabet.lookupIndex("P");
			morphologyValueAlphabet.lookupIndex("I");
		}
	}
	
	public void setAndCheckOffset() {
		// set
		largeOff = getBits(lexAlphabet.size());
		midOff = getBits(posAlphabet.size());
		midOff = Math.max(getBits(cposAlphabet.size()), midOff);
		midOff = Math.max(getBits(typeAlphabet.size()), midOff);
		moridOff = getBits(morphologyIDAlphabet.size());
		morvalOff = getBits(morphologyValueAlphabet.size());
		tempOff = getBits(ArcMST.COUNT.ordinal());
		tempOff = Math.max(getBits(ArcTurbo.COUNT.ordinal()), tempOff);
		
		System.out.println("large offset: " + largeOff);
		System.out.println("mid offset: " + midOff);
		System.out.println("morphology id offset: " + moridOff);
		System.out.println("morphology value offset: " + morvalOff);
		System.out.println("temp offset: " + tempOff);
		
		// check
		if (midOff * 5 + tempOff + flagOff > 64) {
			System.out.println("size too large 1");
			System.exit(0);
		}
		else if (largeOff * 2 + midOff * 2 + tempOff + flagOff > 64) {
			System.out.println("size too large 2");
			System.exit(0);
		}
		else if (midOff * 2 + moridOff * 2 + morvalOff * 2 + tempOff + flagOff > 64) {
			System.out.println("size too large 3");
			System.exit(0);
		}
		else if (largeOff * 3 + tempOff > 64) {
			System.out.println("size too large 4");
			System.exit(0);
		}
		else if (getBits(GlobalFeatureData.maxChildNum) > flagOff
				|| getBits(GlobalFeatureData.maxBinLength) > flagOff
				|| getBits(GlobalFeatureData.maxChildNum) > flagOff) {
			System.out.println("size too large 5");
			System.exit(0);
		}
	}
	
	public String findLang(String file) {
		String[] possibleLang = {"arabic", "bulgarian", "chinese", "czech", "danish", "dutch", 
				"english08", "german", "japanese", "portuguese", "slovene", "spanish",
				"swedish", "turkish", "mlg", "kin", "spmrl"};
		for (int i = 0; i < possibleLang.length; i++)
			if (file.indexOf(possibleLang[i]) != -1) {
				return possibleLang[i];
			}
		return "";
	}

	public void loadCoarseMap(String file) throws IOException {
		int p = file.indexOf("/");
		lang = findLang(file);
		path = file.substring(0, file.indexOf("/", p + 1));
		System.out.println("language: " + lang);
		System.out.println("data path: " + path);

		String mapFile = path + "/lab/" + lang + ".uni.map";
		BufferedReader br = new BufferedReader(new FileReader(mapFile));
		coarseMap = new HashMap<String, String>();
		String str = null;
		while ((str = br.readLine()) != null) {
			String[] data = str.split("\\s+");
			coarseMap.put(data[0], data[1]);
		}
		br.close();
	}

	public void setLabeledAndLemma(String file) throws IOException {
		labeled = DependencyParser.labeled;
		if (labeled && DependencyParser.pruningMode == 1) {
			System.out.println("don't support label parsing with prediction only. Set label to false");
			DependencyParser.labeled = false;
			labeled = false;
		}
		int p = file.indexOf("/");
		lang = findLang(file);
		path = file.substring(0, file.indexOf("/", p + 1));
		System.out.println("language: " + lang);

		if (lang.equals("swedish") || lang.equals("japanese")
				|| lang.equals("german") || lang.equals("english")
				|| lang.equals("danish") || lang.equals("chinese")
				|| lang.equals("bulgarian"))
			hasLemma = false;
		else
			hasLemma = true;
		
		if (lang.equals("arabic") || 
				lang.equals("german") || lang.equals("english08") ||
				lang.equals("swedish") ||lang.equals("turkish") || lang.equals("kin"))
			hasCpos = false;
		else
			hasCpos = true;

		if (lang.equals("arabic") || lang.equals("slovene")
				|| lang.equals("chinese") || lang.equals("czech")
				|| lang.equals("dutch")) {
			ccDepType = 0;
		} else if (lang.equals("bulgarian") || lang.equals("german")
				|| lang.equals("portuguese") || lang.equals("spanish") || lang.equals("kin")) {
			ccDepType = 1;
		} else if (lang.equals("danish") || lang.equals("english08") || lang.equals("spmrl")) {
			ccDepType = 2;
		} else if (lang.equals("japanese") || lang.equals("mlg")) {
			ccDepType = 3;
		} else if (lang.equals("swedish")) {
			ccDepType = 4;
		} else if (lang.equals("turkish")) {
			ccDepType = 5;
		} else {
			System.out.println("undefined language");
			System.exit(0);
		}
		
		if ((lang.equals("german") || lang.equals("czech") || lang.equals("english08"))
				&& DependencyParser.pruningMode != 0) {
			//DependencyParser.maxLength = 70;
			//DependencyParser.maxAvgHead = 25;
			//if (lang.equals("english08")) {
			//	DependencyParser.devSplit = 10;
			//}
		}
		
		if (lang.equals("turkish")) {
			//DependencyParser.regC = 0.001;
		}
		//else if (lang.equals("spanish")) {
		//	DependencyParser.regC = 0.001;
		//}
		
		int devNum = 300;
		if (lang.equals("arabic")) {
			devNum = 110;
		}
		else if (lang.equals("slovene")) {
			devNum = 200;
		}
		else if (lang.equals("spanish")) {
			devNum = 200;
		}
		else if (lang.equals("german") || lang.equals("czech") || lang.equals("english08")) {
			devNum = 400;
		}
		if (DependencyParser.validationSize > 0)
			DependencyParser.validationSize = devNum;
	}

	public void buildChild(DependencyInstance s) {
		for (int i = 1; i < s.dep.length; i++)
			if (s.dep[i] == -1) {
				// don't build it
				return;
			}
		
		s.child = new int[s.dep.length][];
		int[] count = new int[s.dep.length];
		for (int i = 1; i < s.dep.length; i++)
			count[s.dep[i]]++;
		for (int i = 0; i < s.dep.length; i++) {
			s.child[i] = new int[count[i]];
			count[i] = 0;
		}
		for (int i = 1; i < s.dep.length; i++) {
			s.child[s.dep[i]][count[s.dep[i]]] = i;
			count[s.dep[i]]++;
		}

		buildSpan(s);
		if (DependencyParser.useNonproj)
			buildNonproj(s);
	}

	public String[] addRoot(String[] old, String label) {
		String[] ret = new String[old.length + 1];
		ret[0] = label;
		for (int i = 1; i < ret.length; i++)
			ret[i] = old[i - 1];
		return ret;
	}

	public void buildSpan(DependencyInstance s, int id) {
		s.st[id] = id;
		s.en[id] = id + 1;
		for (int i = 0; i < s.child[id].length; i++) {
			int cid = s.child[id][i];
			if (s.en[cid] == 0)
				buildSpan(s, cid);
			if (s.st[cid] < s.st[id])
				s.st[id] = s.st[cid];
			if (s.en[cid] > s.en[id])
				s.en[id] = s.en[cid];
		}
	}

	public void buildSpan(DependencyInstance s) {
		for (int i = 0; i < s.dep.length; i++) {
			s.st[i] = 0;
			s.en[i] = 0;
		}
		for (int i = 0; i < s.dep.length; i++)
			if (s.en[i] == 0)
				buildSpan(s, i);
	}
	
	public void buildNonproj(DependencyInstance s) {
		for (int i = 0; i < s.dep.length; i++) {
			s.isNonproj[i] = 0;
		}
		for (int m = 1; m < s.dep.length; m++) {
			int h = s.dep[m];
			int sm = m < h ? m : h;
			int la = m > h ? m : h;
			for (int tm = sm + 1; tm < la; tm++) {
				// head
				int th = s.dep[tm];
				if (th < sm || th > la) {
					s.isNonproj[m]++;
				}
				// child
				for (int j = 0; j < s.child[tm].length; j++) {
					int tc = s.child[tm][j];
					if (tc < sm || tc > la) {
						s.isNonproj[m]++;
					}
				}
			}
		}
	}
	
	public void updateNonproj(DependencyInstance s, int newH, int oldH, int arg) {
		// clean old head
		int sm = arg < oldH ? arg : oldH;
		int la = arg > oldH ? arg : oldH;

		for (int m = sm + 1; m < la; m++) {
			// head
			int h = s.dep[m];
			if (h < sm || h > la) {
				s.isNonproj[m]--;
			}
			// child
			for (int j = 0; j < s.child[m].length; j++) {
				int c = s.child[m][j];
				if (c < sm || c > la) {
					s.isNonproj[c]--;
				}
			}
		}
		s.isNonproj[arg] = 0;
		
		// add new head
		sm = arg < newH ? arg : newH;
		la = arg > newH ? arg : newH;
		for (int m = sm + 1; m < la; m++) {
			// head
			int h = s.dep[m];
			if (h < sm || h > la) {
				s.isNonproj[m]++;
				s.isNonproj[arg]++;
			}
			// child
			for (int j = 0; j < s.child[m].length; j++) {
				int c = s.child[m][j];
				if (c < sm || c > la) {
					s.isNonproj[c]++;
					s.isNonproj[arg]++;
				}
			}
		}
	}

	public Morphology processConllMorphology(String s, String lemma) {
		Morphology m = new Morphology();
		m.lemma = lexAlphabet.lookupIndex(lemma);

		String[] d = s.split("\\|");
		
        m.id = new int[Math.min(DependencyParser.maxFeatureNum, d.length)];
        m.val = new int[Math.min(DependencyParser.maxFeatureNum, d.length)];

        for (int i = 0; i < Math.min(DependencyParser.maxFeatureNum, d.length); i++) {
                if (d[i].isEmpty()  //){
                 				|| d[i].equals("_") || d[i].equals("-")) {
                	m.id[i] = -1;
                	m.val[i] = -1;
                	continue;
                }
                int p = d[i].indexOf("=");
                // if (p == -1) {
                // continue;
                // }
                String name = null;
                String val = null;
                if (p != -1) {
                	name = d[i].substring(0, p);
                	//val = d[i];
                	val = d[i].substring(p + 1);
                } else {
                	name = d[i];
                	//val = d[i];
                	val = "";
                }

               	int id = morphologyIDAlphabet.lookupIndex(name);
               	int valid = morphologyValueAlphabet.lookupIndex(val);
               	m.id[i] = id;
               	m.val[i] = valid;
        }
		
		return m;
	}

	public CatibMorphology processCatibMorphology(String s, String pos) {
		CatibMorphology cm = new CatibMorphology();
		String[] d = s.split("\\|");

		for (int i = 0; i < d.length; i++) {
			if (d[i].isEmpty()  //){
					|| d[i].equals("_") || d[i].equals("-")) {
				continue;
			}
			int p = d[i].indexOf("=");
			if (p == -1) {
				System.out.println("process catib morphology bug: " + s);
			}
			String name = d[i].substring(0, p);
			String val = d[i].substring(p + 1);
			if (name.equals("fgen")) {
				cm.fgen = morphologyValueAlphabet.lookupIndex(val);
			}
			else if (name.equals("fnum")) {
				cm.fnum = morphologyValueAlphabet.lookupIndex(val);
			}
			else if (name.equals("rat")) {
				cm.rat = morphologyValueAlphabet.lookupIndex(val);
			}
			else if (name.equals("det")) {
				cm.dt = val.equals("true") ? true : false;
			}
			else if (name.equals("def")) {
				cm.dt = val.equals("D") ? true : false;
			}
		}
		
		if (cm.dt)
			cm.dtPos = posAlphabet.lookupIndex("DT+" + pos);
		else
			cm.dtPos = posAlphabet.lookupIndex(pos);
		return cm;

	}

	public String normalize(String s) {
		if (lang.equals("arabic") && s.indexOf("_") != -1) {
			s = s.substring(s.indexOf("_") + 1);
		}
		if (s.equals("-LRB-"))
			s = "(";
		else if (s.equals("-RRB-"))
			s = ")";
		if (s.matches("[0-9]+|[0-9]+\\.[0-9]+|[0-9]+[0-9,]+"))
			return "<num>";

		return s;
	}
	
	public boolean isPunc(String w) {
		return w.matches("[-!\"#%&'()*,./:;?@\\[\\]_{}、]+");
		//return w.matches("[\"()]");
	}
	
	public boolean isCoord(String w) {
		if (lang.equals("turkish")) {
			if (w.equals("ve") || w.equals("veya"))
				return true;
			else
				return false;
		}
		else if (lang.equals("arabic")) {
			if (w.equals("w") || w.equals(">w") || w.equals(">n")) 
				return true;
			else
				return false;
		}
		else if (lang.equals("bulgarian")) {
			if (w.equals("и") || w.equals("или"))
				return true;
			else 
				return false;
		}
		else if (lang.equals("chinese")) {
			if (w.equals("和") || w.equals("或"))
				return true;
			else 
				return false;
		}
		else if (lang.equals("czech")) {
			if (w.equals("a") || w.equals("ale") || w.equals("i") || w.equals("nebo"))
				return true;
			else 
				return false;
		}
		else if (lang.equals("danish")) {
			if (w.equals("og") || w.equals("eller"))
				return true;
			else 
				return false;
		}
		else if (lang.equals("dutch")) {
			if (w.equals("en") || w.equals("of"))
				return true;
			else 
				return false;
		}
		else if (lang.equals("german")) {
			if (w.equals("und") || w.equals("oder"))
				return true;
			else 
				return false;
		}
		else if (lang.equals("japanese")) {
			if (w.equals("to"))
				return true;
			else 
				return false;
		}
		else if (lang.equals("portuguese")) {
			if (w.equals("e") || w.equals("ou"))
				return true;
			else 
				return false;
		}
		else if (lang.equals("slovene")) {
			if (w.equals("in") || w.equals("ali"))
				return true;
			else 
				return false;
		}
		else if (lang.equals("spanish")) {
			if (w.equals("y") || w.equals("e") || w.equals("o"))
				return true;
			else 
				return false;
		}
		else if (lang.equals("swedish")) {
			if (w.equals("och") || w.equals("eller"))
				return true;
			else 
				return false;
		}
		else if (lang.equals("kin")) {
			if (w.equals("na") || w.equals("no") || w.equals("cyangwa") || w.equals("kyangwa"))
				return true;
			else 
				return false;
		}
		else {
			return true;
		}
	}

	public DependencyInstance readConllSent(BufferedReader br, boolean labeled)
			throws IOException {
		DependencyInstance s = new DependencyInstance();
		String str = br.readLine();
		if (str == null)
			return null;
		ArrayList<String> data = new ArrayList<String>();
		while (!str.isEmpty()) {
			data.add(str);
			str = br.readLine();
		}

		int len = data.size() + 1;
		s.sentence = new int[len];
		s.pos = new int[len];
		s.posA = new int[len];
		s.labs = new int[len];
		s.dep = new int[len];
		s.m = new Morphology[len];
		s.specialPos = new SpecialPos[len];
		s.origWord = new String[len];

		s.sentence[0] = lexAlphabet.lookupIndex("<root>");
		s.pos[0] = posAlphabet.lookupIndex("<root-POS>");
		s.posA[0] = cposAlphabet.lookupIndex("<root-CPOS>");
		s.labs[0] = typeAlphabet.lookupIndex("<no-type>");
		s.dep[0] = -1;
		s.m[0] = processConllMorphology("_", "<root>");
		s.specialPos[0] = SpecialPos.OTHER;
		s.origWord[0] = "<root>";

		if (DependencyParser.isSpmrl) {
			s.cm = new CatibMorphology[len];
			s.cm[0] = processCatibMorphology("_", "<root-CPOS>");
			if (DependencyParser.correctPos) {
				s.candPos = new int[len][];
				s.candCpos = new int[len][];
				s.candSpecialPos = new SpecialPos[len][];
				s.goldPos = new int[len];
				s.currPosCandID = new int[len];

				s.candPos[0] = new int[1];
				s.candPos[0][0] = posAlphabet.lookupIndex("<root-POS>");
				s.candCpos[0] = new int[1];
				s.candCpos[0][0] = cposAlphabet.lookupIndex("<root-CPOS>");
				s.candSpecialPos[0] = new SpecialPos[1];
				s.candSpecialPos[0][0] = SpecialPos.OTHER;
				s.goldPos[0] = posAlphabet.lookupIndex("<root-POS>");
				s.currPosCandID[0] = 0;
			}
		}

		possibleLab.add("<no-type>");

		for (int i = 0; i < data.size(); i++) {
			String[] tmp = data.get(i).split("\t");
			s.origWord[i + 1] = tmp[1];
			s.sentence[i + 1] = lexAlphabet.lookupIndex(normalize(tmp[1]));
			s.labs[i + 1] = typeAlphabet.lookupIndex(labeled ? tmp[7] : "<no-type>");
			possibleLab.add(tmp[7]);
			s.dep[i + 1] = Integer.parseInt(tmp[6]);
			s.m[i + 1] = processConllMorphology(tmp[5],
					hasLemma ? normalize(tmp[2]) : normalize(tmp[1]));

			String pos = tmp[4];
			String word = tmp[1];
			
			if (DependencyParser.isSpmrl) {
				String[] tmpPos = tmp[4].split("\\|\\|");
				pos = tmpPos[0];
				if (DependencyParser.correctPos) {
					s.candPos[i + 1] = new int[tmpPos.length];
					s.candCpos[i + 1] = new int[tmpPos.length];
					s.candSpecialPos[i + 1] = new SpecialPos[tmpPos.length];
					for (int j = 0; j < tmpPos.length; j++) {
						s.candPos[i + 1][j] = posAlphabet.lookupIndex(tmpPos[j]);
						s.candCpos[i + 1][j] = cposAlphabet.lookupIndex(removeMorphology(tmpPos[j]));
						s.candSpecialPos[i + 1][j] = getCatibSpecialPos(tmpPos[j], word);
					}
					s.goldPos[i + 1] = posAlphabet.lookupIndex(tmp[3]);
					s.currPosCandID[i + 1] = 0;
				}
			}
			
			if (DependencyParser.isSpmrl) {
				s.cm[i + 1] = processCatibMorphology(tmp[5], removeMorphology(pos));
			}
			
			if (!lang.equals("spmrl") && !coarseMap.containsKey(pos)) {
				System.out.println("can't find map: " + pos);
				coarseMap.put(pos, pos);
			}
			
			//if (lang.equals("dutch")) {
			//	if (tmp[4].indexOf("_") != -1)
			//		tmp[4] = tmp[4].substring(tmp[4].lastIndexOf("_"));
			//	if (tmp[3].indexOf("_") != -1)
			//		tmp[3] = tmp[3].substring(tmp[3].lastIndexOf("_"));
			//}

			if (lang.equals("spmrl")) {
				s.specialPos[i + 1] = getCatibSpecialPos(pos, word);
			}
			else {
				String cpos = coarseMap.get(pos);
				if (cpos.equals("CONJ") && isCoord(word))
					s.specialPos[i + 1] = SpecialPos.C;
				else if (cpos.equals("ADP"))
					s.specialPos[i + 1] = SpecialPos.P;
				else if (/*cpos.equals(".") || */isPunc(word))
					s.specialPos[i + 1] = SpecialPos.PNX;
				else if (cpos.equals("VERB"))
					s.specialPos[i + 1] = SpecialPos.V;
				else
					s.specialPos[i + 1] = SpecialPos.OTHER;
			}

			s.pos[i + 1] = posAlphabet.lookupIndex(pos);
			if (lang.equals("spmrl")) {
				s.posA[i + 1] = cposAlphabet.lookupIndex(removeMorphology(pos));
			}
			else {
				if (!hasCpos && tmp[3].equals("_"))
					s.posA[i + 1] = cposAlphabet.lookupIndex(tmp[4]);
				else
					s.posA[i + 1] = cposAlphabet.lookupIndex(tmp[3]);
			}
		}

		s.st = new int[len];
		s.en = new int[len];
		if (DependencyParser.useNonproj) {
			s.isNonproj = new int[len];
		}
		buildChild(s);

		s.parserRerankScore = 0.0;
		s.length = s.dep.length;
		
		return s;
	}
	
	public SpecialPos getCatibSpecialPos(String pos, String word) {
		if (pos.equals("C") && isCoord(word))
			return SpecialPos.C;
		else if (pos.equals("P"))
			return SpecialPos.P;
		else if (pos.equals("PNX"))
			return SpecialPos.PNX;
		else if (pos.indexOf("AJ") != -1)
			return SpecialPos.AJ;
		else if (isNoun(pos))
			return SpecialPos.N;
		else
			return SpecialPos.OTHER;
	}

	public boolean isNoun(String headP) {
		return headP.equals("N") || headP.equals("PN") || headP.equals("AB");
	}

	public String removeMorphology(String str) {
		if (str.indexOf("_") != -1)
			str = str.substring(str.indexOf("_") + 1);
		if (str.equals("pron"))
			str = "PRO";
		else if (str.equals("prep"))
			str = "P";
		return str;
	}

	public void updateChildList(DependencyInstance inst, int newH, int oldH, int arg) {
		if (newH != inst.dep[arg]) {
			System.out.println("update child list bug");
		}
		// remove arg in oldH
		int[] list = new int[inst.child[oldH].length - 1];
		int p = 0;
		for (int i = 0; i < inst.child[oldH].length; i++) {
			if (inst.child[oldH][i] == arg)
				continue;
			list[p] = inst.child[oldH][i];
			p++;
		}
		inst.child[oldH] = list;
		// add arg in newH
		list = new int[inst.child[newH].length + 1];
		p = 0;
		boolean add = false;
		for (int i = 0; i < inst.child[newH].length; i++) {
			if (!add && arg < inst.child[newH][i]) {
				list[p] = arg;
				add = true;
				p++;
			}
			list[p] = inst.child[newH][i];
			p++;
		}
		if (!add) {
			list[p] = arg;
		}
		inst.child[newH] = list;

		updateSpan(inst, newH, oldH, arg);
		if (DependencyParser.useNonproj) {
			updateNonproj(inst, newH, oldH, arg);
			
			int[] tmp = new int[inst.isNonproj.length];
			for (int i = 0; i < tmp.length; i++)
				tmp[i] = inst.isNonproj[i];
			buildNonproj(inst);
			for (int i = 0; i < tmp.length; i++) {
				if (tmp[i] != inst.isNonproj[i]) {
					System.out.println("non proj bug: " + arg + " " + oldH + " " + newH + " " + i);
					for (int j = 1; j < inst.dep.length; j++) {
						System.out.println(j + "\t" + inst.isNonproj[j] + "\t" + tmp[j] + "\t" + inst.dep[j]);
					}
					System.exit(0);
				}
			}
			
		}
		/*
		int[] tmpst = new int[inst.st.length];
		int[] tmpen = new int[inst.en.length];
		for (int i = 0; i < tmpst.length; i++) {
			tmpst[i] = inst.st[i];
			tmpen[i] = inst.en[i];
		}
		buildSpan(inst);
		for (int i = 0; i < tmpst.length; i++) {
			if (tmpst[i] != inst.st[i] || tmpen[i] != inst.en[i]) {
				System.out.println("span bug");
				System.out.println(arg + "\t" + oldH + "\t" + newH);
				for (int j = 0; j < tmpst.length; j++) {
					System.out.print(j);
					for (int k = 0; k < inst.child[j].length; k++)
						System.out.print("\t" + inst.child[j][k]);
					System.out.println();
					System.out.println("\t" + tmpst[j] + "\t" + tmpen[j] + "\t" + inst.st[j] + "\t" + inst.en[j]);
				}
				System.exit(0);
			}
		}
		*/
	}

	public void updateSpan(DependencyInstance inst, int newH, int oldH, int arg) {
		int tmpH = newH;
		while (tmpH != -1) {
			inst.st[tmpH] = Math.min(inst.st[tmpH], inst.st[arg]);
			inst.en[tmpH] = Math.max(inst.en[tmpH], inst.en[arg]);
			tmpH = inst.dep[tmpH];
		}

		// assume that child list is updated
		tmpH = oldH;
		while (tmpH != -1) {
			if (inst.st[tmpH] == inst.st[arg]) {
				inst.st[tmpH] = tmpH;
				for (int i = 0; i < inst.child[tmpH].length; i++)
					inst.st[tmpH] = Math.min(inst.st[tmpH],
							inst.st[inst.child[tmpH][i]]);
			}
			if (inst.en[tmpH] == inst.en[arg]) {
				inst.en[tmpH] = tmpH + 1;
				for (int i = 0; i < inst.child[tmpH].length; i++)
					inst.en[tmpH] = Math.max(inst.en[tmpH],
							inst.en[inst.child[tmpH][i]]);
			}
			tmpH = inst.dep[tmpH];
		}
	}

	public FeatureVector add(String feat, double val, FeatureVector fv) {
		int num = dataAlphabet.lookupIndex(feat);
		if (num > 0)
			return new FeatureVector(num, val, fv);
		return fv;
	}

	public FeatureVector add(String feat, FeatureVector fv) {
		return add(feat, 1.0, fv);
	}

	public FeatureVector add(int feat, double val, FeatureVector fv) {
		if (feat > 0)
			return new FeatureVector(feat, val, fv);
		return fv;
	}
	
	public FeatureVector addCode(TemplateType type, long code, double val, FeatureVector fv) {
		int feat = dataAlphabet.lookupIndex(type, code, true);
		if (feat > 0)
			return new FeatureVector(feat, val, fv);
		return fv;
	}

	public FeatureVector addCode(TLongIntHashMap intmap, long code, double val, FeatureVector fv) {
		// avoid computation to select mapping
		int feat = dataAlphabet.lookupIndex(intmap, code, true);
		if (feat > 0)
			return new FeatureVector(feat, val, fv);
		return fv;
	}

	public void copyPredDepLab(DependencyInstance inst, int[] dep, int[] lab) {
		inst.parserPredDep = new int[dep.length];
		inst.parserPredLab = new int[lab.length];
		for (int i = 0; i < dep.length; i++) {
			inst.parserPredDep[i] = dep[i];
			inst.parserPredLab[i] = lab[i];
		}
	}

	public void buildDictionary(String goldfile) throws IOException {
		System.out.print("Creating Dictionary ... ");

		BufferedReader in = new BufferedReader(new InputStreamReader(
				new FileInputStream(goldfile), "UTF8"));

		ArrayList<Integer> length = new ArrayList<Integer>();
		DependencyInstance gold = readConllSent(in, labeled);

		int cnt = 0;

		while (gold != null) {
			if ((cnt + 1) % 1000 == 0)
				System.out.print((cnt + 1) + "  ");
			
			length.add(gold.dep.length);
			gold = readConllSent(in, labeled);
			cnt++;
			if (cnt >= DependencyParser.trainSentences)
				break;
		}
		in.close();

		System.out.println("Done.");

		setAndCheckOffset();
		buildDevSet(cnt, length);
	}
	
	public boolean isNotDev(int i) {
		// because when usedAsDev is null, it's testing file, everyone should be used in training
		return (usedAsDev == null || !usedAsDev[i]);
	}
	
	public boolean isDev(int i) {
		// because when usedAsDev is null, it's testing file, everyone should be used in development
		return (usedAsDev == null || usedAsDev[i]);
	}
	
	public void buildDevSet(int cnt, ArrayList<Integer> length) {
		if (DependencyParser.validationSize <= 0) {
			System.out.println("no validation, use test file to validate");
			return;
		}
		System.out.println("validation size: " + DependencyParser.validationSize);
		usedAsDev = new boolean[cnt];
		
		System.out.println("begin select dev");
		Random r = new Random(0);
		int pointer = 0;
		for (int i = 0; i < DependencyParser.validationSize; i++) {
			int jump = r.nextInt(cnt);
			pointer = (pointer + jump) % cnt;
			if (pointer == usedAsDev.length)
				pointer = 0;
			while (usedAsDev[pointer] || length.get(pointer) > 70) {
				pointer++;
				if (pointer == usedAsDev.length)
					pointer = 0;
			}
			if (usedAsDev[pointer])
				System.out.println("build dev set bug 1: " + pointer);
			usedAsDev[pointer] = true;
		}
		System.out.println("select dev done");
		
	}
	
	public void createAlphabet(String goldfile) throws IOException {

		buildDictionary(goldfile);

		System.out.print("Creating Alphabet ... ");

		BufferedReader in = new BufferedReader(new InputStreamReader(
				new FileInputStream(goldfile), "UTF8"));
		// int typeLength = 9; // may be figured out how to automatically get
		// later

		DependencyInstance gold = readConllSent(in, labeled);
		
		int cnt = 0;

		while (gold != null) {
			if ((cnt + 1) % 1000 == 0)
				System.out.print((cnt + 1) + "  ");
			//for (int i = 0; i < gold.labs.length; i++) {
			//	typeAlphabet.lookupIndex(gold.labs[i]);
			//}

			gold.parserPredDep = new int[gold.dep.length];
			gold.parserPredLab = new int[gold.labs.length];
			for (int i = 0; i < gold.labs.length; i++) {
				gold.parserPredLab[i] = gold.labs[i]; // build empty pred result
				gold.parserPredDep[i] = gold.dep[i];
			}
			if (DependencyParser.pruningMode == 3)
				gold.parserPredScore = new LocalFeatureData(gold.dep.length, typeAlphabet.size(), true);
			
			if (DependencyParser.learningMode == 2 || DependencyParser.learningMode == 3) {
				gold.parserRerankScore = 1.0;
			}

			if (isNotDev(cnt))
				createFeatureVector(gold);

			gold = readConllSent(in, labeled);
			cnt++;
			if (cnt >= DependencyParser.trainSentences)
				break;
		}

		closeAlphabets();

		in.close();

		System.out.println("Done.");
	}

	public void closeAlphabets() {
		dataAlphabet.stopGrowth();
		typeAlphabet.stopGrowth();
		cposAlphabet.stopGrowth();
		posAlphabet.stopGrowth();
		morphologyIDAlphabet.stopGrowth();
		morphologyValueAlphabet.stopGrowth();
		lexAlphabet.stopGrowth();

		types = new String[typeAlphabet.size()];
		Object[] keys = typeAlphabet.toArray();
		for (int i = 0; i < keys.length; i++) {
			int indx = typeAlphabet.lookupIndex((String)keys[i]);
			types[indx] = (String) keys[i];
		}
		for (int i = 1; i < types.length; i++) {
			if (types[i] == null)
				System.out.println("bug 1");
		}
		
		posName = new String[posAlphabet.size()];
		keys = posAlphabet.toArray();
		for (int i = 0; i < keys.length; i++) {
			int indx = posAlphabet.lookupIndex((String)keys[i]);
			posName[indx] = (String) keys[i];
		}
		
		System.out.println("string alphabet: " + dataAlphabet.strmap.size());
		System.out.println("arc mst alphabet: " + dataAlphabet.arcMSTMap.size());
		System.out.println("arc turbo alphabet: " + dataAlphabet.arcTurboMap.size());
		System.out.println("second order alphabet: " + dataAlphabet.secondOrderMap.size());
		System.out.println("third order alphabet: " + dataAlphabet.thirdOrderMap.size());
		System.out.println("high order alphabet: " + dataAlphabet.highOrderMap.size());
		System.out.println("catib alphabet: " + dataAlphabet.catibMap.size());
		
	}

	public FeatureVector createFeatureVector(DependencyInstance inst) {
		FeatureVector fv = end;
		fv = createLocalFeatureVector(inst, fv);
		fv = createGlobalFeatureVector(inst, fv);
		return fv;
	}

	public FeatureVector createFeatureVector(DependencyInstance inst, 
			LocalFeatureData lfd, int singleChange) {
		FeatureVector fv = end;
		fv = createLocalFeatureVectorNoCache(inst, lfd, fv, singleChange);
		fv = createGlobalFeatureVector(inst, fv);
		return fv;
	}

	public FeatureVector createFeatureVector(DependencyInstance inst,
			LocalFeatureData lfd, GlobalFeatureData gfd, int singleChange) {
		FeatureVector fv = end;
		fv = createLocalFeatureVector(inst, lfd, fv, singleChange);
		fv = createGlobalFeatureVector(inst, gfd, fv);
		return fv;
	}

	public FeatureVector createFeatureVector(DependencyInstance inst,
			LocalFeatureData lfd, GlobalFeatureData gfd, int[] multipleChange) {
		FeatureVector fv = end;
		fv = createLocalFeatureVector(inst, lfd, fv, multipleChange);
		fv = createGlobalFeatureVector(inst, gfd, fv);
		return fv;
	}

	public int[] findConjArg(DependencyInstance s, int arg) {
		// 0: head; 1:left; 2:right
		int head = -1;
		int left = -1;
		int right = -1;
		int[] deps = s.dep;

		if (ccDepType == 0) {
			// head left arg right
			//   0   2    1    2
			right = findRightNearestChild(s.child[arg], arg);
			left = findLeftNearestChild(s.child[arg], arg);
			//if (s.specialPos[left] == SpecialPos.C) {
			//	left = findLeftNearestChild(s.child[left], arg);
			//}
			head = deps[arg];
		} else if (ccDepType == 1) {
			// head left arg right
			//   0   1    2    2
			if (deps[arg] < arg) {
				left = deps[arg];
				if (left != -1) {
					right = findRightNearestChild(s.child[left], arg);
					head = deps[left];
				}
			}
		} else if (ccDepType == 2) {
			// head left arg right
			//   0   1    2    3
			if (deps[arg] < arg) {
				left = deps[arg];
				if (left != -1)
					head = deps[left];
				right = findRightNearestChild(s.child[arg], arg);
			}
		} else if (ccDepType == 3) {

		} else if (ccDepType == 4) {
			// head left arg right
			//  0    1    4    2
			if (deps[arg] > arg) {
				right = deps[arg];
				if (deps[right] < arg) {
					left = deps[right];
					if (left != -1) {
						head = deps[left];
					}
				}
			}
		} else if (ccDepType == 5) {
			// left arg right head
			//  2    3    4    0
			if (deps[arg] > arg) {
				right = deps[arg];
				left = findLeftNearestChild(s.child[arg], arg);
				if (right != -1)
					head = deps[right];
			}
		} else {
			System.out.println("undefined cc type");
			System.exit(0);
		}

		int[] ret = new int[3];
		ret[0] = head;
		ret[1] = left;
		ret[2] = right;
		return ret;
	}

	public int[] findPPArg(DependencyInstance s, int arg) {
		int c = s.child[arg].length == 0 ? -1 : s.child[arg][0];
		int c2 = -1;
		if (c != -1 && s.specialPos[c] == SpecialPos.C) {
			if (ccDepType == 0) {
				// prep is head
				c2 = findRightNearestChild(s.child[c], c);
				c = findLeftNearestChild(s.child[c], c);
			}
			else if (ccDepType == 1) {
				// prep is left, should not happen
				// c = findRightNearestChild(s.child[arg], c);
			}
			else if (ccDepType == 2) {
				// prep is left
				c = findRightNearestChild(s.child[arg], c);
			}
			// others not valid
		}
		ArrayList<Integer> tmp = new ArrayList<Integer>();
		int head = s.dep[arg];
		if (c != -1 && c != head)
			tmp.add(c);
		if (c2 != -1 && c != head)
			tmp.add(c2);
		int[] ret = new int[tmp.size()];
		for (int i = 0; i < ret.length; i++)
			ret[i] = tmp.get(i);
		return ret;
	}

	public FeatureVector createLocalFeatureVector(DependencyInstance inst,
			FeatureVector fv) {
		int[] toks = inst.sentence;
		int[] deps = inst.dep;
		int[] labs = inst.labs;

		// first order
		for (int i = 0; i < toks.length; i++) {
			if (deps[i] == -1)
				continue;
			int small = i < deps[i] ? i : deps[i];
			int large = i > deps[i] ? i : deps[i];
			boolean attR = i < deps[i] ? false : true;
			fv = createFeatureVector(inst, small, large, attR, fv);

			if (labeled) {
				fv = createFeatureVector(inst, i, labs[i], attR, true, fv);
				fv = createFeatureVector(inst, deps[i], labs[i], attR, false,
						fv);
				fv = createLabFeatureVector(inst, deps[i], i, labs[i], fv);
				// dep2
				if (DependencyParser.useDep2) {
					if (deps[i] != 0)
						fv = createFeatureVectorDep2(inst, deps[i], i,
								labs[deps[i]], labs[i], fv);
				}
			}
		}

		// sibs
		for (int i = 0; i < toks.length; i++) {
			if (deps[i] == -1 && i != 0)
				continue;
			int aid = findRightNearestChildID(inst.child[i], i);
			// right children
			int prev = i;
			for (int j = aid; j < inst.child[i].length; j++) {
				int curr = inst.child[i][j];
				if (DependencyParser.useSecondOrder) {
					fv = createFeatureVector(inst, i, prev, curr, fv);
					fv = createFeatureVectorSib(inst, prev, curr, prev == i, fv);
				}

				if (DependencyParser.useThirdOrder) {
					if (deps[i] != -1) {
						fv = createFeatureVectorGPSib(inst, deps[i], i, prev, curr,
								fv);
					}
	
					if (j + 1 < inst.child[i].length) {
						int next = inst.child[i][j + 1];
						fv = createFeatureVectorTriSib(inst, i, prev, curr, next,
								fv);
					}
				}

				// parent-sibling-child
				if (DependencyParser.usePSC) {
					if (prev != i) {
						for (int k = 0; k < inst.child[prev].length; k++)
							fv = createFeatureVectorPSC(inst, i, prev,
									inst.child[prev][k], curr, fv);
						for (int k = 0; k < inst.child[curr].length; k++)
							fv = createFeatureVectorPSC(inst, i, curr,
									inst.child[curr][k], prev, fv);
					}
				}

				prev = curr;
			}

			// left children
			prev = i;
			for (int j = aid - 1; j >= 0; j--) {
				int curr = inst.child[i][j];
				
				if (DependencyParser.useSecondOrder) {
					fv = createFeatureVector(inst, i, prev, curr, fv);
					fv = createFeatureVectorSib(inst, prev, curr, prev == i, fv);
				}

				if (DependencyParser.useThirdOrder) {
					if (deps[i] != -1) {
						fv = createFeatureVectorGPSib(inst, deps[i], i, prev, curr,
								fv);
					}
	
					if (j > 0) {
						int next = inst.child[i][j - 1];
						fv = createFeatureVectorTriSib(inst, i, prev, curr, next,
								fv);
					}
				}
				
				// parent-sibling-child
				if (DependencyParser.usePSC) {
					if (prev != i) {
						for (int k = 0; k < inst.child[prev].length; k++)
							fv = createFeatureVectorPSC(inst, i, prev,
									inst.child[prev][k], curr, fv);
						for (int k = 0; k < inst.child[curr].length; k++)
							fv = createFeatureVectorPSC(inst, i, curr,
									inst.child[curr][k], prev, fv);
					}
				}

				prev = curr;
			}

			// arbitrary sib
			
			if (DependencyParser.useAS) {
				for (int j = 0; j < inst.child[i].length; j++) { 
					for (int k = j + 1; k < inst.child[i].length; k++) { 
						fv = createFeatureVectorArbSib(inst, i, inst.child[i][j],
								inst.child[i][k], fv); 
					} 
				}
			}
		}

		// head bigram
		if (DependencyParser.useHB) {
			for (int i = 0; i < toks.length - 1; i++) {
				if (deps[i] == -1 || deps[i + 1] == -1)
					continue;
				fv = createFeatureVectorHeadBi(inst, i, deps[i], deps[i + 1],
						fv);
			}
		}

		// grandparent-parent-child
		for (int i = 0; i < toks.length; i++) {
			if (deps[i] == -1 || deps[deps[i]] == -1)
				continue;
			int p = deps[i];
			int gp = deps[p];
			if (DependencyParser.useSecondOrder) {
				fv = createFeatureVectorGPC(inst, gp, p, i, fv);
			}

			// ggp->gp->p->c
			if (!DependencyParser.useGGPC || deps[gp] == -1)
				continue;
			int ggp = deps[gp];
			fv = createFeatureVectorGGPC(inst, ggp, gp, p, i, fv);
		}
		
		if (DependencyParser.useHO) {
			for (int i = 0; i < toks.length; i++) {
				// pp attachment
				if (inst.specialPos[i] == SpecialPos.P) {
					int par = deps[i];
					int[] c = findPPArg(inst, i);
					for (int z = 0; z < c.length; z++) {
						if (par != -1 && c[z] != -1) {
							fv = createFeatureVectorPP(inst, par, i, c[z], fv);
						}
					}
				}
				// conjunction pos
				if (inst.specialPos[i] == SpecialPos.C) {
					int[] arg = findConjArg(inst, i);
					int head = arg[0];
					int left = arg[1];
					int right = arg[2];
					if (left != -1 && right != -1 && left < right) {
						fv = createFeatureVectorCC1(inst, left, i, right, fv);
						if (head != -1) {
							fv = createFeatureVectorCC2(inst, i, head, left, fv);
							fv = createFeatureVectorCC2(inst, i, head, right, fv);
						}
					}
				}
				// punc head
				if (inst.specialPos[i] == SpecialPos.PNX) {
					int j = findPuncCounterpart(inst.sentence, i);
					if (j != -1 && deps[i] == deps[j])
						fv = createFeatureVectorPNX(inst, deps[i], i, j, fv);
				}
				
				// non-proj
				if (DependencyParser.useNonproj && deps[i] != -1) {
					fv = createFeatureVectorNonproj(inst, deps[i], i, fv);
				}
			}
		}
		
		return fv;
	}

	public FeatureVector createLocalFeatureVectorNoCache(DependencyInstance inst, 
			LocalFeatureData lfd, FeatureVector fv, int singleChange) {
		// create and save in the cache
		
		int[] toks = inst.sentence;
		int[] deps = inst.dep;
		int[] labs = inst.labs;

		// first order
		for (int i = 0; i < toks.length; i++) {
			if (deps[i] == -1)
				continue;
			if (singleChange != -1 && i != singleChange)
				continue;
			int small = i < deps[i] ? i : deps[i];
			int large = i > deps[i] ? i : deps[i];
			boolean attR = i < deps[i] ? false : true;
			fv = createFeatureVector(inst, small, large, attR, fv);

			if (labeled) {
				int type = labs[i];
				if (type <= 0) {
					//System.out.println("unseen type");
					type = NOTYPE;
				}
				fv = createFeatureVector(inst, deps[i], type, attR, false, fv);
				fv = createFeatureVector(inst, i, type, attR, true, fv);
				fv = createLabFeatureVector(inst, deps[i], i, type, fv);
				// dep2
				if (DependencyParser.useDep2) {
					if (deps[i] != 0) {
						int typehead = labs[deps[i]];
						if (typehead <= 0) {
							typehead = NOTYPE;
						}
						fv = createFeatureVectorDep2(inst, deps[i], i,
								labs[deps[i]], labs[i], fv);
					}
				}
			}
		}

		// sibs
		for (int i = 0; i < toks.length; i++) {
			if (deps[i] == -1 && i != 0)
				continue;
			int aid = findRightNearestChildID(inst.child[i], i);

			// right children
			int prev = i;
			for (int j = aid; j < inst.child[i].length; j++) {
				int curr = inst.child[i][j];
				
				if (DependencyParser.useSecondOrder) {
					if (singleChange == -1
							|| (singleChange >= prev && singleChange <= curr)) {
						fv = createFeatureVector(inst, i, prev, curr, fv);
						fv = createFeatureVectorSib(inst, prev, curr, prev == i, fv);
					}
				}
				
				if (DependencyParser.useThirdOrder) {
					if (deps[i] != -1) {
						if (singleChange == -1 || singleChange == i
								|| (singleChange >= prev && singleChange <= curr)) {
							fv = createFeatureVectorGPSib(inst, deps[i], i, prev, curr,
									fv);
						}
					}
	
					if (j + 1 < inst.child[i].length) {
						int next = inst.child[i][j + 1];
						if (singleChange == -1
								|| (singleChange >= prev && singleChange <= next)) {
							fv = createFeatureVectorTriSib(inst, i, prev, curr, next,
									fv);
						}
					}
				}

				// parent-sibling-child
				if (DependencyParser.usePSC) {
					if (prev != i) {
						for (int k = 0; k < inst.child[prev].length; k++) {
							int c = inst.child[prev][k];
							if (singleChange == -1
									|| (singleChange >= prev && singleChange <= curr)
									|| singleChange == c) {
								fv = createFeatureVectorPSC(inst, i, prev,
										inst.child[prev][k], curr, fv);
							}
						}
						for (int k = 0; k < inst.child[curr].length; k++) {
							int c = inst.child[curr][k];
							if (singleChange == -1
									|| (singleChange >= prev && singleChange <= curr)
									|| singleChange == c) {
								fv = createFeatureVectorPSC(inst, i, curr,
										inst.child[curr][k], prev, fv);
							}
						}
					}
				}

				prev = curr;
			}

			// left children
			prev = i;
			for (int j = aid - 1; j >= 0; j--) {
				int curr = inst.child[i][j];
				
				if (DependencyParser.useSecondOrder) {
					if (singleChange == -1
							|| (singleChange >= curr && singleChange <= prev)) {
						fv = createFeatureVector(inst, i, prev, curr, fv);
						fv = createFeatureVectorSib(inst, prev, curr, prev == i, fv);
					}
				}

				if (DependencyParser.useThirdOrder) {
					if (deps[i] != -1) {
						if (singleChange == -1 || singleChange == i
								|| (singleChange >= curr && singleChange <= prev)) {
							fv = createFeatureVectorGPSib(inst, deps[i], i, prev, curr,
									fv);
						}
					}
	
					if (j > 0) {
						int next = inst.child[i][j - 1];
						if (singleChange == -1
								|| (singleChange >= next && singleChange <= prev)) {
							fv = createFeatureVectorTriSib(inst, i, prev, curr, next,
									fv);
						}
					}
				}

				// parent-sibling-child

				if (DependencyParser.usePSC) {
					if (prev != i) {
						for (int k = 0; k < inst.child[prev].length; k++) {
							int c = inst.child[prev][k];
							if (singleChange == -1
									|| (singleChange >= curr && singleChange <= prev)
									|| singleChange == c) {
								fv = createFeatureVectorPSC(inst, i, prev,
										inst.child[prev][k], curr, fv);
							}
						}
						for (int k = 0; k < inst.child[curr].length; k++) {
							int c = inst.child[curr][k];
							if (singleChange == -1
									|| (singleChange >= curr && singleChange <= prev)
									|| singleChange == c) {
								fv = createFeatureVectorPSC(inst, i, curr,
										inst.child[curr][k], prev, fv);
							}
						}
					}
				}

				prev = curr;
			}

			// arbitrary sib

			if (DependencyParser.useAS) {
				for (int j = 0; j < inst.child[i].length; j++) { 
					for (int k = j + 1; k < inst.child[i].length; k++) { 
						int ch1 = inst.child[i][j];
						int ch2 = inst.child[i][k]; 
						if (singleChange == -1 || singleChange == ch1 || singleChange == ch2) { 
							fv = createFeatureVectorArbSib(inst, i, ch1, ch2, fv); 
						} 
					}
				}
			}
			
		}

		// head bigram
		if (DependencyParser.useHB) {
			for (int i = 0; i < toks.length - 1; i++) {
				if (deps[i] == -1 || deps[i + 1] == -1)
					continue;
				if (singleChange == -1 || singleChange == i
						|| singleChange == i + 1) {
					fv = createFeatureVectorHeadBi(inst, i, deps[i], deps[i + 1],
							fv);
				}
			}
		}

		// grandparent-parent-child
		for (int i = 0; i < toks.length; i++) {
			if (deps[i] == -1 || deps[deps[i]] == -1)
				continue;
			int p = deps[i];
			int gp = deps[p];
			if (DependencyParser.useSecondOrder) {
				if (singleChange == -1 || singleChange == p || singleChange == i) {
					fv = createFeatureVectorGPC(inst, gp, p, i, fv);
				}
			}
			
			// ggp->gp->p->c
			if (!DependencyParser.useGGPC || deps[gp] == -1)
				continue;
			int ggp = deps[gp];
			if (singleChange == -1 || singleChange == p || singleChange == i
					|| singleChange == gp) {
				fv = createFeatureVectorGGPC(inst, ggp, gp, p, i, fv);
			}
		}

		if (DependencyParser.useHO) {
			for (int i = 0; i < toks.length; i++) {
				int par = deps[i];
				// pp attachment
				if (inst.specialPos[i] == SpecialPos.P) {
					int[] c = findPPArg(inst, i);
					for (int z = 0; z < c.length; z++) {
						if (par != -1 && c[z] != -1) {
							fv = createFeatureVectorPP(inst, par, i, c[z], fv);
						}
					}
				}
				// conjunction pos
				if (inst.specialPos[i] == SpecialPos.C) {
					int[] arg = findConjArg(inst, i);
					int head = arg[0];
					int left = arg[1];
					int right = arg[2];
					if (left != -1 && right != -1 && left < right) {
						fv = createFeatureVectorCC1(inst, left, i, right, fv);
						if (head != -1) {
							fv = createFeatureVectorCC2(inst, i, head, left, fv);
							fv = createFeatureVectorCC2(inst, i, head, right, fv);
						}
					}
				}
				// punc head
				if (inst.specialPos[i] == SpecialPos.PNX) {
					int j = lfd.puncPair[i];
					if (j != -1 && par == deps[j]) {
						fv = createFeatureVectorPNX(inst, deps[i], i, j, fv);
					}
				}
				
				// non-proj
				if (DependencyParser.useNonproj && deps[i] != -1) {
					fv = createFeatureVectorNonproj(inst, deps[i], i, fv);
				}
			}
		}

		return fv;
	}

	public FeatureVector createLocalFeatureVector(DependencyInstance inst,
			LocalFeatureData lfd, FeatureVector fv, int singleChange) {
		// create and save in the cache
		
		int[] toks = inst.sentence;
		int[] deps = inst.dep;
		int[] labs = inst.labs;

		// first order
		for (int i = 0; i < toks.length; i++) {
			if (deps[i] == -1)
				continue;
			if (singleChange != -1 && i != singleChange)
				continue;
			int small = i < deps[i] ? i : deps[i];
			int large = i > deps[i] ? i : deps[i];
			boolean attR = i < deps[i] ? false : true;
			if (lfd.fvs[deps[i]][i] == null) {
				lfd.fvs[deps[i]][i] = createFeatureVector(inst, small, large, attR, end);
			}
			fv = fv.concat(lfd.fvs[deps[i]][i]);

			if (labeled) {
				int dir = deps[i] < i ? 0 : 1;
				int type = labs[i];
				if (type <= 0) {
					//System.out.println("unseen type");
					type = NOTYPE;
				}
				if (lfd.fvs_lab[deps[i]][type][dir][1] == null) {
					lfd.fvs_lab[deps[i]][type][dir][1] = 
							createFeatureVector(inst, deps[i], type, attR, false, end);
				}
				if (lfd.fvs_lab[i][type][dir][0] == null) {
					lfd.fvs_lab[i][type][dir][0] =
							createFeatureVector(inst, i, type, attR, true, end);
				}
				if (lfd.fvs_dep[deps[i]][i][type] == null) {
					lfd.fvs_dep[deps[i]][i][type] = 
							createLabFeatureVector(inst, deps[i], i, type, end);
				}
				fv = fv.concat(lfd.fvs_lab[deps[i]][type][dir][1]).concat(
						lfd.fvs_lab[i][type][dir][0]);
				fv = fv.concat(lfd.fvs_dep[deps[i]][i][type]);
				// dep2
				if (DependencyParser.useDep2) {
					if (deps[i] != 0) {
						int id = lfd.dep2id[deps[i]][i];
						int typehead = labs[deps[i]];
						if (typehead <= 0) {
							typehead = NOTYPE;
						}
						if (lfd.fvs_dep2[id][typehead][type] == null) {
							lfd.fvs_dep2[id][typehead][type] = 
									createFeatureVectorDep2(inst, deps[i], i, typehead, type, end);
						}
						fv = fv.concat(lfd.fvs_dep2[id][typehead][type]);
					}
				}
			}
		}

		// sibs
		for (int i = 0; i < toks.length; i++) {
			if (deps[i] == -1 && i != 0)
				continue;
			int aid = findRightNearestChildID(inst.child[i], i);

			// right children
			int prev = i;
			for (int j = aid; j < inst.child[i].length; j++) {
				int curr = inst.child[i][j];
				
				if (DependencyParser.useSecondOrder) {
					if (singleChange == -1
							|| (singleChange >= prev && singleChange <= curr)) {
						if (lfd.fvs_sibs[prev][curr][prev == i ? 0 : 1] == null) {
							lfd.fvs_sibs[prev][curr][prev == i ? 0 : 1] =
									createFeatureVectorSib(inst, prev, curr, prev == i, end);
						}
						if (lfd.fvs_trips[i][prev][curr] == null) {
							lfd.fvs_trips[i][prev][curr] =
									createFeatureVector(inst, i, prev, curr, end);	
						}
					
						fv = fv.concat(lfd.fvs_trips[i][prev][curr]).concat(
								lfd.fvs_sibs[prev][curr][prev == i ? 0 : 1]);
					}
				}
				
				if (DependencyParser.useThirdOrder) {
					if (deps[i] != -1) {
						int id = lfd.dep2id[deps[i]][i];
						if (singleChange == -1 || singleChange == i
								|| (singleChange >= prev && singleChange <= curr)) {
							if (lfd.fvs_gpsib[id][prev][curr] == null) {
								lfd.fvs_gpsib[id][prev][curr] = 
										createFeatureVectorGPSib(inst, deps[i], i, prev, curr, end);
							}
							fv = fv.concat(lfd.fvs_gpsib[id][prev][curr]);
						}
					}
	
					if (j + 1 < inst.child[i].length) {
						int next = inst.child[i][j + 1];
						int id = lfd.dep2id[i][curr];
						if (singleChange == -1
								|| (singleChange >= prev && singleChange <= next)) {
							if (lfd.fvs_trisib[id][prev][next] == null) {
								lfd.fvs_trisib[id][prev][next] = 
										createFeatureVectorTriSib(inst, i, prev, curr, next, end);
							}
							fv = fv.concat(lfd.fvs_trisib[id][prev][next]);
						}
					}
				}

				// parent-sibling-child
				if (DependencyParser.usePSC) {
					if (prev != i) {
						for (int k = 0; k < inst.child[prev].length; k++) {
							int c = inst.child[prev][k];
							if (singleChange == -1
									|| (singleChange >= prev && singleChange <= curr)
									|| singleChange == c) {
								int id1 = lfd.dep2id[i][curr];
								int id2 = lfd.dep2id[prev][c];
								if (lfd.fvs_psc[id1][id2] == null) {
									lfd.fvs_psc[id1][id2] = 
											createFeatureVectorPSC(inst, i, prev, c, curr, end);
								}
								fv = fv.concat(lfd.fvs_psc[id1][id2]);
							}
						}
						for (int k = 0; k < inst.child[curr].length; k++) {
							int c = inst.child[curr][k];
							if (singleChange == -1
									|| (singleChange >= prev && singleChange <= curr)
									|| singleChange == c) {
								int id1 = lfd.dep2id[i][prev];
								int id2 = lfd.dep2id[curr][c];
								if (lfd.fvs_psc[id1][id2] == null) {
									lfd.fvs_psc[id1][id2] = 
											createFeatureVectorPSC(inst, i, curr, c, prev, end);
								}
								fv = fv.concat(lfd.fvs_psc[id1][id2]);
							}
						}
					}
				}

				prev = curr;
			}

			// left children
			prev = i;
			for (int j = aid - 1; j >= 0; j--) {
				int curr = inst.child[i][j];
				
				if (DependencyParser.useSecondOrder) {
					if (singleChange == -1
							|| (singleChange >= curr && singleChange <= prev)) {
						if (lfd.fvs_sibs[prev][curr][prev == i ? 0 : 1] == null) {
							lfd.fvs_sibs[prev][curr][prev == i ? 0 : 1] =
									createFeatureVectorSib(inst, prev, curr, prev == i, end);
						}
						if (lfd.fvs_trips[i][prev][curr] == null) {
							lfd.fvs_trips[i][prev][curr] =
									createFeatureVector(inst, i, prev, curr, end);	
						}
	
						fv = fv.concat(lfd.fvs_trips[i][prev][curr]).concat(
								lfd.fvs_sibs[prev][curr][prev == i ? 0 : 1]);
					}
				}

				if (DependencyParser.useThirdOrder) {
					if (deps[i] != -1) {
						int id = lfd.dep2id[deps[i]][i];
						if (singleChange == -1 || singleChange == i
								|| (singleChange >= curr && singleChange <= prev)) {
							if (lfd.fvs_gpsib[id][prev][curr] == null) {
								lfd.fvs_gpsib[id][prev][curr] = 
										createFeatureVectorGPSib(inst, deps[i], i, prev, curr, end);
							}
							fv = fv.concat(lfd.fvs_gpsib[id][prev][curr]);
						}
					}
	
					if (j > 0) {
						int next = inst.child[i][j - 1];
						int id = lfd.dep2id[i][curr];
						if (singleChange == -1
								|| (singleChange >= next && singleChange <= prev)) {
							if (lfd.fvs_trisib[id][prev][next] == null) {
								lfd.fvs_trisib[id][prev][next] = 
										createFeatureVectorTriSib(inst, i, prev, curr, next, end);
							}
							fv = fv.concat(lfd.fvs_trisib[id][prev][next]);
						}
					}
				}

				// parent-sibling-child

				if (DependencyParser.usePSC) {
					if (prev != i) {
						for (int k = 0; k < inst.child[prev].length; k++) {
							int c = inst.child[prev][k];
							if (singleChange == -1
									|| (singleChange >= curr && singleChange <= prev)
									|| singleChange == c) {
								int id1 = lfd.dep2id[i][curr];
								int id2 = lfd.dep2id[prev][c];
								if (lfd.fvs_psc[id1][id2] == null) {
									lfd.fvs_psc[id1][id2] = 
											createFeatureVectorPSC(inst, i, prev, c, curr, end);
								}
								fv = fv.concat(lfd.fvs_psc[id1][id2]);
							}
						}
						for (int k = 0; k < inst.child[curr].length; k++) {
							int c = inst.child[curr][k];
							if (singleChange == -1
									|| (singleChange >= curr && singleChange <= prev)
									|| singleChange == c) {
								int id1 = lfd.dep2id[i][prev];
								int id2 = lfd.dep2id[curr][c];
								if (lfd.fvs_psc[id1][id2] == null) {
									lfd.fvs_psc[id1][id2] = 
											createFeatureVectorPSC(inst, i, curr, c, prev, end);
								}
								fv = fv.concat(lfd.fvs_psc[id1][id2]);
							}
						}
					}
				}

				prev = curr;
			}

			// arbitrary sib
			
			if (DependencyParser.useAS) {
				for (int j = 0; j < inst.child[i].length; j++) { 
					for (int k = j + 1; k < inst.child[i].length; k++) { 
						int ch1 = inst.child[i][j];
						int ch2 = inst.child[i][k]; 
						if (singleChange == -1 || singleChange == ch1 || singleChange == ch2) { 
							int id = lfd.dep2id[i][ch1]; 
							if (lfd.fvs_arbsib[id][ch2] == null) {
								lfd.fvs_arbsib[id][ch2] = createFeatureVectorArbSib(inst, i, ch1, ch2, end); 
							}
							fv = fv.concat(lfd.fvs_arbsib[id][ch2]); 
						} 
					}
				}
			}
			
		}

		// head bigram
		if (DependencyParser.useHB) {
			for (int i = 0; i < toks.length - 1; i++) {
				if (deps[i] == -1 || deps[i + 1] == -1)
					continue;
				if (singleChange == -1 || singleChange == i
						|| singleChange == i + 1) {
					int id1 = lfd.dep2id[deps[i]][i];
					if (lfd.fvs_headbi[id1][deps[i + 1]] == null) {
						lfd.fvs_headbi[id1][deps[i + 1]] = 
								createFeatureVectorHeadBi(inst, i, deps[i], deps[i + 1], end);
					}
					fv = fv.concat(lfd.fvs_headbi[id1][deps[i + 1]]);
				}
			}
		}

		// grandparent-parent-child
		for (int i = 0; i < toks.length; i++) {
			if (deps[i] == -1 || deps[deps[i]] == -1)
				continue;
			int p = deps[i];
			int gp = deps[p];
			if (DependencyParser.useSecondOrder) {
				if (singleChange == -1 || singleChange == p || singleChange == i) {
					if (lfd.fvs_gpc[gp][p][i] == null) {
						lfd.fvs_gpc[gp][p][i] = createFeatureVectorGPC(inst, gp, p, i, end);
					}
					fv = fv.concat(lfd.fvs_gpc[gp][p][i]);
				}
			}
			
			// ggp->gp->p->c
			if (!DependencyParser.useGGPC || deps[gp] == -1)
				continue;
			int ggp = deps[gp];
			if (singleChange == -1 || singleChange == p || singleChange == i
					|| singleChange == gp) {
				int id1 = lfd.dep2id[ggp][gp];
				int id2 = lfd.dep2id[p][i];
				if (lfd.fvs_ggpc[id1][id2] == null) {
					lfd.fvs_ggpc[id1][id2] = createFeatureVectorGGPC(inst, ggp, gp, p, i, end);
				}
				fv = fv.concat(lfd.fvs_ggpc[id1][id2]);
			}
		}

		if (DependencyParser.useHO) {
			for (int i = 0; i < toks.length; i++) {
				int par = deps[i];
				// pp attachment
				if (inst.specialPos[i] == SpecialPos.P) {
					int[] c = findPPArg(inst, i);
					for (int z = 0; z < c.length; z++) {
						if (par != -1 && c[z] != -1) {
							if (lfd.fvs_ppcc1[par][i][c[z]] == null) {
								lfd.fvs_ppcc1[par][i][c[z]] = 
										createFeatureVectorPP(inst, par, i, c[z], end);
							}
							fv = fv.concat(lfd.fvs_ppcc1[par][i][c[z]]);
						}
					}
				}
				// conjunction pos
				if (inst.specialPos[i] == SpecialPos.C) {
					int[] arg = findConjArg(inst, i);
					int head = arg[0];
					int left = arg[1];
					int right = arg[2];
					if (left != -1 && right != -1 && left < right) {
						if (lfd.fvs_ppcc1[left][i][right] == null) {
							lfd.fvs_ppcc1[left][i][right] = 
									createFeatureVectorCC1(inst, left, i, right, end);
						}
						fv = fv.concat(lfd.fvs_ppcc1[left][i][right]);
						if (head != -1) {
							if (lfd.fvs_cc2[i][head][left] == null) {
								lfd.fvs_cc2[i][head][left] = 
										createFeatureVectorCC2(inst, i, head, left, end);
							}
							if (lfd.fvs_cc2[i][head][right] == null) {
								lfd.fvs_cc2[i][head][right] = 
										createFeatureVectorCC2(inst, i, head, right, end);
							}
 							fv = fv.concat(lfd.fvs_cc2[i][head][left]).concat(
								lfd.fvs_cc2[i][head][right]);
						}
					}
				}
				// punc head
				if (inst.specialPos[i] == SpecialPos.PNX) {
					int j = lfd.puncPair[i];
					if (j != -1 && par == deps[j]) {
						if (lfd.fvs_ppcc1[par][i][j] == null) {
							lfd.fvs_ppcc1[par][i][j] = 
									createFeatureVectorPNX(inst, par, i, j, end);
						}
						fv = fv.concat(lfd.fvs_ppcc1[par][i][j]);
					}
				}
				
				// non-proj
				if (DependencyParser.useNonproj && deps[i] != -1) {
					int id = lfd.dep2id[deps[i]][i];
					int num = getBinnedDistance(inst.isNonproj[i]) - 1;
					if (lfd.fvs_nonproj[id][num] == null) {
						lfd.fvs_nonproj[id][num] = createFeatureVectorNonproj(inst, deps[i], i, end);
					}
					fv = fv.concat(lfd.fvs_nonproj[id][num]);
				}
			}
		}

		return fv;
	}

	public boolean changeArc(int[] multipleChange, int i) {
		for (int z = 0; z < multipleChange.length; z++)
			if (multipleChange[z] == -1 || multipleChange[z] == i)
				return true;
		return false;
	}
	
	public boolean changeNextSib(int[] multipleChange, int prev, int curr) {
		int sm = prev < curr ? prev : curr;
		int la = prev > curr ? prev : curr;
		for (int z = 0; z < multipleChange.length; z++) {
			int id = multipleChange[z];
			if (id == -1 || (id >= sm && id <= la))
				return true;
		}
		return false;
	}
	
	public boolean changeGPSib(int[] multipleChange, int i, int prev, int curr) {
		int sm = prev < curr ? prev : curr;
		int la = prev > curr ? prev : curr;
		for (int z = 0; z < multipleChange.length; z++) {
			int id = multipleChange[z];
			if (id == -1 || id == i || (id >= sm && id <= la))
				return true;
		}
		return false;
	}
	
	public boolean changeTwo(int[] multipleChange, int i, int j) {
		for (int z = 0; z < multipleChange.length; z++)
			if (multipleChange[z] == -1 || multipleChange[z] == i || multipleChange[z] == j)
				return true;
		return false;
	}
	
	public boolean changeThree(int[] multipleChange, int i, int j, int k) {
		for (int z = 0; z < multipleChange.length; z++)
			if (multipleChange[z] == -1 || multipleChange[z] == i || multipleChange[z] == j|| multipleChange[z] == k)
				return true;
		return false;
	}
	
	public FeatureVector createLocalFeatureVector(DependencyInstance inst,
			LocalFeatureData lfd, FeatureVector fv, int[] multipleChange) {
		// create and save in the cache
		
		int[] toks = inst.sentence;
		int[] deps = inst.dep;
		int[] labs = inst.labs;

		// first order
		for (int i = 0; i < toks.length; i++) {
			if (deps[i] == -1)
				continue;
			if (!changeArc(multipleChange, i))
				continue;
			int small = i < deps[i] ? i : deps[i];
			int large = i > deps[i] ? i : deps[i];
			boolean attR = i < deps[i] ? false : true;
			if (lfd.fvs[deps[i]][i] == null) {
				lfd.fvs[deps[i]][i] = createFeatureVector(inst, small, large, attR, end);
			}
			fv = fv.concat(lfd.fvs[deps[i]][i]);

			if (labeled) {
				int dir = deps[i] < i ? 0 : 1;
				int type = labs[i];
				if (type <= 0) {
					//System.out.println("unseen type");
					type = NOTYPE;
				}
				if (lfd.fvs_lab[deps[i]][type][dir][1] == null) {
					lfd.fvs_lab[deps[i]][type][dir][1] = 
							createFeatureVector(inst, deps[i], type, attR, false, end);
				}
				if (lfd.fvs_lab[i][type][dir][0] == null) {
					lfd.fvs_lab[i][type][dir][0] =
							createFeatureVector(inst, i, type, attR, true, end);
				}
				if (lfd.fvs_dep[deps[i]][i][type] == null) {
					lfd.fvs_dep[deps[i]][i][type] = 
							createLabFeatureVector(inst, deps[i], i, type, end);
				}
				fv = fv.concat(lfd.fvs_lab[deps[i]][type][dir][1]).concat(
						lfd.fvs_lab[i][type][dir][0]);
				fv = fv.concat(lfd.fvs_dep[deps[i]][i][type]);
				// dep2
				if (DependencyParser.useDep2) {
					if (deps[i] != 0) {
						int id = lfd.dep2id[deps[i]][i];
						int typehead = labs[deps[i]];
						if (typehead <= 0) {
							typehead = NOTYPE;
						}
						if (lfd.fvs_dep2[id][typehead][type] == null) {
							lfd.fvs_dep2[id][typehead][type] = 
									createFeatureVectorDep2(inst, deps[i], i, typehead, type, end);
						}
						fv = fv.concat(lfd.fvs_dep2[id][typehead][type]);
					}
				}
			}
		}

		// sibs
		for (int i = 0; i < toks.length; i++) {
			if (deps[i] == -1 && i != 0)
				continue;
			int aid = findRightNearestChildID(inst.child[i], i);

			// right children
			int prev = i;
			for (int j = aid; j < inst.child[i].length; j++) {
				int curr = inst.child[i][j];
				
				if (DependencyParser.useSecondOrder) {
					if (changeNextSib(multipleChange, prev, curr)) {
						if (lfd.fvs_sibs[prev][curr][prev == i ? 0 : 1] == null) {
							lfd.fvs_sibs[prev][curr][prev == i ? 0 : 1] =
									createFeatureVectorSib(inst, prev, curr, prev == i, end);
						}
						if (lfd.fvs_trips[i][prev][curr] == null) {
							lfd.fvs_trips[i][prev][curr] =
									createFeatureVector(inst, i, prev, curr, end);	
						}
					
						fv = fv.concat(lfd.fvs_trips[i][prev][curr]).concat(
								lfd.fvs_sibs[prev][curr][prev == i ? 0 : 1]);
					}
				}
				
				if (DependencyParser.useThirdOrder) {
					if (deps[i] != -1) {
						int id = lfd.dep2id[deps[i]][i];
						if (changeGPSib(multipleChange, i, prev, curr)) {
							if (lfd.fvs_gpsib[id][prev][curr] == null) {
								lfd.fvs_gpsib[id][prev][curr] = 
										createFeatureVectorGPSib(inst, deps[i], i, prev, curr, end);
							}
							fv = fv.concat(lfd.fvs_gpsib[id][prev][curr]);
						}
					}
	
					if (j + 1 < inst.child[i].length) {
						int next = inst.child[i][j + 1];
						int id = lfd.dep2id[i][curr];
						if (changeNextSib(multipleChange, prev, next)) {
							if (lfd.fvs_trisib[id][prev][next] == null) {
								lfd.fvs_trisib[id][prev][next] = 
										createFeatureVectorTriSib(inst, i, prev, curr, next, end);
							}
							fv = fv.concat(lfd.fvs_trisib[id][prev][next]);
						}
					}
				}

				// parent-sibling-child
				if (DependencyParser.usePSC) {
					if (prev != i) {
						for (int k = 0; k < inst.child[prev].length; k++) {
							int c = inst.child[prev][k];
							if (changeGPSib(multipleChange, c, prev, curr)) {
								int id1 = lfd.dep2id[i][curr];
								int id2 = lfd.dep2id[prev][c];
								if (lfd.fvs_psc[id1][id2] == null) {
									lfd.fvs_psc[id1][id2] = 
											createFeatureVectorPSC(inst, i, prev, c, curr, end);
								}
								fv = fv.concat(lfd.fvs_psc[id1][id2]);
							}
						}
						for (int k = 0; k < inst.child[curr].length; k++) {
							int c = inst.child[curr][k];
							if (changeGPSib(multipleChange, c, prev, curr)) {
								int id1 = lfd.dep2id[i][prev];
								int id2 = lfd.dep2id[curr][c];
								if (lfd.fvs_psc[id1][id2] == null) {
									lfd.fvs_psc[id1][id2] = 
											createFeatureVectorPSC(inst, i, curr, c, prev, end);
								}
								fv = fv.concat(lfd.fvs_psc[id1][id2]);
							}
						}
					}
				}

				prev = curr;
			}

			// left children
			prev = i;
			for (int j = aid - 1; j >= 0; j--) {
				int curr = inst.child[i][j];
				
				if (DependencyParser.useSecondOrder) {
					if (changeNextSib(multipleChange, prev, curr)) {
						if (lfd.fvs_sibs[prev][curr][prev == i ? 0 : 1] == null) {
							lfd.fvs_sibs[prev][curr][prev == i ? 0 : 1] =
									createFeatureVectorSib(inst, prev, curr, prev == i, end);
						}
						if (lfd.fvs_trips[i][prev][curr] == null) {
							lfd.fvs_trips[i][prev][curr] =
									createFeatureVector(inst, i, prev, curr, end);	
						}
	
						fv = fv.concat(lfd.fvs_trips[i][prev][curr]).concat(
								lfd.fvs_sibs[prev][curr][prev == i ? 0 : 1]);
					}
				}

				if (DependencyParser.useThirdOrder) {
					if (deps[i] != -1) {
						int id = lfd.dep2id[deps[i]][i];
						if (changeGPSib(multipleChange, i, prev, curr)) {
							if (lfd.fvs_gpsib[id][prev][curr] == null) {
								lfd.fvs_gpsib[id][prev][curr] = 
										createFeatureVectorGPSib(inst, deps[i], i, prev, curr, end);
							}
							fv = fv.concat(lfd.fvs_gpsib[id][prev][curr]);
						}
					}
	
					if (j > 0) {
						int next = inst.child[i][j - 1];
						int id = lfd.dep2id[i][curr];
						if (changeNextSib(multipleChange, prev, next)) {
							if (lfd.fvs_trisib[id][prev][next] == null) {
								lfd.fvs_trisib[id][prev][next] = 
										createFeatureVectorTriSib(inst, i, prev, curr, next, end);
							}
							fv = fv.concat(lfd.fvs_trisib[id][prev][next]);
						}
					}
				}

				// parent-sibling-child

				if (DependencyParser.usePSC) {
					if (prev != i) {
						for (int k = 0; k < inst.child[prev].length; k++) {
							int c = inst.child[prev][k];
							if (changeGPSib(multipleChange, c, prev, curr)) {
								int id1 = lfd.dep2id[i][curr];
								int id2 = lfd.dep2id[prev][c];
								if (lfd.fvs_psc[id1][id2] == null) {
									lfd.fvs_psc[id1][id2] = 
											createFeatureVectorPSC(inst, i, prev, c, curr, end);
								}
								fv = fv.concat(lfd.fvs_psc[id1][id2]);
							}
						}
						for (int k = 0; k < inst.child[curr].length; k++) {
							int c = inst.child[curr][k];
							if (changeGPSib(multipleChange, c, prev, curr)) {
								int id1 = lfd.dep2id[i][prev];
								int id2 = lfd.dep2id[curr][c];
								if (lfd.fvs_psc[id1][id2] == null) {
									lfd.fvs_psc[id1][id2] = 
											createFeatureVectorPSC(inst, i, curr, c, prev, end);
								}
								fv = fv.concat(lfd.fvs_psc[id1][id2]);
							}
						}
					}
				}

				prev = curr;
			}

			// arbitrary sib
			
			if (DependencyParser.useAS) {
				for (int j = 0; j < inst.child[i].length; j++) { 
					for (int k = j + 1; k < inst.child[i].length; k++) { 
						int ch1 = inst.child[i][j];
						int ch2 = inst.child[i][k]; 
						if (changeTwo(multipleChange, ch1, ch2)) { 
							int id = lfd.dep2id[i][ch1]; 
							if (lfd.fvs_arbsib[id][ch2] == null) {
								lfd.fvs_arbsib[id][ch2] = createFeatureVectorArbSib(inst, i, ch1, ch2, end); 
							}
							fv = fv.concat(lfd.fvs_arbsib[id][ch2]); 
						} 
					}
				}
			}
			
		}

		// head bigram
		if (DependencyParser.useHB) {
			for (int i = 0; i < toks.length - 1; i++) {
				if (deps[i] == -1 || deps[i + 1] == -1)
					continue;
				if (changeTwo(multipleChange, i, i + 1)) {
					int id1 = lfd.dep2id[deps[i]][i];
					if (lfd.fvs_headbi[id1][deps[i + 1]] == null) {
						lfd.fvs_headbi[id1][deps[i + 1]] = 
								createFeatureVectorHeadBi(inst, i, deps[i], deps[i + 1], end);
					}
					fv = fv.concat(lfd.fvs_headbi[id1][deps[i + 1]]);
				}
			}
		}

		// grandparent-parent-child
		for (int i = 0; i < toks.length; i++) {
			if (deps[i] == -1 || deps[deps[i]] == -1)
				continue;
			int p = deps[i];
			int gp = deps[p];
			if (DependencyParser.useSecondOrder) {
				if (changeTwo(multipleChange, i, p)) {
					if (lfd.fvs_gpc[gp][p][i] == null) {
						lfd.fvs_gpc[gp][p][i] = createFeatureVectorGPC(inst, gp, p, i, end);
					}
					fv = fv.concat(lfd.fvs_gpc[gp][p][i]);
				}
			}
			
			// ggp->gp->p->c
			if (!DependencyParser.useGGPC || deps[gp] == -1)
				continue;
			int ggp = deps[gp];
			if (changeThree(multipleChange, i, p, gp)) {
				int id1 = lfd.dep2id[ggp][gp];
				int id2 = lfd.dep2id[p][i];
				if (lfd.fvs_ggpc[id1][id2] == null) {
					lfd.fvs_ggpc[id1][id2] = createFeatureVectorGGPC(inst, ggp, gp, p, i, end);
				}
				fv = fv.concat(lfd.fvs_ggpc[id1][id2]);
			}
		}

		if (DependencyParser.useHO) {
			for (int i = 0; i < toks.length; i++) {
				int par = deps[i];
				// pp attachment
				if (inst.specialPos[i] == SpecialPos.P) {
					int[] c = findPPArg(inst, i);
					for (int z = 0; z < c.length; z++) {
						if (par != -1 && c[z] != -1) {
							if (lfd.fvs_ppcc1[par][i][c[z]] == null) {
								lfd.fvs_ppcc1[par][i][c[z]] = 
										createFeatureVectorPP(inst, par, i, c[z], end);
							}
							fv = fv.concat(lfd.fvs_ppcc1[par][i][c[z]]);
						}
					}
				}
				// conjunction pos
				if (inst.specialPos[i] == SpecialPos.C) {
					int[] arg = findConjArg(inst, i);
					int head = arg[0];
					int left = arg[1];
					int right = arg[2];
					if (left != -1 && right != -1 && left < right) {
						if (lfd.fvs_ppcc1[left][i][right] == null) {
							lfd.fvs_ppcc1[left][i][right] = 
									createFeatureVectorCC1(inst, left, i, right, end);
						}
						fv = fv.concat(lfd.fvs_ppcc1[left][i][right]);
						if (head != -1) {
							if (lfd.fvs_cc2[i][head][left] == null) {
								lfd.fvs_cc2[i][head][left] = 
										createFeatureVectorCC2(inst, i, head, left, end);
							}
							if (lfd.fvs_cc2[i][head][right] == null) {
								lfd.fvs_cc2[i][head][right] = 
										createFeatureVectorCC2(inst, i, head, right, end);
							}
 							fv = fv.concat(lfd.fvs_cc2[i][head][left]).concat(
								lfd.fvs_cc2[i][head][right]);
						}
					}
				}
				// punc head
				if (inst.specialPos[i] == SpecialPos.PNX) {
					int j = lfd.puncPair[i];
					if (j != -1 && par == deps[j]) {
						if (lfd.fvs_ppcc1[par][i][j] == null) {
							lfd.fvs_ppcc1[par][i][j] = 
									createFeatureVectorPNX(inst, par, i, j, end);
						}
						fv = fv.concat(lfd.fvs_ppcc1[par][i][j]);
					}
				}
				
				// non-proj
				if (DependencyParser.useNonproj && deps[i] != -1) {
					int id = lfd.dep2id[deps[i]][i];
					int num = getBinnedDistance(inst.isNonproj[i]) - 1;
					if (lfd.fvs_nonproj[id][num] == null) {
						lfd.fvs_nonproj[id][num] = createFeatureVectorNonproj(inst, deps[i], i, end);
					}
					fv = fv.concat(lfd.fvs_nonproj[id][num]);
				}
			}
		}

		return fv;
	}

	public double getLocalScore(DependencyInstance inst, LocalFeatureData lfd,
			Parameters params, int singleChange) {
		int[] toks = inst.sentence;
		int[] deps = inst.dep;
		int[] labs = inst.labs;

		double score = 0.0;

		// first order
		for (int i = 0; i < toks.length; i++) {
			if (deps[i] == -1)
				continue;
			if (singleChange != -1 && i != singleChange)
				continue;
			int small = i < deps[i] ? i : deps[i];
			int large = i > deps[i] ? i : deps[i];
			boolean attR = i < deps[i] ? false : true;

			int dir = deps[i] < i ? 0 : 1;
			int type = labs[i];

			if (lfd.fvs[deps[i]][i] == null) {
				lfd.fvs[deps[i]][i] = createFeatureVector(inst, small, large, attR, end);
				lfd.probs[deps[i]][i] = params.getScore(lfd.fvs[deps[i]][i]);
			}
			score += lfd.probs[deps[i]][i];

			if (labeled) {
				if (lfd.fvs_lab[deps[i]][type][dir][1] == null) {
					lfd.fvs_lab[deps[i]][type][dir][1] = 
							createFeatureVector(inst, deps[i], labs[i], attR, false, end);
					lfd.probs_lab[deps[i]][type][dir][1] = params.getScore(lfd.fvs_lab[deps[i]][type][dir][1]);
				}
				if (lfd.fvs_lab[i][type][dir][0] == null) {
					lfd.fvs_lab[i][type][dir][0] =
							createFeatureVector(inst, i, labs[i], attR, true, end);
					lfd.probs_lab[i][type][dir][0] = params.getScore(lfd.fvs_lab[i][type][dir][0]);
				}
				if (lfd.fvs_dep[deps[i]][i][type] == null) {
					lfd.fvs_dep[deps[i]][i][type] = 
							createLabFeatureVector(inst, deps[i], i, labs[i], end);
					lfd.probs_dep[deps[i]][i][type] = params.getScore(lfd.fvs_dep[deps[i]][i][type]);
				}

				score += lfd.probs_lab[deps[i]][type][dir][1]
						+ lfd.probs_lab[i][type][dir][0];
				score += lfd.probs_dep[deps[i]][i][type];
				// dep2
				if (DependencyParser.useDep2) {
					if (deps[i] != 0) {
						int id = lfd.dep2id[deps[i]][i];
						int typehead = labs[deps[i]];
						if (lfd.fvs_dep2[id][typehead][type] == null) {
							lfd.fvs_dep2[id][typehead][type] = 
									createFeatureVectorDep2(inst, deps[i], i, labs[deps[i]], labs[i], end);
							lfd.probs_dep2[id][typehead][type] = params.getScore(lfd.fvs_dep2[id][typehead][type]);
						}
						score += lfd.probs_dep2[id][typehead][type];
					}
				}
			}
		}

		// sibs
		for (int i = 0; i < toks.length; i++) {
			if (deps[i] == -1 && i != 0)
				continue;
			int aid = findRightNearestChildID(inst.child[i], i);

			// right children
			int prev = i;
			for (int j = aid; j < inst.child[i].length; j++) {
				int curr = inst.child[i][j];
				
				if (DependencyParser.useSecondOrder) {
					if (singleChange == -1
							|| (singleChange >= prev && singleChange <= curr)) {
						if (lfd.fvs_sibs[prev][curr][prev == i ? 0 : 1] == null) {
							lfd.fvs_sibs[prev][curr][prev == i ? 0 : 1] =
									createFeatureVectorSib(inst, prev, curr, prev == i, end);
							lfd.probs_sibs[prev][curr][prev == i ? 0 : 1] = params.getScore(lfd.fvs_sibs[prev][curr][prev == i ? 0 : 1]);
						}
						if (lfd.fvs_trips[i][prev][curr] == null) {
							lfd.fvs_trips[i][prev][curr] =
									createFeatureVector(inst, i, prev, curr, end);	
							lfd.probs_trips[i][prev][curr] = params.getScore(lfd.fvs_trips[i][prev][curr]);
						}
	
						score += lfd.probs_trips[i][prev][curr]
								+ lfd.probs_sibs[prev][curr][prev == i ? 0 : 1];
					}
				}
				
				if (DependencyParser.useThirdOrder) {
					if (deps[i] != -1) {
						int id = lfd.dep2id[deps[i]][i];
						if (singleChange == -1 || singleChange == i
								|| (singleChange >= prev && singleChange <= curr)) {
							if (lfd.fvs_gpsib[id][prev][curr] == null) {
								lfd.fvs_gpsib[id][prev][curr] = 
										createFeatureVectorGPSib(inst, deps[i], i, prev, curr, end);
								lfd.probs_gpsib[id][prev][curr] = params.getScore(lfd.fvs_gpsib[id][prev][curr]);
							}
							score += lfd.probs_gpsib[id][prev][curr];
						}
					}
	
					if (j + 1 < inst.child[i].length) {
						int next = inst.child[i][j + 1];
						int id = lfd.dep2id[i][curr];
						if (singleChange == -1
								|| (singleChange >= prev && singleChange <= next)) {
							if (lfd.fvs_trisib[id][prev][next] == null) {
								lfd.fvs_trisib[id][prev][next] = 
										createFeatureVectorTriSib(inst, i, prev, curr, next, end);
								lfd.probs_trisib[id][prev][next] = params.getScore(lfd.fvs_trisib[id][prev][next]);
							}
							score += lfd.probs_trisib[id][prev][next];
						}
					}
				}

				// parent-sibling-child

				if (DependencyParser.usePSC) {
					if (prev != i) {
						for (int k = 0; k < inst.child[prev].length; k++) {
							int c = inst.child[prev][k];
							if (singleChange == -1
									|| (singleChange >= prev && singleChange <= curr)
									|| singleChange == c) {
								int id1 = lfd.dep2id[i][curr];
								int id2 = lfd.dep2id[prev][c];
								if (lfd.fvs_psc[id1][id2] == null) {
									lfd.fvs_psc[id1][id2] = 
											createFeatureVectorPSC(inst, i, prev, c, curr, end);
									lfd.probs_psc[id1][id2] = params.getScore(lfd.fvs_psc[id1][id2]);
								}
								score += lfd.probs_psc[id1][id2];
							}
						}
						for (int k = 0; k < inst.child[curr].length; k++) {
							int c = inst.child[curr][k];
							if (singleChange == -1
									|| (singleChange >= prev && singleChange <= curr)
									|| singleChange == c) {
								int id1 = lfd.dep2id[i][prev];
								int id2 = lfd.dep2id[curr][c];
								if (lfd.fvs_psc[id1][id2] == null) {
									lfd.fvs_psc[id1][id2] = 
											createFeatureVectorPSC(inst, i, curr, c, prev, end);
									lfd.probs_psc[id1][id2] = params.getScore(lfd.fvs_psc[id1][id2]);
								}

								score += lfd.probs_psc[id1][id2];
							}
						}
					}
				}

				prev = curr;
			}

			// left children
			prev = i;
			for (int j = aid - 1; j >= 0; j--) {
				int curr = inst.child[i][j];
				
				if (DependencyParser.useSecondOrder) {
					if (singleChange == -1
							|| (singleChange >= curr && singleChange <= prev)) {
						if (lfd.fvs_sibs[prev][curr][prev == i ? 0 : 1] == null) {
							lfd.fvs_sibs[prev][curr][prev == i ? 0 : 1] =
									createFeatureVectorSib(inst, prev, curr, prev == i, end);
							lfd.probs_sibs[prev][curr][prev == i ? 0 : 1] = params.getScore(lfd.fvs_sibs[prev][curr][prev == i ? 0 : 1]);
						}
						if (lfd.fvs_trips[i][prev][curr] == null) {
							lfd.fvs_trips[i][prev][curr] =
									createFeatureVector(inst, i, prev, curr, end);	
							lfd.probs_trips[i][prev][curr] = params.getScore(lfd.fvs_trips[i][prev][curr]);
						}
						
						score += lfd.probs_trips[i][prev][curr]
								+ lfd.probs_sibs[prev][curr][prev == i ? 0 : 1];
					}
				}
				
				if (DependencyParser.useThirdOrder) {
					if (deps[i] != -1) {
						int id = lfd.dep2id[deps[i]][i];
						if (singleChange == -1 || singleChange == i
								|| (singleChange >= curr && singleChange <= prev)) {
							if (lfd.fvs_gpsib[id][prev][curr] == null) {
								lfd.fvs_gpsib[id][prev][curr] = 
										createFeatureVectorGPSib(inst, deps[i], i, prev, curr, end);
								lfd.probs_gpsib[id][prev][curr] = params.getScore(lfd.fvs_gpsib[id][prev][curr]);
							}
							score += lfd.probs_gpsib[id][prev][curr];
						}
					}
	
					if (j > 0) {
						int next = inst.child[i][j - 1];
						int id = lfd.dep2id[i][curr];
						if (singleChange == -1
								|| (singleChange >= next && singleChange <= prev)) {
							if (lfd.fvs_trisib[id][prev][next] == null) {
								lfd.fvs_trisib[id][prev][next] = 
										createFeatureVectorTriSib(inst, i, prev, curr, next, end);
								lfd.probs_trisib[id][prev][next] = params.getScore(lfd.fvs_trisib[id][prev][next]);
							}
							score += lfd.probs_trisib[id][prev][next];
						}
					}
				}

				// parent-sibling-child

				if (DependencyParser.usePSC) {
					if (prev != i) {
						for (int k = 0; k < inst.child[prev].length; k++) {
							int c = inst.child[prev][k];
							if (singleChange == -1
									|| (singleChange >= curr && singleChange <= prev)
									|| singleChange == c) {
								int id1 = lfd.dep2id[i][curr];
								int id2 = lfd.dep2id[prev][c];
								if (lfd.fvs_psc[id1][id2] == null) {
									lfd.fvs_psc[id1][id2] = 
											createFeatureVectorPSC(inst, i, prev, c, curr, end);
									lfd.probs_psc[id1][id2] = params.getScore(lfd.fvs_psc[id1][id2]);
								}
								score += lfd.probs_psc[id1][id2];
							}
						}
						for (int k = 0; k < inst.child[curr].length; k++) {
							int c = inst.child[curr][k];
							if (singleChange == -1
									|| (singleChange >= curr && singleChange <= prev)
									|| singleChange == c) {
								int id1 = lfd.dep2id[i][prev];
								int id2 = lfd.dep2id[curr][c];
								if (lfd.fvs_psc[id1][id2] == null) {
									lfd.fvs_psc[id1][id2] = 
											createFeatureVectorPSC(inst, i, curr, c, prev, end);
									lfd.probs_psc[id1][id2] = params.getScore(lfd.fvs_psc[id1][id2]);
								}
								score += lfd.probs_psc[id1][id2];
							}
						}
					}
				}

				prev = curr;
			}

			// arbitrary sib
			if (DependencyParser.useAS) {
				for (int j = 0; j < inst.child[i].length; j++) { 
					for (int k = j + 1; k < inst.child[i].length; k++) { 
						int ch1 = inst.child[i][j];
						int ch2 = inst.child[i][k]; 
						if (singleChange == -1 || singleChange == ch1 || singleChange == ch2) { 
							int id = lfd.dep2id[i][ch1]; 
							if (lfd.fvs_arbsib[id][ch2] == null) {
								lfd.fvs_arbsib[id][ch2] = createFeatureVectorArbSib(inst, i, ch1, ch2, end); 
								lfd.probs_arbsib[id][ch2] = params.getScore(lfd.fvs_arbsib[id][ch2]);
							}
							score += lfd.probs_arbsib[id][ch2]; 
						} 
					} 
				}
			}			 
		}

		// head bigram
		if (DependencyParser.useHB) {
			for (int i = 0; i < toks.length - 1; i++) {
				if (deps[i] == -1 || deps[i + 1] == -1)
					continue;
				if (singleChange == -1 || singleChange == i
						|| singleChange == i + 1) {
					int id1 = lfd.dep2id[deps[i]][i];
					if (lfd.fvs_headbi[id1][deps[i + 1]] == null) {
						lfd.fvs_headbi[id1][deps[i + 1]] = 
								createFeatureVectorHeadBi(inst, i, deps[i], deps[i + 1], end);
						lfd.probs_headbi[id1][deps[i + 1]] = params.getScore(lfd.fvs_headbi[id1][deps[i + 1]]);
					}
					score += lfd.probs_headbi[id1][deps[i + 1]];
				}
			}
		}

		// grandparent-parent-child
		for (int i = 0; i < toks.length; i++) {
			if (deps[i] == -1 || deps[deps[i]] == -1)
				continue;
			int p = deps[i];
			int gp = deps[p];

			if (DependencyParser.useSecondOrder) {
				if (singleChange == -1 || singleChange == p || singleChange == i) {
					if (lfd.fvs_gpc[gp][p][i] == null) {
						lfd.fvs_gpc[gp][p][i] = createFeatureVectorGPC(inst, gp, p, i, end);
						lfd.probs_gpc[gp][p][i] = params.getScore(lfd.fvs_gpc[gp][p][i]);
					}
	
					score += lfd.probs_gpc[gp][p][i];
				}
			}

			// ggp->gp->p->c
			if (!DependencyParser.useGGPC || deps[gp] == -1)
				continue;
			int ggp = deps[gp];
			if (singleChange == -1 || singleChange == p || singleChange == i
					|| singleChange == gp) {
				int id1 = lfd.dep2id[ggp][gp];
				int id2 = lfd.dep2id[p][i];
				if (lfd.fvs_ggpc[id1][id2] == null) {
					lfd.fvs_ggpc[id1][id2] = createFeatureVectorGGPC(inst, ggp, gp, p, i, end);
					lfd.probs_ggpc[id1][id2] = params.getScore(lfd.fvs_ggpc[id1][id2]);
				}
				score += lfd.probs_ggpc[id1][id2];
			}
		}

		if (DependencyParser.useHO) {
			for (int i = 0; i < toks.length; i++) {
				int par = deps[i];
				// pp attachment
				if (inst.specialPos[i] == SpecialPos.P) {
					int[] c = findPPArg(inst, i);
					for (int z = 0; z < c.length; z++) {
						if (par != -1 && c[z] != -1) {
							if (lfd.fvs_ppcc1[par][i][c[z]] == null) {
								lfd.fvs_ppcc1[par][i][c[z]] = 
										createFeatureVectorPP(inst, par, i, c[z], end);
								lfd.probs_ppcc1[par][i][c[z]] = params.getScore(lfd.fvs_ppcc1[par][i][c[z]]);
							}
							score += lfd.probs_ppcc1[par][i][c[z]];
						}
					}
				}
				// conjunction pos
				if (inst.specialPos[i] == SpecialPos.C) {
					int[] arg = findConjArg(inst, i);
					int head = arg[0];
					int left = arg[1];
					int right = arg[2];
					if (left != -1 && right != -1 && left < right) {
						if (lfd.fvs_ppcc1[left][i][right] == null) {
							lfd.fvs_ppcc1[left][i][right] = 
									createFeatureVectorCC1(inst, left, i, right, end);
							lfd.probs_ppcc1[left][i][right] = params.getScore(lfd.fvs_ppcc1[left][i][right]);
						}
						score += lfd.probs_ppcc1[left][i][right];
						if (head != -1) {
							if (lfd.fvs_cc2[i][head][left] == null) {
								lfd.fvs_cc2[i][head][left] = 
										createFeatureVectorCC2(inst, i, head, left, end);
								lfd.probs_cc2[i][head][left] = params.getScore(lfd.fvs_cc2[i][head][left]);
							}
							if (lfd.fvs_cc2[i][head][right] == null) {
								lfd.fvs_cc2[i][head][right] = 
										createFeatureVectorCC2(inst, i, head, right, end);
								lfd.probs_cc2[i][head][right] = params.getScore(lfd.fvs_cc2[i][head][right]);
							}
							score += lfd.probs_cc2[i][head][left]
								+ lfd.probs_cc2[i][head][right];
						}
					}
				}
				// punc head
				if (inst.specialPos[i] == SpecialPos.PNX) {
					int j = lfd.puncPair[i];
					if (j != -1 && par == deps[j]) {
						if (lfd.fvs_ppcc1[par][i][j] == null) {
							lfd.fvs_ppcc1[par][i][j] = 
									createFeatureVectorPNX(inst, par, i, j, end);
							lfd.probs_ppcc1[par][i][j] = params.getScore(lfd.fvs_ppcc1[par][i][j]);
						}
						score += lfd.probs_ppcc1[par][i][j];
					}
				}
				
				// non proj
				if (DependencyParser.useNonproj && deps[i] != -1) {
					int id = lfd.dep2id[deps[i]][i];
					int num = getBinnedDistance(inst.isNonproj[i]) - 1;
					if (lfd.fvs_nonproj[id][num] == null) {
						lfd.fvs_nonproj[id][num] = createFeatureVectorNonproj(inst, deps[i], i, end);
						lfd.probs_nonproj[id][num] = params.getScore(lfd.fvs_nonproj[id][num]);
					}
					score += lfd.probs_nonproj[id][num];
				}
			}
		}
		
		return score;
	}

	public FeatureVector createFirstOrderFeatureVector(DependencyInstance inst,
			LocalFeatureData lfd, FeatureVector fv, int arg) {
		int[] toks = inst.sentence;
		int[] deps = inst.dep;
		int[] labs = inst.labs;

		// first order
		for (int i = 0; i < toks.length; i++) {
			if (deps[i] == -1)
				continue;
			if (arg != -1 && i != arg)
				continue;
			fv = fv.concat(lfd.fvs[deps[i]][i]);

			if (labeled) {
				int dir = deps[i] < i ? 0 : 1;
				int type = labs[i];
				if (type == -1)
					type = 0;
				fv = fv.concat(lfd.fvs_lab[deps[i]][type][dir][1]).concat(
						lfd.fvs_lab[i][type][dir][0]);
				fv = fv.concat(lfd.fvs_dep[deps[i]][i][type]);
			}
		}
		return fv;
	}

	public double getFirstOrderScore(DependencyInstance inst,
			LocalFeatureData lfd, int arg) {
		int[] toks = inst.sentence;
		int[] deps = inst.dep;
		int[] labs = inst.labs;

		double score = 0.0;

		// first order
		for (int i = 0; i < toks.length; i++) {
			if (deps[i] == -1)
				continue;
			if (arg != -1 && i != arg)
				continue;
			int dir = deps[i] < i ? 0 : 1;
			int type = labs[i];

			score += lfd.probs[deps[i]][i];

			if (labeled) {
				score += lfd.probs_lab[deps[i]][type][dir][1]
						+ lfd.probs_lab[i][type][dir][0];
				score += lfd.probs_dep[deps[i]][i][type];
			}
		}
		return score;
	}

    public int getBinnedDistance(int x) {
    	int flag = 0;
    	if (x <= 0) {
    		x = -x;
    		flag = 8;
    	}
    	if (x > 10)          // x > 10
    		flag |= 0x7;
    	else if (x > 5)		 // x = 6 .. 10
    		flag |= 0x6;
    	else
    		flag |= x;   	 // x = 1 .. 5    	
    	if (flag == 0) {
    		System.out.println("bad distance");
    		new Exception().printStackTrace();
    		System.exit(0);
    	}
    	return flag;
    }
    
    public int getTurboBinnedDistance(int x) {
    	int flag = 0;
    	if (x < 0) {
    		x = -x;
    		flag = 8;
    	}
    	if (x > 40) {
    		flag |= 0x7;
    	} else if (x > 30) {
    		flag |= 0x6;
    	} else if (x > 20) {
    		flag |= 0x5;
    	} else if (x > 10) {
    		flag |= 0x4;
    	} else if (x > 5) {
    		flag |= 0x3;
    	} else if (x > 2) {
    		flag |= 0x2;
    	} else {
    		flag |= 0x1;
    	}
    	return flag;
    }
        
	public FeatureVector createFeatureVector(DependencyInstance inst,
			int small, int large, boolean attR, FeatureVector fv) {
		int[] pos = inst.pos;
		int[] posA = inst.posA;
		int[] toks = inst.sentence;
		Morphology[] m = inst.m;
		int len = toks.length;

		int headIndex = attR ? small : large;
		int childIndex = attR ? large : small;
		
		int distFlag = getBinnedDistance(headIndex - childIndex) << tempOff;
		//int distFlag = getTurboBinnedDistance(headIndex - childIndex) << tempOff;
		
		// use head->mod, rather than small->large
		int LP = pos[headIndex];
		int RP = pos[childIndex];
		int LC = posA[headIndex];
		int RC = posA[childIndex];

		int pLP = headIndex > 0 ? pos[headIndex - 1] : START;
		pLP = headIndex == childIndex + 1 ? MID : pLP;
		
		int nLP = headIndex < len - 1 ? pos[headIndex + 1] : END;
		nLP = headIndex == childIndex - 1 ? MID : nLP;

		int pRP = childIndex > 0 ? pos[childIndex - 1] : START;
		pRP = childIndex == headIndex + 1 ? MID : pRP;

		int nRP = childIndex < len - 1 ? pos[childIndex + 1] : END;
		nRP = childIndex == headIndex - 1 ? MID : nRP;

		int pLC = headIndex > 0 ? posA[headIndex - 1] : START;
		pLC = headIndex == childIndex + 1 ? MID : pLC;

		int nLC = headIndex < len - 1 ? posA[headIndex + 1] : END;
		nLC = headIndex == childIndex - 1 ? MID : nLC;

		int pRC = childIndex > 0 ? posA[childIndex - 1] : START;
		pRC = childIndex == headIndex + 1 ? MID : pRC;

		int nRC = childIndex < len - 1 ? posA[childIndex + 1] : END;
		nRC = childIndex == headIndex - 1 ? MID : nRC;
		
		long code = 0;
		

		// feature posR posMid posL
		for (int i = small + 1; i < large; i++) {
			int MP = pos[i];
			int MC = posA[i];
			code = genCodePPPF(ArcMST.LP_MP_RP.ordinal(), LP, MP, RP);
			fv = addCode(TArcMST, code, 1.0, fv);
			fv = addCode(TArcMST, code | distFlag, 1.0, fv);
			
			code = genCodePPPF(ArcMST.LC_MC_RC.ordinal(), LC, MC, RC);
			fv = addCode(TArcMST, code, 1.0, fv);
			fv = addCode(TArcMST, code | distFlag, 1.0, fv);
		}

		// feature posL-1 posL posR posR+1
		code = genCodePPPPF(ArcMST.pLP_LP_RP_nRP.ordinal(), pLP, LP, RP, nRP);
		fv = addCode(TArcMST, code, 1.0, fv);
		fv = addCode(TArcMST, code | distFlag, 1.0, fv);
		
		code = genCodePPPF(ArcMST.LP_RP_nRP.ordinal(), LP, RP, nRP);
		fv = addCode(TArcMST, code, 1.0, fv);
		fv = addCode(TArcMST, code | distFlag, 1.0, fv);
		
		code = genCodePPPF(ArcMST.pLP_RP_nRP.ordinal(), pLP, RP, nRP);
		fv = addCode(TArcMST, code, 1.0, fv);
		fv = addCode(TArcMST, code | distFlag, 1.0, fv);
		
		code = genCodePPPF(ArcMST.pLP_LP_nRP.ordinal(), pLP, LP, nRP);
		fv = addCode(TArcMST, code, 1.0, fv);
		fv = addCode(TArcMST, code | distFlag, 1.0, fv);
		
		code = genCodePPPF(ArcMST.pLP_LP_RP.ordinal(), pLP, LP, RP);
		fv = addCode(TArcMST, code, 1.0, fv);
		fv = addCode(TArcMST, code | distFlag, 1.0, fv);
		
		code = genCodePPPPF(ArcMST.pLC_LC_RC_nRC.ordinal(), pLC, LC, RC, nRC);
		fv = addCode(TArcMST, code, 1.0, fv);
		fv = addCode(TArcMST, code | distFlag, 1.0, fv);
		
		code = genCodePPPF(ArcMST.LC_RC_nRC.ordinal(), LC, RC, nRC);
		fv = addCode(TArcMST, code, 1.0, fv);
		fv = addCode(TArcMST, code | distFlag, 1.0, fv);
		
		code = genCodePPPF(ArcMST.pLC_RC_nRC.ordinal(), pLC, RC, nRC);
		fv = addCode(TArcMST, code, 1.0, fv);
		fv = addCode(TArcMST, code | distFlag, 1.0, fv);
		
		code = genCodePPPF(ArcMST.pLC_LC_nRC.ordinal(), pLC, LC, nRC);
		fv = addCode(TArcMST, code, 1.0, fv);
		fv = addCode(TArcMST, code | distFlag, 1.0, fv);
		
		code = genCodePPPF(ArcMST.pLC_LC_RC.ordinal(), pLC, LC, RC);
		fv = addCode(TArcMST, code, 1.0, fv);
		fv = addCode(TArcMST, code | distFlag, 1.0, fv);
		
		// feature posL posL+1 posR-1 posR
		code = genCodePPPPF(ArcMST.LP_nLP_pRP_RP.ordinal(), LP, nLP, pRP, RP);
		fv = addCode(TArcMST, code, 1.0, fv);
		fv = addCode(TArcMST, code | distFlag, 1.0, fv);
		
		code = genCodePPPF(ArcMST.nLP_pRP_RP.ordinal(), nLP, pRP, RP);
		fv = addCode(TArcMST, code, 1.0, fv);
		fv = addCode(TArcMST, code | distFlag, 1.0, fv);
		
		code = genCodePPPF(ArcMST.LP_pRP_RP.ordinal(), LP, pRP, RP);
		fv = addCode(TArcMST, code, 1.0, fv);
		fv = addCode(TArcMST, code | distFlag, 1.0, fv);
		
		code = genCodePPPF(ArcMST.LP_nLP_RP.ordinal(), LP, nLP, RP);
		fv = addCode(TArcMST, code, 1.0, fv);
		fv = addCode(TArcMST, code | distFlag, 1.0, fv);
		
		code = genCodePPPF(ArcMST.LP_nLP_pRP.ordinal(), LP, nLP, pRP);
		fv = addCode(TArcMST, code, 1.0, fv);
		fv = addCode(TArcMST, code | distFlag, 1.0, fv);
		
		code = genCodePPPPF(ArcMST.LC_nLC_pRC_RC.ordinal(), LC, nLC, pRC, RC);
		fv = addCode(TArcMST, code, 1.0, fv);
		fv = addCode(TArcMST, code | distFlag, 1.0, fv);
		
		code = genCodePPPF(ArcMST.nLC_pRC_RC.ordinal(), nLC, pRC, RC);
		fv = addCode(TArcMST, code, 1.0, fv);
		fv = addCode(TArcMST, code | distFlag, 1.0, fv);
		
		code = genCodePPPF(ArcMST.LC_pRC_RC.ordinal(), LC, pRC, RC);
		fv = addCode(TArcMST, code, 1.0, fv);
		fv = addCode(TArcMST, code | distFlag, 1.0, fv);
		
		code = genCodePPPF(ArcMST.LC_nLC_RC.ordinal(), LC, nLC, RC);
		fv = addCode(TArcMST, code, 1.0, fv);
		fv = addCode(TArcMST, code | distFlag, 1.0, fv);
		
		code = genCodePPPF(ArcMST.LC_nLC_pRC.ordinal(), LC, nLC, pRC);
		fv = addCode(TArcMST, code, 1.0, fv);
		fv = addCode(TArcMST, code | distFlag, 1.0, fv);
		
		// feature posL-1 posL posR-1 posR
		// feature posL posL+1 posR posR+1
		code = genCodePPPPF(ArcMST.pLP_LP_pRP_RP.ordinal(), pLP, LP, pRP, RP);
		fv = addCode(TArcMST, code, 1.0, fv);
		fv = addCode(TArcMST, code | distFlag, 1.0, fv);

		code = genCodePPPPF(ArcMST.pLC_LC_pRC_RC.ordinal(), pLC, LC, pRC, RC);
		fv = addCode(TArcMST, code, 1.0, fv);
		fv = addCode(TArcMST, code | distFlag, 1.0, fv);

		code = genCodePPPPF(ArcMST.LP_nLP_RP_nRP.ordinal(), LP, nLP, RP, nRP);
		fv = addCode(TArcMST, code, 1.0, fv);
		fv = addCode(TArcMST, code | distFlag, 1.0, fv);

		code = genCodePPPPF(ArcMST.LC_nLC_RC_nRC.ordinal(), LC, nLC, RC, nRC);
		fv = addCode(TArcMST, code, 1.0, fv);
		fv = addCode(TArcMST, code | distFlag, 1.0, fv);

		// arc feature
		int HP = pos[headIndex];
		int HW = toks[headIndex];
		int MP = pos[childIndex];
		int MW = toks[childIndex];
		
		code = genCodePF(ArcMST.HP.ordinal(), HP);
		fv = addCode(TArcMST, code, 1.0, fv);
		fv = addCode(TArcMST, code | distFlag, 1.0, fv);
		
		code = genCodeWF(ArcMST.HW.ordinal(), HW);
		fv = addCode(TArcMST, code, 1.0, fv);
		fv = addCode(TArcMST, code | distFlag, 1.0, fv);
		
		//code = genCodePF(ArcMST.MP.ordinal(), MP);
		//fv = addCode(TArcMST, code, 1.0, fv);
		//fv = addCode(TArcMST, code | distFlag, 1.0, fv);
		
		//code = genCodeWF(ArcMST.MW.ordinal(), MW);
		//fv = addCode(TArcMST, code, 1.0, fv);
		//fv = addCode(TArcMST, code | distFlag, 1.0, fv);
		
		code = genCodePPF(ArcMST.HP_MP.ordinal(), HP, MP);
		fv = addCode(TArcMST, code, 1.0, fv);
		fv = addCode(TArcMST, code | distFlag, 1.0, fv);
		
		code = genCodePWF(ArcMST.HP_MW.ordinal(), HP, MW);
		fv = addCode(TArcMST, code, 1.0, fv);
		fv = addCode(TArcMST, code | distFlag, 1.0, fv);
		
		code = genCodePWF(ArcMST.HW_MP.ordinal(), MP, HW);
		fv = addCode(TArcMST, code, 1.0, fv);
		fv = addCode(TArcMST, code | distFlag, 1.0, fv);
		
		code = genCodeWWF(ArcMST.HW_MW.ordinal(), HW, MW);
		fv = addCode(TArcMST, code, 1.0, fv);
		fv = addCode(TArcMST, code | distFlag, 1.0, fv);
		
		code = genCodePWF(ArcMST.HW_HP.ordinal(), HP, HW);
		fv = addCode(TArcMST, code, 1.0, fv);
		fv = addCode(TArcMST, code | distFlag, 1.0, fv);
		
		//code = genCodePWF(ArcMST.MW_MP.ordinal(), pos[childIndex], toks[childIndex]);
		//fv = addCode(TArcMST, code, 1.0, fv);
		//fv = addCode(TArcMST, code | distFlag, 1.0, fv);
		
		code = genCodePPWF(ArcMST.HP_MP_MW.ordinal(), HP, MP, MW);
		fv = addCode(TArcMST, code, 1.0, fv);
		fv = addCode(TArcMST, code | distFlag, 1.0, fv);
		
		code = genCodePPWF(ArcMST.HP_HW_MP.ordinal(), HP, MP, HW);
		fv = addCode(TArcMST, code, 1.0, fv);
		fv = addCode(TArcMST, code | distFlag, 1.0, fv);
		
		code = genCodePWWF(ArcMST.HW_MP_MW.ordinal(), MP, HW, MW);
		fv = addCode(TArcMST, code, 1.0, fv);
		fv = addCode(TArcMST, code | distFlag, 1.0, fv);
		
		code = genCodePWWF(ArcMST.HP_HW_MW.ordinal(), HP, HW, MW);
		fv = addCode(TArcMST, code, 1.0, fv);
		fv = addCode(TArcMST, code | distFlag, 1.0, fv);
		
		code = genCodePPWWF(ArcMST.HP_HW_MP_MW.ordinal(), HP, MP, HW, MW);
		fv = addCode(TArcMST, code, 1.0, fv);
		fv = addCode(TArcMST, code | distFlag, 1.0, fv);

		if (hasLemma) {
			
			int HL = m[headIndex].lemma;
			int ML = m[childIndex].lemma;
			
			code = genCodeWF(ArcMST.HL.ordinal(), HL);
			fv = addCode(TArcMST, code, 1.0, fv);
			fv = addCode(TArcMST, code | distFlag, 1.0, fv);
			
			//code = genCodeWF(ArcMST.ML.ordinal(), ML);
			//fv = addCode(TArcMST, code, 1.0, fv);
			//fv = addCode(TArcMST, code | distFlag, 1.0, fv);
			
			code = genCodePWF(ArcMST.HP_ML.ordinal(), HP, ML);
			fv = addCode(TArcMST, code, 1.0, fv);
			fv = addCode(TArcMST, code | distFlag, 1.0, fv);
			
			code = genCodePWF(ArcMST.HL_MP.ordinal(), MP, HL);
			fv = addCode(TArcMST, code, 1.0, fv);
			fv = addCode(TArcMST, code | distFlag, 1.0, fv);
			
			code = genCodeWWF(ArcMST.HL_ML.ordinal(), HL, ML);
			fv = addCode(TArcMST, code, 1.0, fv);
			fv = addCode(TArcMST, code | distFlag, 1.0, fv);
			
			//code = genCodePWF(ArcMST.MP_ML.ordinal(), MP, ML);
			//fv = addCode(TArcMST, code, 1.0, fv);
			//fv = addCode(TArcMST, code | distFlag, 1.0, fv);
			
			code = genCodePWF(ArcMST.HP_HL.ordinal(), HP, HL);
			fv = addCode(TArcMST, code, 1.0, fv);
			fv = addCode(TArcMST, code | distFlag, 1.0, fv);
			
			code = genCodePPWF(ArcMST.HP_MP_ML.ordinal(), HP, MP, ML);
			fv = addCode(TArcMST, code, 1.0, fv);
			fv = addCode(TArcMST, code | distFlag, 1.0, fv);
			
			code = genCodePPWF(ArcMST.HP_HL_MP.ordinal(), HP, MP, HL);
			fv = addCode(TArcMST, code, 1.0, fv);
			fv = addCode(TArcMST, code | distFlag, 1.0, fv);
			
			code = genCodePWWF(ArcMST.HL_MP_ML.ordinal(), HP, HL, ML);
			fv = addCode(TArcMST, code, 1.0, fv);
			fv = addCode(TArcMST, code | distFlag, 1.0, fv);
			
			code = genCodePWWF(ArcMST.HP_HL_ML.ordinal(), MP, HL, ML);
			fv = addCode(TArcMST, code, 1.0, fv);
			fv = addCode(TArcMST, code | distFlag, 1.0, fv);
			
			code = genCodePPWWF(ArcMST.HP_HL_MP_ML.ordinal(), HP, MP, HL, ML);
			fv = addCode(TArcMST, code, 1.0, fv);
			fv = addCode(TArcMST, code | distFlag, 1.0, fv);
			
		}

		// morphology feature
		if (DependencyParser.useMorphology) {
			for (int fh = 0; fh < m[headIndex].id.length; fh++) {
				if (m[headIndex].id[fh] <= 0)
					continue;
				for (int fc = 0; fc < m[childIndex].id.length; fc++) {
					if (m[childIndex].id[fc] <= 0)
						continue;
					
					int IDH = m[headIndex].id[fh];
					int IDM = m[childIndex].id[fc];
					int HV = m[headIndex].val[fh];
					int MV = m[childIndex].val[fc];
					
					code = genCodeIIVF(ArcMST.FF_IDH_IDM_HV.ordinal(), IDH, IDM, HV);
					fv = addCode(TArcMST, code, 1.0, fv);
					fv = addCode(TArcMST, code | distFlag, 1.0, fv);
					
					code = genCodeIIVF(ArcMST.FF_IDH_IDM_MV.ordinal(), IDH, IDM, MV);
					fv = addCode(TArcMST, code, 1.0, fv);
					fv = addCode(TArcMST, code | distFlag, 1.0, fv);
					
					code = genCodeIIVPF(ArcMST.FF_IDH_IDM_HP_MV.ordinal(), IDH, IDM, MV, HP);
					fv = addCode(TArcMST, code, 1.0, fv);
					fv = addCode(TArcMST, code | distFlag, 1.0, fv);
					
					code = genCodeIIVPF(ArcMST.FF_IDH_IDM_HV_MP.ordinal(), IDH, IDM, HV, MP);
					fv = addCode(TArcMST, code, 1.0, fv);
					fv = addCode(TArcMST, code | distFlag, 1.0, fv);
					
					code = genCodeIIVVF(ArcMST.FF_IDH_IDM_HV_MV.ordinal(), IDH, IDM, HV, MV);
					fv = addCode(TArcMST, code, 1.0, fv);
					fv = addCode(TArcMST, code | distFlag, 1.0, fv);
					
					code = genCodeIIVPF(ArcMST.FF_IDH_IDM_HV_HP.ordinal(), IDH, IDM, HV, HP);
					fv = addCode(TArcMST, code, 1.0, fv);
					fv = addCode(TArcMST, code | distFlag, 1.0, fv);
					
					code = genCodeIIVPF(ArcMST.FF_IDH_IDM_MV_MP.ordinal(), IDH, IDM, MV, MP);
					fv = addCode(TArcMST, code, 1.0, fv);
					fv = addCode(TArcMST, code | distFlag, 1.0, fv);
					
					code = genCodeIIVPPF(ArcMST.FF_IDH_IDM_HP_MP_MV.ordinal(), IDH, IDM, MV, HP, MP);
					fv = addCode(TArcMST, code, 1.0, fv);
					fv = addCode(TArcMST, code | distFlag, 1.0, fv);
					
					code = genCodeIIVPPF(ArcMST.FF_IDH_IDM_HP_HV_MP.ordinal(), IDH, IDM, HV, HP, MP);
					fv = addCode(TArcMST, code, 1.0, fv);
					fv = addCode(TArcMST, code | distFlag, 1.0, fv);
					
					code = genCodeIIVVPF(ArcMST.FF_IDH_IDM_HV_MP_MV.ordinal(), IDH, IDM, HV, MV, MP);
					fv = addCode(TArcMST, code, 1.0, fv);
					fv = addCode(TArcMST, code | distFlag, 1.0, fv);
					
					code = genCodeIIVVPF(ArcMST.FF_IDH_IDM_HP_HV_MV.ordinal(), IDH, IDM, HV, MV, HP);
					fv = addCode(TArcMST, code, 1.0, fv);
					fv = addCode(TArcMST, code | distFlag, 1.0, fv);
					
					code = genCodeIIVVPPF(ArcMST.FF_IDH_IDM_HP_HV_MP_MV.ordinal(), IDH, IDM, HV, MV, HP, MP);
					fv = addCode(TArcMST, code, 1.0, fv);
					fv = addCode(TArcMST, code | distFlag, 1.0, fv);
				}
			}
		}
	
		// parser prediction feature, assume that inst.parserPredDep gives it
		//fv = add("Pred=" + (inst.parserPredDep[childIndex] == headIndex ? "1" : "0"), 1.0, fv);
    	if (DependencyParser.pruningMode != 0 && DependencyParser.addPredAsFeature
    			&& inst.parserPredDep[childIndex] == headIndex) {
    		int HC = posA[headIndex];
    		int MC = posA[childIndex];

    		code = genCodePP(ArcMST.PRED_HC_MC.ordinal(), HC, MC);
    		fv = addCode(TArcMST, code, 1.0, fv);

    		//code = ArcMST.PRED.ordinal();
    		//fv = addCode(TArcMST, code, 1.0, fv);
    	}

		// score in inst.parserPredScore
		if (DependencyParser.pruningMode == 3) {
			code = ArcMST.SCORE1.ordinal();
			fv = addCode(TArcMST, code, inst.parserPredScore.probs[headIndex][childIndex], fv);
		}

		if (DependencyParser.useTurbo) {
			fv = addTurboArc(inst, headIndex, childIndex, distFlag, fv);
		}

		if (DependencyParser.isSpmrl) {
			fv = addCatibArc(inst, headIndex, childIndex, distFlag, fv);
		}
		
		if (DependencyParser.usePOSFeature) {
			fv = addPOSFeature(inst, childIndex, fv);
		}
		
		// contextual
		int pHW = headIndex > 0 ? toks[headIndex - 1] : START;
		pHW = headIndex == childIndex + 1 ? MID : pHW;
		
		int nHW = headIndex < len - 1 ? toks[headIndex + 1] : END;
		nHW = headIndex == childIndex - 1 ? MID : nHW;

		int pMW = childIndex > 0 ? toks[childIndex - 1] : START;
		pMW = childIndex == headIndex + 1 ? MID : pMW;

		int nMW = childIndex < len - 1 ? toks[childIndex + 1] : END;
		nMW = childIndex == headIndex - 1 ? MID : nMW;

		code = genCodeWF(ArcMST.pHW.ordinal(), pHW);
		fv = addCode(TArcMST, code, 1.0, fv);
		fv = addCode(TArcMST, code | distFlag, 1.0, fv);

		code = genCodeWF(ArcMST.nHW.ordinal(), nHW);
		fv = addCode(TArcMST, code, 1.0, fv);
		fv = addCode(TArcMST, code | distFlag, 1.0, fv);

		code = genCodeWF(ArcMST.pMW.ordinal(), pMW);
		fv = addCode(TArcMST, code | distFlag, 1.0, fv);

		code = genCodeWF(ArcMST.MW.ordinal(), MW);
		fv = addCode(TArcMST, code | distFlag, 1.0, fv);

		code = genCodeWF(ArcMST.nMW.ordinal(), nMW);
		fv = addCode(TArcMST, code | distFlag, 1.0, fv);
		
		return fv;

	}

	public FeatureVector addCatibArc(DependencyInstance inst,
			int head, int mod, int distFlag, FeatureVector fv) {

		CatibMorphology[] cm = inst.cm;
		int len = inst.dep.length;

		int HD = cm[head].dtPos;
		int MD = cm[mod].dtPos;

		int pHD = head > 0 ? cm[head - 1].dtPos : START;
		pHD = head == mod + 1 ? MID : pHD;

		int nHD = head < len - 1 ? cm[head + 1].dtPos : END;
		nHD = head == mod - 1 ? MID : nHD;

		int pMD = mod > 0 ? cm[mod - 1].dtPos : START;
		pMD = mod == head + 1 ? MID : pMD;

		int nMD = mod < len - 1 ? cm[mod + 1].dtPos : END;
		nMD = mod == head - 1 ? MID : nMD;

		long code = 0;

		int small = head < mod ? head : mod;
		int large = head > mod ? head : mod;
		// feature posR posMid posL
		for (int i = small + 1; i < large; i++) {
			int BD = cm[i].dtPos;
			code = genCodePPPF(CatibFeature.HD_BD_MD.ordinal(), HD, BD, MD);
			fv = addCode(TCatib, code, 1.0, fv);
			fv = addCode(TCatib, code | distFlag, 1.0, fv);
		}

		code = genCodePPPPF(CatibFeature.pHD_HD_MD_nMD.ordinal(), pHD, HD, MD, nMD);
		fv = addCode(TCatib, code, 1.0, fv);
		fv = addCode(TCatib, code | distFlag, 1.0, fv);

		code = genCodePPPF(CatibFeature.HD_MD_nMD.ordinal(), HD, MD, nMD);
		fv = addCode(TCatib, code, 1.0, fv);
		fv = addCode(TCatib, code | distFlag, 1.0, fv);

		code = genCodePPPF(CatibFeature.pHD_MD_nMD.ordinal(), pHD, MD, nMD);
		fv = addCode(TCatib, code, 1.0, fv);
		fv = addCode(TCatib, code | distFlag, 1.0, fv);

		code = genCodePPPF(CatibFeature.pHD_HD_nMD.ordinal(), pHD, HD, nMD);
		fv = addCode(TCatib, code, 1.0, fv);
		fv = addCode(TCatib, code | distFlag, 1.0, fv);

		code = genCodePPPF(CatibFeature.pHD_HD_MD.ordinal(), pHD, HD, MD);
		fv = addCode(TCatib, code, 1.0, fv);
		fv = addCode(TCatib, code | distFlag, 1.0, fv);

		code = genCodePPPPF(CatibFeature.HD_nHD_pMD_MD.ordinal(), HD, nHD, pMD, MD);
		fv = addCode(TCatib, code, 1.0, fv);
		fv = addCode(TCatib, code | distFlag, 1.0, fv);

		code = genCodePPPF(CatibFeature.nHD_pMD_MD.ordinal(), nHD, pMD, MD);
		fv = addCode(TCatib, code, 1.0, fv);
		fv = addCode(TCatib, code | distFlag, 1.0, fv);

		code = genCodePPPF(CatibFeature.HD_pMD_MD.ordinal(), HD, pMD, MD);
		fv = addCode(TCatib, code, 1.0, fv);
		fv = addCode(TCatib, code | distFlag, 1.0, fv);

		code = genCodePPPF(CatibFeature.HD_nHD_MD.ordinal(), HD, nHD, MD);
		fv = addCode(TCatib, code, 1.0, fv);
		fv = addCode(TCatib, code | distFlag, 1.0, fv);

		code = genCodePPPF(CatibFeature.HD_nHD_pMD.ordinal(), HD, nHD, pMD);
		fv = addCode(TCatib, code, 1.0, fv);
		fv = addCode(TCatib, code | distFlag, 1.0, fv);

		code = genCodePPPPF(CatibFeature.pHD_HD_pMD_MD.ordinal(), pHD, HD, pMD, MD);
		fv = addCode(TCatib, code, 1.0, fv);
		fv = addCode(TCatib, code | distFlag, 1.0, fv);

		code = genCodePPPPF(CatibFeature.HD_nHD_MD_nMD.ordinal(), HD, nHD, MD, nMD);
		fv = addCode(TCatib, code, 1.0, fv);
		fv = addCode(TCatib, code | distFlag, 1.0, fv);

		code = genCodePF(CatibFeature.HD.ordinal(), HD);
		fv = addCode(TCatib, code, 1.0, fv);
		fv = addCode(TCatib, code | distFlag, 1.0, fv);

		code = genCodePPF(CatibFeature.HD_MD.ordinal(), HD, MD);
		fv = addCode(TCatib, code, 1.0, fv);
		fv = addCode(TCatib, code | distFlag, 1.0, fv);

		if (inst.specialPos[head] == SpecialPos.N && inst.specialPos[mod] == SpecialPos.AJ) {
			if (cm[head].fgen > NA && cm[head].fnum > NA && cm[head].rat > NA
					&& cm[mod].fgen > NA && cm[mod].fnum > NA) {
				code = genCodeVVVVV(CatibFeature.NA_HGEN_HNUM_HRAT_MGEN_MNUM.ordinal(),
						cm[head].fgen, cm[head].fnum, cm[head].rat, cm[mod].fgen, cm[mod].fnum);
				fv = addCode(TCatib, code, 1.0, fv);
			}
			fv = adjectiveAgreement(inst, head, mod, fv);
		}
		return fv;
	}
	
	public FeatureVector addPOSFeature(DependencyInstance inst, int mod, FeatureVector fv) {
		
		int[] pos = inst.pos;
		Morphology[] m = inst.m;
		
		int ppP = mod > 1 ? pos[mod - 2] : START;
		int pP = mod > 0 ? pos[mod - 1] : START;
		int P = pos[mod];
		int nP = mod < pos.length - 1 ? pos[mod + 1] : END;
		int nnP = mod < pos.length - 2 ? pos[mod + 2] : END;
		
		int ppL = mod > 1 ? m[mod - 2].lemma : START;
		int pL = mod > 0 ? m[mod - 1].lemma : START;
		int L = m[mod].lemma;
		int nL = mod < m.length - 1 ? m[mod + 1].lemma : END;
		int nnL = mod < m.length - 2 ? m[mod + 2].lemma : END;
		
		long code = 0;
		
		code = genCodePPF(HighOrder.ppP_P.ordinal(), ppP, P);
		fv = addCode(THighOrder, code, 1.0, fv);
		
		code = genCodePPF(HighOrder.pP_P.ordinal(), pP, P);
		fv = addCode(THighOrder, code, 1.0, fv);
		
		code = genCodePPF(HighOrder.P_nP.ordinal(), P, nP);
		fv = addCode(THighOrder, code, 1.0, fv);
		
		code = genCodePPF(HighOrder.P_nnP.ordinal(), P, nnP);
		fv = addCode(THighOrder, code, 1.0, fv);

		code = genCodePPPF(HighOrder.ppP_pP_P.ordinal(), ppP, pP, P);
		fv = addCode(THighOrder, code, 1.0, fv);
		
		code = genCodePPPF(HighOrder.pP_P_nP.ordinal(), pP, P, nP);
		fv = addCode(THighOrder, code, 1.0, fv);
		
		code = genCodePPPF(HighOrder.P_nP_nnP.ordinal(), P, nP, nnP);
		fv = addCode(THighOrder, code, 1.0, fv);
		
		code = genCodePPPF(HighOrder.ppP_P_nP.ordinal(), ppP, P, nP);
		fv = addCode(THighOrder, code, 1.0, fv);

		code = genCodePPPF(HighOrder.pP_P_nnP.ordinal(), pP, P, nnP);
		fv = addCode(THighOrder, code, 1.0, fv);

		code = genCodePPPF(HighOrder.ppP_P_nnP.ordinal(), ppP, P, nnP);
		fv = addCode(THighOrder, code, 1.0, fv);

		code = genCodePPPPF(HighOrder.ppP_pP_P_nP.ordinal(), ppP, pP, P, nP);
		fv = addCode(THighOrder, code, 1.0, fv);
		
		code = genCodePPPPF(HighOrder.ppP_pP_P_nnP.ordinal(), ppP, pP, P, nnP);
		fv = addCode(THighOrder, code, 1.0, fv);

		code = genCodePPPPF(HighOrder.pP_P_nP_nnP.ordinal(), pP, P, nP, nnP);
		fv = addCode(THighOrder, code, 1.0, fv);

		code = genCodePPPPF(HighOrder.ppP_P_nP_nnP.ordinal(), ppP, P, nP, nnP);
		fv = addCode(THighOrder, code, 1.0, fv);

		code = genCodePPPPPF(HighOrder.ppP_pP_P_nP_nnP.ordinal(), ppP, pP, P, nP, nnP);
		fv = addCode(THighOrder, code, 1.0, fv);

		code = genCodePWF(HighOrder.ppL_P.ordinal(), P, ppL);
		fv = addCode(THighOrder, code, 1.0, fv);
		
		code = genCodePWF(HighOrder.pL_P.ordinal(), P, pL);
		fv = addCode(THighOrder, code, 1.0, fv);
		
		code = genCodePWF(HighOrder.L_P.ordinal(), P, L);
		fv = addCode(THighOrder, code, 1.0, fv);
		
		code = genCodePWF(HighOrder.P_nL.ordinal(), P, nL);
		fv = addCode(THighOrder, code, 1.0, fv);

		code = genCodePWF(HighOrder.P_nnL.ordinal(), P, nnL);
		fv = addCode(THighOrder, code, 1.0, fv);

		code = genCodePPWF(HighOrder.pP_L_P.ordinal(), pP, P, L);
		fv = addCode(THighOrder, code, 1.0, fv);

		code = genCodePPWF(HighOrder.L_P_nP.ordinal(), P, nP, L);
		fv = addCode(THighOrder, code, 1.0, fv);

		code = genCodePPPWF(HighOrder.pP_L_P_nP.ordinal(), pP, P, nP, L);
		fv = addCode(THighOrder, code, 1.0, fv);

		code = genCodePPPWF(HighOrder.ppP_pP_L_P.ordinal(), ppP, pP, P, L);
		fv = addCode(THighOrder, code, 1.0, fv);

		code = genCodePPPWF(HighOrder.L_P_nP_nnP.ordinal(), P, nP, nnP, L);
		fv = addCode(THighOrder, code, 1.0, fv);

		return fv;
	}

	public FeatureVector adjectiveAgreement(DependencyInstance s, int h, int m, FeatureVector fv) {
		boolean hM = s.cm[h].fgen == M || s.cm[h].fgen == B;
		boolean hF = s.cm[h].fgen == F || s.cm[h].fgen == B;
		boolean mM = s.cm[m].fgen == M || s.cm[m].fgen == B;
		boolean mF = s.cm[m].fgen == F || s.cm[m].fgen == B;
		boolean genAgree = (hM && mM) || (hF && mF) || s.cm[h].fgen == s.cm[m].fgen;

		boolean hS = s.cm[h].fnum == S || s.cm[h].fnum == B;
		boolean hP = s.cm[h].fnum == P || s.cm[h].fnum == B;
		boolean mS = s.cm[m].fnum == S || s.cm[m].fnum == B;
		boolean mP = s.cm[m].fnum == P || s.cm[m].fnum == B;
		boolean numAgree = (hS && mS) || (hP && mP) || s.cm[h].fnum == s.cm[m].fnum;

		boolean agree = false;
		if (s.cm[h].fnum == P && s.cm[h].rat == I) {
			if (mF && mS)
				agree = true;
			else
				agree =false;
		}
		else {
			if (genAgree && numAgree)
				agree = true;
			else
				agree =false;
		}
		if (agree) {
			long code = CatibFeature.NA_AGREE.ordinal();
			fv = addCode(TCatib, code, 1.0, fv);
		}
		return fv;
	}

	public FeatureVector addTurboArc(DependencyInstance inst,
			int head, int mod, int distFlag, FeatureVector fv) {
		int[] posA = inst.posA;
		Morphology[] m = inst.m;
	
		int pH = head > 0 ? posA[head - 1] : START;
		int nH = head < posA.length - 1 ? posA[head + 1] : END;
		int pM = mod > 0 ? posA[mod - 1] : START;
		int nM = mod < posA.length - 1 ? posA[mod + 1] : END;
		
		int HC = posA[head];
		int MC = posA[mod];
		int HL = m[head].lemma;
		int ML = m[mod].lemma;
		
		long code = 0;
		
		code = genCodePPWF(ArcTurbo.pHC_HL_MC.ordinal(), pH, MC, HL);
		fv = addCode(TArcTurbo, code, 1.0, fv);
		fv = addCode(TArcTurbo, code | distFlag, 1.0, fv);

		code = genCodePPWF(ArcTurbo.HL_nHC_MC.ordinal(), nH, MC, HL);
		fv = addCode(TArcTurbo, code, 1.0, fv);
		fv = addCode(TArcTurbo, code | distFlag, 1.0, fv);

		code = genCodePPWF(ArcTurbo.HL_pMC_MC.ordinal(), pM, MC, HL);
		fv = addCode(TArcTurbo, code, 1.0, fv);
		fv = addCode(TArcTurbo, code | distFlag, 1.0, fv);

		code = genCodePPWF(ArcTurbo.HL_MC_nMC.ordinal(), MC, nM, HL);
		fv = addCode(TArcTurbo, code, 1.0, fv);
		fv = addCode(TArcTurbo, code | distFlag, 1.0, fv);

		code = genCodePPPWF(ArcTurbo.pHC_HL_MC_nMC.ordinal(), pH, MC, nM, HL);
		fv = addCode(TArcTurbo, code, 1.0, fv);
		fv = addCode(TArcTurbo, code | distFlag, 1.0, fv);

		code = genCodePPPWF(ArcTurbo.HL_nHC_MC_nMC.ordinal(), nH, MC, nM, HL);
		fv = addCode(TArcTurbo, code, 1.0, fv);
		fv = addCode(TArcTurbo, code | distFlag, 1.0, fv);

		code = genCodePPPWF(ArcTurbo.pHC_HL_pMC_MC.ordinal(), pH, pM, MC, HL);
		fv = addCode(TArcTurbo, code, 1.0, fv);
		fv = addCode(TArcTurbo, code | distFlag, 1.0, fv);

		code = genCodePPPWF(ArcTurbo.HL_nHC_pMC_MC.ordinal(), nH, pM, MC, HL);
		fv = addCode(TArcTurbo, code, 1.0, fv);
		fv = addCode(TArcTurbo, code | distFlag, 1.0, fv);

		code = genCodePPWF(ArcTurbo.pHC_HC_ML.ordinal(), pH, HC, ML);
		fv = addCode(TArcTurbo, code, 1.0, fv);
		fv = addCode(TArcTurbo, code | distFlag, 1.0, fv);

		code = genCodePPWF(ArcTurbo.HC_nHC_ML.ordinal(), HC, nH, ML);
		fv = addCode(TArcTurbo, code, 1.0, fv);
		fv = addCode(TArcTurbo, code | distFlag, 1.0, fv);

		code = genCodePPWF(ArcTurbo.HC_pMC_ML.ordinal(), HC, pM, ML);
		fv = addCode(TArcTurbo, code, 1.0, fv);
		fv = addCode(TArcTurbo, code | distFlag, 1.0, fv);

		code = genCodePPWF(ArcTurbo.HC_ML_nMC.ordinal(), HC, nM, ML);
		fv = addCode(TArcTurbo, code, 1.0, fv);
		fv = addCode(TArcTurbo, code | distFlag, 1.0, fv);

		code = genCodePPPWF(ArcTurbo.pHC_HC_ML_nMC.ordinal(), pH, HC, nM, ML);
		fv = addCode(TArcTurbo, code, 1.0, fv);
		fv = addCode(TArcTurbo, code | distFlag, 1.0, fv);

		code = genCodePPPWF(ArcTurbo.HC_nHC_ML_nMC.ordinal(), HC, nH, nM, ML);
		fv = addCode(TArcTurbo, code, 1.0, fv);
		fv = addCode(TArcTurbo, code | distFlag, 1.0, fv);

		code = genCodePPPWF(ArcTurbo.pHC_HC_pMC_ML.ordinal(), pH, HC, pM, ML);
		fv = addCode(TArcTurbo, code, 1.0, fv);
		fv = addCode(TArcTurbo, code | distFlag, 1.0, fv);

		code = genCodePPPWF(ArcTurbo.HC_nHC_pMC_ML.ordinal(), HC, nH, pM, ML);
		fv = addCode(TArcTurbo, code, 1.0, fv);
		fv = addCode(TArcTurbo, code | distFlag, 1.0, fv);

		code = genCodePWWF(ArcTurbo.pHC_HL_ML.ordinal(), pH, HL, ML);
		fv = addCode(TArcTurbo, code, 1.0, fv);
		fv = addCode(TArcTurbo, code | distFlag, 1.0, fv);

		code = genCodePWWF(ArcTurbo.HL_nHC_ML.ordinal(), nH, HL, ML);
		fv = addCode(TArcTurbo, code, 1.0, fv);
		fv = addCode(TArcTurbo, code | distFlag, 1.0, fv);

		code = genCodePWWF(ArcTurbo.HL_pMC_ML.ordinal(), pM, HL, ML);
		fv = addCode(TArcTurbo, code, 1.0, fv);
		fv = addCode(TArcTurbo, code | distFlag, 1.0, fv);

		code = genCodePWWF(ArcTurbo.HL_ML_nMC.ordinal(), nM, HL, ML);
		fv = addCode(TArcTurbo, code, 1.0, fv);
		fv = addCode(TArcTurbo, code | distFlag, 1.0, fv);

		code = genCodePPWWF(ArcTurbo.pHC_HL_ML_nMC.ordinal(), pH, nM, HL, ML);
		fv = addCode(TArcTurbo, code, 1.0, fv);
		fv = addCode(TArcTurbo, code | distFlag, 1.0, fv);

		code = genCodePPWWF(ArcTurbo.HL_nHC_ML_nMC.ordinal(), nH, nM, HL, ML);
		fv = addCode(TArcTurbo, code, 1.0, fv);
		fv = addCode(TArcTurbo, code | distFlag, 1.0, fv);

		code = genCodePPWWF(ArcTurbo.pHC_HL_pMC_ML.ordinal(), pH, pM, HL, ML);
		fv = addCode(TArcTurbo, code, 1.0, fv);
		fv = addCode(TArcTurbo, code | distFlag, 1.0, fv);

		code = genCodePPWWF(ArcTurbo.HL_nHC_pMC_ML.ordinal(), nH, pM, HL, ML);
		fv = addCode(TArcTurbo, code, 1.0, fv);
		fv = addCode(TArcTurbo, code | distFlag, 1.0, fv);

		int flagVerb = 0x1;
		int flagCoord = 0x2;
		int flagPunc = 0x3;
		int verbNum = 0;
		int coordNum = 0;
		int puncNum = 0;
		int small = head < mod ? head : mod;
		int large = head > mod ? head : mod;
		for (int i = small + 1; i < large; i++) {
			if (inst.specialPos[i] == SpecialPos.V)
				verbNum++;
			else if (inst.specialPos[i] == SpecialPos.C)
				coordNum++;
			else if (inst.specialPos[i] == SpecialPos.PNX)
				puncNum++;
		}
		flagVerb = (flagVerb << 4) | getBinnedDistance(verbNum);
		flagCoord = (flagCoord << 4) | getBinnedDistance(coordNum);
		flagPunc = (flagPunc << 4) | getBinnedDistance(puncNum);
		
		code = genCodePPPF(ArcTurbo.HC_MC_FLAG.ordinal(), HC, MC, flagVerb);
		fv = addCode(TArcTurbo, code, 1.0, fv);
		fv = addCode(TArcTurbo, code | distFlag, 1.0, fv);
		
		code = genCodePPPF(ArcTurbo.HC_MC_FLAG.ordinal(), HC, MC, flagCoord);
		fv = addCode(TArcTurbo, code, 1.0, fv);
		fv = addCode(TArcTurbo, code | distFlag, 1.0, fv);
		
		code = genCodePPPF(ArcTurbo.HC_MC_FLAG.ordinal(), HC, MC, flagPunc);
		fv = addCode(TArcTurbo, code, 1.0, fv);
		fv = addCode(TArcTurbo, code | distFlag, 1.0, fv);
		
		code = genCodePWWF(ArcTurbo.HL_ML_FLAG.ordinal(), flagVerb, HL, ML);
		fv = addCode(TArcTurbo, code, 1.0, fv);
		fv = addCode(TArcTurbo, code | distFlag, 1.0, fv);
		
		code = genCodePWWF(ArcTurbo.HL_ML_FLAG.ordinal(), flagCoord, HL, ML);
		fv = addCode(TArcTurbo, code, 1.0, fv);
		fv = addCode(TArcTurbo, code | distFlag, 1.0, fv);
		
		code = genCodePWWF(ArcTurbo.HL_ML_FLAG.ordinal(), flagPunc, HL, ML);
		fv = addCode(TArcTurbo, code, 1.0, fv);
		fv = addCode(TArcTurbo, code | distFlag, 1.0, fv);
		
		return fv;
	}

	public FeatureVector createFeatureVector(DependencyInstance inst, int word,
			int type, boolean attR, boolean childFeatures, FeatureVector fv) {

		int[] toks = inst.sentence;
		int[] pos = inst.pos;
		Morphology[] m = inst.m;

		int attFlag = ((((attR ? 0 : 1) << 1) | (childFeatures ? 0 : 1)) << 1) | 1;
		attFlag = attFlag << tempOff;

		int w = toks[word];
		int wP = pos[word];

		int pWP = word > 0 ? pos[word - 1] : START;
		int nWP = word < pos.length - 1 ? pos[word + 1] : END;

		long code = 0;
		
		code = genCodePF(ArcMST.LAB.ordinal(), type);
		fv = addCode(TArcMST, code, 1.0, fv);
		fv = addCode(TArcMST, code | attFlag, 1.0, fv);
		
		code = genCodePPWF(ArcMST.LAB_W_WP.ordinal(), type, wP, w);
		fv = addCode(TArcMST, code, 1.0, fv);
		fv = addCode(TArcMST, code | attFlag, 1.0, fv);
		
		code = genCodePPF(ArcMST.LAB_WP.ordinal(), type, wP);
		fv = addCode(TArcMST, code, 1.0, fv);
		fv = addCode(TArcMST, code | attFlag, 1.0, fv);
		
		code = genCodePPPF(ArcMST.LAB_pWP_WP.ordinal(), type, pWP, wP);
		fv = addCode(TArcMST, code, 1.0, fv);
		fv = addCode(TArcMST, code | attFlag, 1.0, fv);
		
		code = genCodePPPF(ArcMST.LAB_WP_nWP.ordinal(), type, wP, nWP);
		fv = addCode(TArcMST, code, 1.0, fv);
		fv = addCode(TArcMST, code | attFlag, 1.0, fv);
		
		code = genCodePPPPF(ArcMST.LAB_pWP_WP_nWP.ordinal(), type, pWP, wP, nWP);
		fv = addCode(TArcMST, code, 1.0, fv);
		fv = addCode(TArcMST, code | attFlag, 1.0, fv);
		
		code = genCodePWF(ArcMST.LAB_W.ordinal(), type, w);
		fv = addCode(TArcMST, code, 1.0, fv);
		fv = addCode(TArcMST, code | attFlag, 1.0, fv);

		// using lemma
		int l = m[word].lemma;

		code = genCodePPWF(ArcMST.LAB_L_WP.ordinal(), type, wP, l);
		fv = addCode(TArcMST, code, 1.0, fv);
		fv = addCode(TArcMST, code | attFlag, 1.0, fv);
	
		code = genCodePWF(ArcMST.LAB_L.ordinal(), type, l);
		fv = addCode(TArcMST, code, 1.0, fv);
		fv = addCode(TArcMST, code | attFlag, 1.0, fv);
	
		// score in inst.parserPredScore
		if (DependencyParser.pruningMode == 3) {
			code = ArcMST.SCORE2.ordinal();
			fv = addCode(TArcMST, code, 
					inst.parserPredScore.probs_lab[word][type][attR ? 0 : 1][childFeatures ? 0 : 1], fv);
		}

		if (DependencyParser.isSpmrl) {
			int WD = inst.cm[word].dtPos;
			int pWD = word > 0 ? inst.cm[word - 1].dtPos : START;
			int nWD = word < pos.length - 1 ? inst.cm[word + 1].dtPos : END;

			code = genCodePPF(CatibFeature.LAB_WD.ordinal(), type, WD);
			fv = addCode(TCatib, code, 1.0, fv);
			fv = addCode(TCatib, code | attFlag, 1.0, fv);

			code = genCodePPPF(CatibFeature.LAB_pWD_WD.ordinal(), type, pWD, WD);
			fv = addCode(TCatib, code, 1.0, fv);
			fv = addCode(TCatib, code | attFlag, 1.0, fv);

			code = genCodePPPF(CatibFeature.LAB_WD_nWD.ordinal(), type, WD, nWD);
			fv = addCode(TCatib, code, 1.0, fv);
			fv = addCode(TCatib, code | attFlag, 1.0, fv);

			code = genCodePPPPF(CatibFeature.LAB_pWD_WD_nWD.ordinal(), type, pWD, WD, nWD);
			fv = addCode(TCatib, code, 1.0, fv);
			fv = addCode(TCatib, code | attFlag, 1.0, fv);
		}

		return fv;
	}

	public FeatureVector createLabFeatureVector(DependencyInstance s, int head,
			int arg, int type, FeatureVector fv) {

		int HP = s.pos[head];
		int MP = s.pos[arg];
		int HL = s.m[head].lemma;
		int ML = s.m[arg].lemma;
		// parent-lab-child
		long code = 0;
		
		code = genCodePPPF(ArcMST.LAB_HP_MP.ordinal(), type, HP, MP);
		fv = addCode(TArcMST, code, 1.0, fv);

		code = genCodePPWF(ArcMST.LAB_HL_MP.ordinal(), type, MP, HL);
		fv = addCode(TArcMST, code, 1.0, fv);

		code = genCodePPWF(ArcMST.LAB_HP_ML.ordinal(), type, HP, ML);
		fv = addCode(TArcMST, code, 1.0, fv);

		code = genCodePWWF(ArcMST.LAB_HL_ML.ordinal(), type, HL, ML);
		fv = addCode(TArcMST, code, 1.0, fv);

		// parser prediction feature, assume that inst.parserPredDep gives it
    	if (DependencyParser.pruningMode != 0 && DependencyParser.addPredAsFeature
    			&& s.parserPredDep[arg] == head && s.parserPredLab[arg] == type) {
    		//fv = add("PredL=" + s.pos[arg] + "_" + s.pos[head] + "_" + type, 1.0, fv);
    		//fv = add("PredL=", 1.0, fv);
    		int HC = s.posA[head];
    		int MC = s.posA[arg];
			code = genCodePPP(ArcMST.PREDL_LAB_HC_MC.ordinal(), type, HC, MC);
			fv = addCode(TArcMST, code, 1.0, fv);
			//code = ArcMST.PREDL.ordinal();
			//fv = addCode(TArcMST, code, 1.0, fv);
    	}

    	if (DependencyParser.isSpmrl) {
    		int HD = s.cm[head].dtPos;
    		int MD = s.cm[arg].dtPos;

    		code = genCodePPPF(CatibFeature.LAB_HD_MD.ordinal(), type, HD, MD);
    		fv = addCode(TCatib, code, 1.0, fv);

    		if (type == SBJ) {
    			if (s.cm[head].fgen > NA && s.cm[head].fnum > NA && s.cm[arg].rat > NA
    					&& s.cm[arg].fgen > NA && s.cm[arg].fnum > NA) {
    				code = genCodeVVVVV(CatibFeature.SBJ_HGEN_HNUM_MRAT_MGEN_MNUM.ordinal(),
    						s.cm[head].fgen, s.cm[head].fnum,
    						s.cm[arg].rat, s.cm[arg].fgen, s.cm[arg].fnum);
    				fv = addCode(TCatib, code, 1.0, fv);
    			}
    			fv = subjectAgreement(s, head, arg, fv);
    		}
    	}

    	return fv;
	}

	public FeatureVector subjectAgreement(DependencyInstance s, int h, int m, FeatureVector fv) {
		boolean hM = s.cm[h].fgen == M || s.cm[h].fgen == B;
		boolean hF = s.cm[h].fgen == F || s.cm[h].fgen == B;
		boolean mM = s.cm[m].fgen == M || s.cm[m].fgen == B;
		boolean mF = s.cm[m].fgen == F || s.cm[m].fgen == B;
		boolean genAgree = (hM && mM) || (hF && mF) || s.cm[h].fgen == s.cm[m].fgen;

		boolean hS = s.cm[h].fnum == S || s.cm[h].fnum == B;
		boolean hP = s.cm[h].fnum == P || s.cm[h].fnum == B;
		boolean mS = s.cm[m].fnum == S || s.cm[m].fnum == B;
		boolean mP = s.cm[m].fnum == P || s.cm[m].fnum == B;
		boolean numAgree = (hS && mS) || (hP && mP) || s.cm[h].fnum == s.cm[m].fnum;

		if (h < m) {
			// verb-subject
			long code = 0;

			code = genCodeV(CatibFeature.SBJ_VS1.ordinal(), s.cm[h].fnum);
			fv = addCode(TCatib, code, 1.0, fv);

			boolean agree = false;
			if (hS) {
				if (s.cm[m].fnum == P && s.cm[m].rat == I) {
					if (hF)
						agree = true;
					else
						agree = false;
				}
				else {
					if (genAgree)
						agree = true;
					else
						agree = false;
				}
			}
			if (agree) {
				code = CatibFeature.SBJ_VS_AGREE.ordinal();
				fv = addCode(TCatib, code, 1.0, fv);
			}
		}
		else {
			// subject-verb
			boolean agree = false;
			if (s.cm[m].fnum == P && s.cm[m].rat == I) {
				if (hF && hS)
					agree = true;
				else
					agree =false;
			}
			else {
				if (genAgree && numAgree)
					agree = true;
				else
					agree =false;
			}
			if (agree) {
				long code = CatibFeature.SBJ_SV_AGREE.ordinal();
				fv = addCode(TCatib, code, 1.0, fv);
			}
		}
		return fv;
	}

	public FeatureVector createFeatureVectorDep2(DependencyInstance s,
			int head, int arg, int headtype, int argtype, FeatureVector fv) {

		String type = headtype + " " + argtype;
		fv = add("Dep2PP=" + s.pos[head] + " " + s.pos[arg] + " " + type, 1.0,
				fv);
		fv = add("Dep2PW=" + s.pos[head] + " " + s.m[arg].lemma + " " + type,
				1.0, fv);
		fv = add("Dep2WP=" + s.m[head].lemma + " " + s.pos[arg] + " " + type,
				1.0, fv);
		fv = add("Dep2WW=" + s.m[head].lemma + " " + s.m[arg].lemma + " "
				+ type, 1.0, fv);

		return fv;
	}

	public FeatureVector createFeatureVector(DependencyInstance inst, int par,
			int ch1, int ch2, FeatureVector fv) {

		
		int[] pos = inst.pos;
		int[] posA = inst.posA;

		// ch1 is always the closes to par
		int dirFlag = (((par < ch2 ? 0 : 1) << 1) | 1) << tempOff;

		int HP = pos[par];
		int SP = ch1 == par ? START : pos[ch1];
		int MP = pos[ch2];
		int HC = posA[par];
		int SC = ch1 == par ? START : posA[ch1];
		int MC  = posA[ch2];
		
		long code = 0;
		
		code = genCodePPPF(SecondOrder.HP_SP_MP.ordinal(), HP, SP, MP);
		fv = addCode(TSecondOrder, code, 1.0, fv);
		fv = addCode(TSecondOrder, code | dirFlag, 1.0, fv);

		code = genCodePPPF(SecondOrder.HC_SC_MC.ordinal(), HC, SC, MC);
		fv = addCode(TSecondOrder, code, 1.0, fv);
		fv = addCode(TSecondOrder, code | dirFlag, 1.0, fv);

		if (DependencyParser.useTurbo) {
			fv = addTurboSib(inst, par, ch1, ch2, dirFlag, fv);
		}

		// score in inst.parserPredScore
		if (DependencyParser.pruningMode == 3) {
			code = SecondOrder.SCORE3.ordinal();
			fv = addCode(TSecondOrder, code, inst.parserPredScore.probs_trips[par][ch1][ch2], fv);
		}

		return fv;
	}
	
	public FeatureVector addTurboSib(DependencyInstance inst, int par,
			int ch1, int ch2, int dirFlag, FeatureVector fv) {
		int[] posA = inst.posA;
		Morphology[] m = inst.m;
		
		int HC = posA[par];
		int SC = ch1 == par ? START : posA[ch1];
		int MC = posA[ch2];

		int HL = m[par].lemma;
		int SL = ch1 == par ? START : m[ch1].lemma;
		int ML = m[ch2].lemma;
		
		int pHC = par > 0 ? posA[par - 1] : START;
		int nHC = par < posA.length - 1 ? posA[par + 1] : END;
		int pSC = ch1 > 0 ? posA[ch1 - 1] : START;
		int nSC = ch1 < posA.length - 1 ? posA[ch1 + 1] : END;
		int pMC = ch2 > 0 ? posA[ch2 - 1] : START;
		int nMC = ch2 < posA.length - 1 ? posA[ch2 + 1] : END;
		
		long code = 0;
		
		// CCC
		code = genCodePPPPF(SecondOrder.pHC_HC_SC_MC.ordinal(), pHC, HC, SC, MC);
		fv = addCode(TSecondOrder, code, 1.0, fv);
		fv = addCode(TSecondOrder, code | dirFlag, 1.0, fv);
		
		code = genCodePPPPF(SecondOrder.HC_nHC_SC_MC.ordinal(), HC, nHC, SC, MC);
		fv = addCode(TSecondOrder, code, 1.0, fv);
		fv = addCode(TSecondOrder, code | dirFlag, 1.0, fv);
		
		code = genCodePPPPF(SecondOrder.HC_pSC_SC_MC.ordinal(), HC, pSC, SC, MC);
		fv = addCode(TSecondOrder, code, 1.0, fv);
		fv = addCode(TSecondOrder, code | dirFlag, 1.0, fv);
		
		code = genCodePPPPF(SecondOrder.HC_SC_nSC_MC.ordinal(), HC, SC, nSC, MC);
		fv = addCode(TSecondOrder, code, 1.0, fv);
		fv = addCode(TSecondOrder, code | dirFlag, 1.0, fv);
		
		code = genCodePPPPF(SecondOrder.HC_SC_pMC_MC.ordinal(), HC, SC, pMC, MC);
		fv = addCode(TSecondOrder, code, 1.0, fv);
		fv = addCode(TSecondOrder, code | dirFlag, 1.0, fv);
		
		code = genCodePPPPF(SecondOrder.HC_SC_MC_nMC.ordinal(), HC, SC, MC, nMC);
		fv = addCode(TSecondOrder, code, 1.0, fv);
		fv = addCode(TSecondOrder, code | dirFlag, 1.0, fv);
		
		// LCC
		code = genCodePPPWF(SecondOrder.pHC_HL_SC_MC.ordinal(), pHC, SC, MC, HL);
		fv = addCode(TSecondOrder, code, 1.0, fv);
		fv = addCode(TSecondOrder, code | dirFlag, 1.0, fv);
		
		code = genCodePPPWF(SecondOrder.HL_nHC_SC_MC.ordinal(), nHC, SC, MC, HL);
		fv = addCode(TSecondOrder, code, 1.0, fv);
		fv = addCode(TSecondOrder, code | dirFlag, 1.0, fv);
		
		code = genCodePPPWF(SecondOrder.HL_pSC_SC_MC.ordinal(), pSC, SC, MC, HL);
		fv = addCode(TSecondOrder, code, 1.0, fv);
		fv = addCode(TSecondOrder, code | dirFlag, 1.0, fv);
		
		code = genCodePPPWF(SecondOrder.HL_SC_nSC_MC.ordinal(), SC, nSC, MC, HL);
		fv = addCode(TSecondOrder, code, 1.0, fv);
		fv = addCode(TSecondOrder, code | dirFlag, 1.0, fv);
		
		code = genCodePPPWF(SecondOrder.HL_SC_pMC_MC.ordinal(), SC, pMC, MC, HL);
		fv = addCode(TSecondOrder, code, 1.0, fv);
		fv = addCode(TSecondOrder, code | dirFlag, 1.0, fv);
		
		code = genCodePPPWF(SecondOrder.HL_SC_MC_nMC.ordinal(), SC, MC, nMC, HL);
		fv = addCode(TSecondOrder, code, 1.0, fv);
		fv = addCode(TSecondOrder, code | dirFlag, 1.0, fv);
		
		// CLC
		code = genCodePPPWF(SecondOrder.pHC_HC_SL_MC.ordinal(), pHC, HC, MC, SL);
		fv = addCode(TSecondOrder, code, 1.0, fv);
		fv = addCode(TSecondOrder, code | dirFlag, 1.0, fv);
		
		code = genCodePPPWF(SecondOrder.HC_nHC_SL_MC.ordinal(), HC, nHC, MC, SL);
		fv = addCode(TSecondOrder, code, 1.0, fv);
		fv = addCode(TSecondOrder, code | dirFlag, 1.0, fv);
		
		code = genCodePPPWF(SecondOrder.HC_pSC_SL_MC.ordinal(), HC, pSC, MC, SL);
		fv = addCode(TSecondOrder, code, 1.0, fv);
		fv = addCode(TSecondOrder, code | dirFlag, 1.0, fv);
		
		code = genCodePPPWF(SecondOrder.HC_SL_nSC_MC.ordinal(), HC, nSC, MC, SL);
		fv = addCode(TSecondOrder, code, 1.0, fv);
		fv = addCode(TSecondOrder, code | dirFlag, 1.0, fv);
		
		code = genCodePPPWF(SecondOrder.HC_SL_pMC_MC.ordinal(), HC, pMC, MC, SL);
		fv = addCode(TSecondOrder, code, 1.0, fv);
		fv = addCode(TSecondOrder, code | dirFlag, 1.0, fv);
		
		code = genCodePPPWF(SecondOrder.HC_SL_MC_nMC.ordinal(), HC, MC, nMC, SL);
		fv = addCode(TSecondOrder, code, 1.0, fv);
		fv = addCode(TSecondOrder, code | dirFlag, 1.0, fv);
		
		// CCL
		code = genCodePPPWF(SecondOrder.pHC_HC_SC_ML.ordinal(), pHC, HC, SC, ML);
		fv = addCode(TSecondOrder, code, 1.0, fv);
		fv = addCode(TSecondOrder, code | dirFlag, 1.0, fv);
		
		code = genCodePPPWF(SecondOrder.HC_nHC_SC_ML.ordinal(), HC, nHC, SC, ML);
		fv = addCode(TSecondOrder, code, 1.0, fv);
		fv = addCode(TSecondOrder, code | dirFlag, 1.0, fv);
		
		code = genCodePPPWF(SecondOrder.HC_pSC_SC_ML.ordinal(), HC, pSC, SC, ML);
		fv = addCode(TSecondOrder, code, 1.0, fv);
		fv = addCode(TSecondOrder, code | dirFlag, 1.0, fv);
		
		code = genCodePPPWF(SecondOrder.HC_SC_nSC_ML.ordinal(), HC, SC, nSC, ML);
		fv = addCode(TSecondOrder, code, 1.0, fv);
		fv = addCode(TSecondOrder, code | dirFlag, 1.0, fv);
		
		code = genCodePPPWF(SecondOrder.HC_SC_pMC_ML.ordinal(), HC, SC, pMC, ML);
		fv = addCode(TSecondOrder, code, 1.0, fv);
		fv = addCode(TSecondOrder, code | dirFlag, 1.0, fv);
		
		code = genCodePPPWF(SecondOrder.HC_SC_ML_nMC.ordinal(), HC, SC, nMC, ML);
		fv = addCode(TSecondOrder, code, 1.0, fv);
		fv = addCode(TSecondOrder, code | dirFlag, 1.0, fv);

		if (DependencyParser.useLargeTurbo) {
			
			code = genCodePPPPPF(SecondOrder.HC_MC_SC_pHC_pMC.ordinal(), HC, MC, SC, pHC, pMC);
			fv = addCode(TSecondOrder, code, 1.0, fv);
			fv = addCode(TSecondOrder, code | dirFlag, 1.0, fv);

			code = genCodePPPPPF(SecondOrder.HC_MC_SC_pHC_pSC.ordinal(), HC, MC, SC, pHC, pSC);
			fv = addCode(TSecondOrder, code, 1.0, fv);
			fv = addCode(TSecondOrder, code | dirFlag, 1.0, fv);

			code = genCodePPPPPF(SecondOrder.HC_MC_SC_pMC_pSC.ordinal(), HC, MC, SC, pMC, pSC);
			fv = addCode(TSecondOrder, code, 1.0, fv);
			fv = addCode(TSecondOrder, code | dirFlag, 1.0, fv);

			code = genCodePPPPPF(SecondOrder.HC_MC_SC_nHC_nMC.ordinal(), HC, MC, SC, nHC, nMC);
			fv = addCode(TSecondOrder, code, 1.0, fv);
			fv = addCode(TSecondOrder, code | dirFlag, 1.0, fv);

			code = genCodePPPPPF(SecondOrder.HC_MC_SC_nHC_nSC.ordinal(), HC, MC, SC, nHC, nSC);
			fv = addCode(TSecondOrder, code, 1.0, fv);
			fv = addCode(TSecondOrder, code | dirFlag, 1.0, fv);

			code = genCodePPPPPF(SecondOrder.HC_MC_SC_nMC_nSC.ordinal(), HC, MC, SC, nMC, nSC);
			fv = addCode(TSecondOrder, code, 1.0, fv);
			fv = addCode(TSecondOrder, code | dirFlag, 1.0, fv);

			code = genCodePPPPPF(SecondOrder.HC_MC_SC_pHC_nMC.ordinal(), HC, MC, SC, pHC, nMC);
			fv = addCode(TSecondOrder, code, 1.0, fv);
			fv = addCode(TSecondOrder, code | dirFlag, 1.0, fv);

			code = genCodePPPPPF(SecondOrder.HC_MC_SC_pHC_nSC.ordinal(), HC, MC, SC, pHC, nSC);
			fv = addCode(TSecondOrder, code, 1.0, fv);
			fv = addCode(TSecondOrder, code | dirFlag, 1.0, fv);

			code = genCodePPPPPF(SecondOrder.HC_MC_SC_pMC_nSC.ordinal(), HC, MC, SC, pMC, nSC);
			fv = addCode(TSecondOrder, code, 1.0, fv);
			fv = addCode(TSecondOrder, code | dirFlag, 1.0, fv);

			code = genCodePPPPPF(SecondOrder.HC_MC_SC_nHC_pMC.ordinal(), HC, MC, SC, nHC, pMC);
			fv = addCode(TSecondOrder, code, 1.0, fv);
			fv = addCode(TSecondOrder, code | dirFlag, 1.0, fv);

			code = genCodePPPPPF(SecondOrder.HC_MC_SC_nHC_pSC.ordinal(), HC, MC, SC, nHC, pSC);
			fv = addCode(TSecondOrder, code, 1.0, fv);
			fv = addCode(TSecondOrder, code | dirFlag, 1.0, fv);

			code = genCodePPPPPF(SecondOrder.HC_MC_SC_nMC_pSC.ordinal(), HC, MC, SC, nMC, pSC);
			fv = addCode(TSecondOrder, code, 1.0, fv);
			fv = addCode(TSecondOrder, code | dirFlag, 1.0, fv);

		}
		
		
		return fv;
	}
	
	public FeatureVector createFeatureVectorSib(DependencyInstance inst,
			int ch1, int ch2, boolean isST, FeatureVector fv) {
		int[] pos = inst.pos;
		int[] posA = inst.posA;
		int[] toks = inst.sentence;
		Morphology[] m = inst.m;

		// ch1 is always the closes to par
		
		int SP = isST ? START : pos[ch1];
		int MP = pos[ch2];
		int SW = isST ? START : toks[ch1];
		int MW = toks[ch2];
		int SC = isST ? START : posA[ch1];
		int MC = posA[ch2];
		int SL = isST ? START : m[ch1].lemma;
		int ML = m[ch2].lemma;
		
		int flag = getBinnedDistance(ch1 - ch2) << tempOff;
		
		long code = 0;
		
		code = genCodePPF(SecondOrder.SP_MP.ordinal(), SP, MP);
		fv = addCode(TSecondOrder, code, 1.0, fv);
		fv = addCode(TSecondOrder, code | flag, 1.0, fv);

		code = genCodeWWF(SecondOrder.SW_MW.ordinal(), SW, MW);
		fv = addCode(TSecondOrder, code, 1.0, fv);
		fv = addCode(TSecondOrder, code | flag, 1.0, fv);

		code = genCodePWF(SecondOrder.SW_MP.ordinal(), MP, SW);
		fv = addCode(TSecondOrder, code, 1.0, fv);
		fv = addCode(TSecondOrder, code | flag, 1.0, fv);

		code = genCodePWF(SecondOrder.SP_MW.ordinal(), SP, MW);
		fv = addCode(TSecondOrder, code, 1.0, fv);
		fv = addCode(TSecondOrder, code | flag, 1.0, fv);

		code = genCodePPF(SecondOrder.SC_MC.ordinal(), SC, MC);
		fv = addCode(TSecondOrder, code, 1.0, fv);
		fv = addCode(TSecondOrder, code | flag, 1.0, fv);

		code = genCodeWWF(SecondOrder.SL_ML.ordinal(), SL, ML);
		fv = addCode(TSecondOrder, code, 1.0, fv);
		fv = addCode(TSecondOrder, code | flag, 1.0, fv);

		code = genCodePWF(SecondOrder.SL_MC.ordinal(), MC, SL);
		fv = addCode(TSecondOrder, code, 1.0, fv);
		fv = addCode(TSecondOrder, code | flag, 1.0, fv);

		code = genCodePWF(SecondOrder.SC_ML.ordinal(), SC, ML);
		fv = addCode(TSecondOrder, code, 1.0, fv);
		fv = addCode(TSecondOrder, code | flag, 1.0, fv);

		// score in inst.parserPredScore
		if (DependencyParser.pruningMode == 3) {
			code = SecondOrder.SCORE4.ordinal();
			fv = addCode(TSecondOrder, code, inst.parserPredScore.probs_sibs[ch1][ch2][isST ? 0 : 1], fv);
		}

		return fv;
	}

	public FeatureVector createFeatureVectorArbSib(DependencyInstance inst,
			int par, int ch1, int ch2, FeatureVector fv) {
		int[] posA = inst.posA;
		Morphology[] m = inst.m;
		
		int dirFlag = ((ch1 < par ? 0x0 : 0x1) << 1) | (ch2 < par ? 0x0 : 0x1);
		dirFlag = (dirFlag << 1) | 1;
		dirFlag = dirFlag << tempOff;
		
		int HC = posA[par];
		int SC = posA[ch1];
		int MC = posA[ch2];

		int HL = m[par].lemma;
		int SL = m[ch1].lemma;
		int ML = m[ch2].lemma;
		
		int pHC = par > 0 ? posA[par - 1] : START;
		int nHC = par < posA.length - 1 ? posA[par + 1] : END;
		int pSC = ch1 > 0 ? posA[ch1 - 1] : START;
		int nSC = ch1 < posA.length - 1 ? posA[ch1 + 1] : END;
		int pMC = ch2 > 0 ? posA[ch2 - 1] : START;
		int nMC = ch2 < posA.length - 1 ? posA[ch2 + 1] : END;
		
		long code = 0;
		
		code = genCodePPPF(CatibFeature.HC_SC_MC.ordinal(), HC, SC, MC);
		fv = addCode(TCatib, code, 1.0, fv);
		fv = addCode(TCatib, code | dirFlag, 1.0, fv);

		// CCC
		code = genCodePPPPF(CatibFeature.pHC_HC_SC_MC.ordinal(), pHC, HC, SC, MC);
		fv = addCode(TCatib, code, 1.0, fv);
		fv = addCode(TCatib, code | dirFlag, 1.0, fv);
		
		code = genCodePPPPF(CatibFeature.HC_nHC_SC_MC.ordinal(), HC, nHC, SC, MC);
		fv = addCode(TCatib, code, 1.0, fv);
		fv = addCode(TCatib, code | dirFlag, 1.0, fv);
		
		code = genCodePPPPF(CatibFeature.HC_pSC_SC_MC.ordinal(), HC, pSC, SC, MC);
		fv = addCode(TCatib, code, 1.0, fv);
		fv = addCode(TCatib, code | dirFlag, 1.0, fv);
		
		code = genCodePPPPF(CatibFeature.HC_SC_nSC_MC.ordinal(), HC, SC, nSC, MC);
		fv = addCode(TCatib, code, 1.0, fv);
		fv = addCode(TCatib, code | dirFlag, 1.0, fv);
		
		code = genCodePPPPF(CatibFeature.HC_SC_pMC_MC.ordinal(), HC, SC, pMC, MC);
		fv = addCode(TCatib, code, 1.0, fv);
		fv = addCode(TCatib, code | dirFlag, 1.0, fv);
		
		code = genCodePPPPF(CatibFeature.HC_SC_MC_nMC.ordinal(), HC, SC, MC, nMC);
		fv = addCode(TCatib, code, 1.0, fv);
		fv = addCode(TCatib, code | dirFlag, 1.0, fv);
		
		// LCC
		code = genCodePPPWF(CatibFeature.pHC_HL_SC_MC.ordinal(), pHC, SC, MC, HL);
		fv = addCode(TCatib, code, 1.0, fv);
		fv = addCode(TCatib, code | dirFlag, 1.0, fv);
		
		code = genCodePPPWF(CatibFeature.HL_nHC_SC_MC.ordinal(), nHC, SC, MC, HL);
		fv = addCode(TCatib, code, 1.0, fv);
		fv = addCode(TCatib, code | dirFlag, 1.0, fv);
		
		code = genCodePPPWF(CatibFeature.HL_pSC_SC_MC.ordinal(), pSC, SC, MC, HL);
		fv = addCode(TCatib, code, 1.0, fv);
		fv = addCode(TCatib, code | dirFlag, 1.0, fv);
		
		code = genCodePPPWF(CatibFeature.HL_SC_nSC_MC.ordinal(), SC, nSC, MC, HL);
		fv = addCode(TCatib, code, 1.0, fv);
		fv = addCode(TCatib, code | dirFlag, 1.0, fv);
		
		code = genCodePPPWF(CatibFeature.HL_SC_pMC_MC.ordinal(), SC, pMC, MC, HL);
		fv = addCode(TCatib, code, 1.0, fv);
		fv = addCode(TCatib, code | dirFlag, 1.0, fv);
		
		code = genCodePPPWF(CatibFeature.HL_SC_MC_nMC.ordinal(), SC, MC, nMC, HL);
		fv = addCode(TCatib, code, 1.0, fv);
		fv = addCode(TCatib, code | dirFlag, 1.0, fv);
		
		// CLC
		code = genCodePPPWF(CatibFeature.pHC_HC_SL_MC.ordinal(), pHC, HC, MC, SL);
		fv = addCode(TCatib, code, 1.0, fv);
		fv = addCode(TCatib, code | dirFlag, 1.0, fv);
		
		code = genCodePPPWF(CatibFeature.HC_nHC_SL_MC.ordinal(), HC, nHC, MC, SL);
		fv = addCode(TCatib, code, 1.0, fv);
		fv = addCode(TCatib, code | dirFlag, 1.0, fv);
		
		code = genCodePPPWF(CatibFeature.HC_pSC_SL_MC.ordinal(), HC, pSC, MC, SL);
		fv = addCode(TCatib, code, 1.0, fv);
		fv = addCode(TCatib, code | dirFlag, 1.0, fv);
		
		code = genCodePPPWF(CatibFeature.HC_SL_nSC_MC.ordinal(), HC, nSC, MC, SL);
		fv = addCode(TCatib, code, 1.0, fv);
		fv = addCode(TCatib, code | dirFlag, 1.0, fv);
		
		code = genCodePPPWF(CatibFeature.HC_SL_pMC_MC.ordinal(), HC, pMC, MC, SL);
		fv = addCode(TCatib, code, 1.0, fv);
		fv = addCode(TCatib, code | dirFlag, 1.0, fv);
		
		code = genCodePPPWF(CatibFeature.HC_SL_MC_nMC.ordinal(), HC, MC, nMC, SL);
		fv = addCode(TCatib, code, 1.0, fv);
		fv = addCode(TCatib, code | dirFlag, 1.0, fv);
		
		// CCL
		code = genCodePPPWF(CatibFeature.pHC_HC_SC_ML.ordinal(), pHC, HC, SC, ML);
		fv = addCode(TCatib, code, 1.0, fv);
		fv = addCode(TCatib, code | dirFlag, 1.0, fv);
		
		code = genCodePPPWF(CatibFeature.HC_nHC_SC_ML.ordinal(), HC, nHC, SC, ML);
		fv = addCode(TCatib, code, 1.0, fv);
		fv = addCode(TCatib, code | dirFlag, 1.0, fv);
		
		code = genCodePPPWF(CatibFeature.HC_pSC_SC_ML.ordinal(), HC, pSC, SC, ML);
		fv = addCode(TCatib, code, 1.0, fv);
		fv = addCode(TCatib, code | dirFlag, 1.0, fv);
		
		code = genCodePPPWF(CatibFeature.HC_SC_nSC_ML.ordinal(), HC, SC, nSC, ML);
		fv = addCode(TCatib, code, 1.0, fv);
		fv = addCode(TCatib, code | dirFlag, 1.0, fv);
		
		code = genCodePPPWF(CatibFeature.HC_SC_pMC_ML.ordinal(), HC, SC, pMC, ML);
		fv = addCode(TCatib, code, 1.0, fv);
		fv = addCode(TCatib, code | dirFlag, 1.0, fv);
		
		code = genCodePPPWF(CatibFeature.HC_SC_ML_nMC.ordinal(), HC, SC, nMC, ML);
		fv = addCode(TCatib, code, 1.0, fv);
		fv = addCode(TCatib, code | dirFlag, 1.0, fv);

		if (DependencyParser.useLargeTurbo) {
			
			code = genCodePPPPPF(CatibFeature.HC_MC_SC_pHC_pMC.ordinal(), HC, MC, SC, pHC, pMC);
			fv = addCode(TCatib, code, 1.0, fv);
			fv = addCode(TCatib, code | dirFlag, 1.0, fv);

			code = genCodePPPPPF(CatibFeature.HC_MC_SC_pHC_pSC.ordinal(), HC, MC, SC, pHC, pSC);
			fv = addCode(TCatib, code, 1.0, fv);
			fv = addCode(TCatib, code | dirFlag, 1.0, fv);

			code = genCodePPPPPF(CatibFeature.HC_MC_SC_pMC_pSC.ordinal(), HC, MC, SC, pMC, pSC);
			fv = addCode(TCatib, code, 1.0, fv);
			fv = addCode(TCatib, code | dirFlag, 1.0, fv);

			code = genCodePPPPPF(CatibFeature.HC_MC_SC_nHC_nMC.ordinal(), HC, MC, SC, nHC, nMC);
			fv = addCode(TCatib, code, 1.0, fv);
			fv = addCode(TCatib, code | dirFlag, 1.0, fv);

			code = genCodePPPPPF(CatibFeature.HC_MC_SC_nHC_nSC.ordinal(), HC, MC, SC, nHC, nSC);
			fv = addCode(TCatib, code, 1.0, fv);
			fv = addCode(TCatib, code | dirFlag, 1.0, fv);

			code = genCodePPPPPF(CatibFeature.HC_MC_SC_nMC_nSC.ordinal(), HC, MC, SC, nMC, nSC);
			fv = addCode(TCatib, code, 1.0, fv);
			fv = addCode(TCatib, code | dirFlag, 1.0, fv);

			code = genCodePPPPPF(CatibFeature.HC_MC_SC_pHC_nMC.ordinal(), HC, MC, SC, pHC, nMC);
			fv = addCode(TCatib, code, 1.0, fv);
			fv = addCode(TCatib, code | dirFlag, 1.0, fv);

			code = genCodePPPPPF(CatibFeature.HC_MC_SC_pHC_nSC.ordinal(), HC, MC, SC, pHC, nSC);
			fv = addCode(TCatib, code, 1.0, fv);
			fv = addCode(TCatib, code | dirFlag, 1.0, fv);

			code = genCodePPPPPF(CatibFeature.HC_MC_SC_pMC_nSC.ordinal(), HC, MC, SC, pMC, nSC);
			fv = addCode(TCatib, code, 1.0, fv);
			fv = addCode(TCatib, code | dirFlag, 1.0, fv);

			code = genCodePPPPPF(CatibFeature.HC_MC_SC_nHC_pMC.ordinal(), HC, MC, SC, nHC, pMC);
			fv = addCode(TCatib, code, 1.0, fv);
			fv = addCode(TCatib, code | dirFlag, 1.0, fv);

			code = genCodePPPPPF(CatibFeature.HC_MC_SC_nHC_pSC.ordinal(), HC, MC, SC, nHC, pSC);
			fv = addCode(TCatib, code, 1.0, fv);
			fv = addCode(TCatib, code | dirFlag, 1.0, fv);

			code = genCodePPPPPF(CatibFeature.HC_MC_SC_nMC_pSC.ordinal(), HC, MC, SC, nMC, pSC);
			fv = addCode(TCatib, code, 1.0, fv);
			fv = addCode(TCatib, code | dirFlag, 1.0, fv);

		}

		return fv;
	}

	public FeatureVector createFeatureVectorHeadBi(DependencyInstance inst,
			int ch, int par1, int par2, FeatureVector fv) {
		int[] pos = inst.pos;
		int[] posA = inst.posA;
		// exactly 16 combination, so need one more template
		int flag = 0;
		if (par1 == par2)
			flag = 1;
		else if (par1 == ch + 1)
			flag = 2;
		else if (par2 == ch)
			flag = 3;
		
		int dirFlag = flag;
		dirFlag = (dirFlag << 1) | (par1 < ch ? 1 : 0);
		dirFlag = (dirFlag << 1) | (par2 < ch + 1 ? 1 : 0);
		
		flag = flag << tempOff;
		dirFlag = dirFlag << tempOff;
		
		long code = 0;
		
		int H1P = pos[par1];
		int H2P = pos[par2];
		int M1P = pos[ch];
		int M2P = pos[ch + 1];
		int H1C = posA[par1];
		int H2C = posA[par2];
		int M1C = posA[ch];
		int M2C = posA[ch + 1];
		
		code = genCodePPPPF(SecondOrder.H1P_H2P_M1P_M2P.ordinal(), H1P, H2P, M1P, M2P);
		fv = addCode(TSecondOrder, code | flag, 1.0, fv);
		code = genCodePPPPF(SecondOrder.H1P_H2P_M1P_M2P_DIR.ordinal(), H1P, H2P, M1P, M2P);
		fv = addCode(TSecondOrder, code | dirFlag, 1.0, fv);
		
		code = genCodePPPPF(SecondOrder.H1C_H2C_M1C_M2C.ordinal(), H1C, H2C, M1C, M2C);
		fv = addCode(TSecondOrder, code | flag, 1.0, fv);
		code = genCodePPPPF(SecondOrder.H1C_H2C_M1C_M2C_DIR.ordinal(), H1C, H2C, M1C, M2C);
		fv = addCode(TSecondOrder, code | dirFlag, 1.0, fv);
		
		return fv;
	}
	
	public FeatureVector createFeatureVectorNonproj(DependencyInstance inst, int head, int mod, FeatureVector fv) {
		
		int[] posA = inst.posA;
		Morphology[] m = inst.m;

		int distFlag = getBinnedDistance(head - mod) << tempOff;
		//int distFlag = getTurboBinnedDistance(headIndex - childIndex) << tempOff;
		
		// use head->mod, rather than small->large
		int HC = posA[head];
		int MC = posA[mod];
		int HL = m[head].lemma;
		int ML = m[mod].lemma;
		
		int projFlag = getBinnedDistance(inst.isNonproj[mod]);

		long code = 0;
		
		code = genCodePF(HighOrder.NP.ordinal(), projFlag);
		fv = addCode(THighOrder, code, 1.0, fv);
		fv = addCode(THighOrder, code | distFlag, 1.0, fv);
		
		code = genCodePPF(HighOrder.NP_MC.ordinal(), projFlag, MC);
		fv = addCode(THighOrder, code, 1.0, fv);
		fv = addCode(THighOrder, code | distFlag, 1.0, fv);
		
		code = genCodePPF(HighOrder.NP_HC.ordinal(), projFlag, HC);
		fv = addCode(THighOrder, code, 1.0, fv);
		fv = addCode(THighOrder, code | distFlag, 1.0, fv);
		
		code = genCodePWF(HighOrder.NP_HL.ordinal(), projFlag, HL);
		fv = addCode(THighOrder, code, 1.0, fv);
		fv = addCode(THighOrder, code | distFlag, 1.0, fv);
		
		code = genCodePWF(HighOrder.NP_ML.ordinal(), projFlag, ML);
		fv = addCode(THighOrder, code, 1.0, fv);
		fv = addCode(THighOrder, code | distFlag, 1.0, fv);
		
		code = genCodePPPF(HighOrder.NP_HC_MC.ordinal(), projFlag, HC, MC);
		fv = addCode(THighOrder, code, 1.0, fv);
		fv = addCode(THighOrder, code | distFlag, 1.0, fv);
		
		code = genCodePPWF(HighOrder.NP_HL_MC.ordinal(), projFlag, MC, HL);
		fv = addCode(THighOrder, code, 1.0, fv);
		fv = addCode(THighOrder, code | distFlag, 1.0, fv);
		
		code = genCodePPWF(HighOrder.NP_HC_ML.ordinal(), projFlag, HC, ML);
		fv = addCode(THighOrder, code, 1.0, fv);
		fv = addCode(THighOrder, code | distFlag, 1.0, fv);
		
		code = genCodePWWF(HighOrder.NP_HL_ML.ordinal(), projFlag, HL, ML);
		fv = addCode(THighOrder, code, 1.0, fv);
		fv = addCode(THighOrder, code | distFlag, 1.0, fv);
		
		return fv;
	}

	public FeatureVector createFeatureVectorGPC(DependencyInstance inst,
			int gp, int par, int c, FeatureVector fv) {
		int[] pos = inst.pos;
		Morphology[] m = inst.m;
		int[] posA = inst.posA;

		int flag = (((((gp < par ? 0 : 1) << 1) | (par < c ? 0 : 1)) << 1) | 1) << tempOff;
		
		int GP = pos[gp];
		int HP = pos[par];
		int MP = pos[c];
		int GC = posA[gp];
		int HC = posA[par];
		int MC = posA[c];
		int GL = m[gp].lemma;
		int HL = m[par].lemma;
		int ML = m[c].lemma;
		
		long code = 0;
		
		code = genCodePPPF(SecondOrder.GP_HP_MP.ordinal(), GP, HP, MP);
		fv = addCode(TSecondOrder, code, 1.0, fv);
		fv = addCode(TSecondOrder, code | flag, 1.0, fv);

		code = genCodePPPF(SecondOrder.GC_HC_MC.ordinal(), GC, HC, MC);
		fv = addCode(TSecondOrder, code, 1.0, fv);
		fv = addCode(TSecondOrder, code | flag, 1.0, fv);

		code = genCodePPWF(SecondOrder.GL_HC_MC.ordinal(), HC, MC, GL);
		fv = addCode(TSecondOrder, code, 1.0, fv);
		fv = addCode(TSecondOrder, code | flag, 1.0, fv);

		code = genCodePPWF(SecondOrder.GC_HL_MC.ordinal(), GC, MC, HL);
		fv = addCode(TSecondOrder, code, 1.0, fv);
		fv = addCode(TSecondOrder, code | flag, 1.0, fv);

		code = genCodePPWF(SecondOrder.GC_HC_ML.ordinal(), GC, HC, ML);
		fv = addCode(TSecondOrder, code, 1.0, fv);
		fv = addCode(TSecondOrder, code | flag, 1.0, fv);

		if (DependencyParser.useTurbo) {
			fv = addTurboGPC(inst, gp, par, c, flag, fv);
		}

		//addTwoObsFeatures("GPCB", pos[gp], m[gp].lemma, pos[c], m[c].lemma,
		//		dir, false, false, fv);

		return fv;
	}

	public FeatureVector addTurboGPC(DependencyInstance inst, int gp,
			int par, int c, int dirFlag, FeatureVector fv) {
		int[] posA = inst.posA;
		Morphology[] m = inst.m;
		
		int GC = posA[gp];
		int HC = posA[par];
		int MC = posA[c];
		int GL = m[gp].lemma;
		int HL = m[par].lemma;
		int ML = m[c].lemma;
		
		int pGC = gp > 0 ? posA[gp - 1] : START;
		int nGC = gp < posA.length - 1 ? posA[gp + 1] : END;
		int pHC = par > 0 ? posA[par - 1] : START;
		int nHC = par < posA.length - 1 ? posA[par + 1] : END;
		int pMC = c > 0 ? posA[c - 1] : START;
		int nMC = c < posA.length - 1 ? posA[c + 1] : END;
		
		long code = 0;
		
		// CCC
		code = genCodePPPPF(SecondOrder.pGC_GC_HC_MC.ordinal(), pGC, GC, HC, MC);
		fv = addCode(TSecondOrder, code, 1.0, fv);
		fv = addCode(TSecondOrder, code | dirFlag, 1.0, fv);
		
		code = genCodePPPPF(SecondOrder.GC_nGC_HC_MC.ordinal(), GC, nGC, HC, MC);
		fv = addCode(TSecondOrder, code, 1.0, fv);
		fv = addCode(TSecondOrder, code | dirFlag, 1.0, fv);
		
		code = genCodePPPPF(SecondOrder.GC_pHC_HC_MC.ordinal(), GC, pHC, HC, MC);
		fv = addCode(TSecondOrder, code, 1.0, fv);
		fv = addCode(TSecondOrder, code | dirFlag, 1.0, fv);
		
		code = genCodePPPPF(SecondOrder.GC_HC_nHC_MC.ordinal(), GC, HC, nHC, MC);
		fv = addCode(TSecondOrder, code, 1.0, fv);
		fv = addCode(TSecondOrder, code | dirFlag, 1.0, fv);
		
		code = genCodePPPPF(SecondOrder.GC_HC_pMC_MC.ordinal(), GC, HC, pMC, MC);
		fv = addCode(TSecondOrder, code, 1.0, fv);
		fv = addCode(TSecondOrder, code | dirFlag, 1.0, fv);
		
		code = genCodePPPPF(SecondOrder.GC_HC_MC_nMC.ordinal(), GC, HC, MC, nMC);
		fv = addCode(TSecondOrder, code, 1.0, fv);
		fv = addCode(TSecondOrder, code | dirFlag, 1.0, fv);

		// LCC
		code = genCodePPPWF(SecondOrder.pGC_GL_HC_MC.ordinal(), pGC, HC, MC, GL);
		fv = addCode(TSecondOrder, code, 1.0, fv);
		fv = addCode(TSecondOrder, code | dirFlag, 1.0, fv);
		
		code = genCodePPPWF(SecondOrder.GL_nGC_HC_MC.ordinal(), nGC, HC, MC, GL);
		fv = addCode(TSecondOrder, code, 1.0, fv);
		fv = addCode(TSecondOrder, code | dirFlag, 1.0, fv);
		
		code = genCodePPPWF(SecondOrder.GL_pHC_HC_MC.ordinal(), pHC, HC, MC, GL);
		fv = addCode(TSecondOrder, code, 1.0, fv);
		fv = addCode(TSecondOrder, code | dirFlag, 1.0, fv);
		
		code = genCodePPPWF(SecondOrder.GL_HC_nHC_MC.ordinal(), HC, nHC, MC, GL);
		fv = addCode(TSecondOrder, code, 1.0, fv);
		fv = addCode(TSecondOrder, code | dirFlag, 1.0, fv);
		
		code = genCodePPPWF(SecondOrder.GL_HC_pMC_MC.ordinal(), HC, pMC, MC, GL);
		fv = addCode(TSecondOrder, code, 1.0, fv);
		fv = addCode(TSecondOrder, code | dirFlag, 1.0, fv);
		
		code = genCodePPPWF(SecondOrder.GL_HC_MC_nMC.ordinal(), HC, MC, nMC, GL);
		fv = addCode(TSecondOrder, code, 1.0, fv);
		fv = addCode(TSecondOrder, code | dirFlag, 1.0, fv);
		
		// CLC
		code = genCodePPPWF(SecondOrder.pGC_GC_HL_MC.ordinal(), pGC, GC, MC, HL);
		fv = addCode(TSecondOrder, code, 1.0, fv);
		fv = addCode(TSecondOrder, code | dirFlag, 1.0, fv);
		
		code = genCodePPPWF(SecondOrder.GC_nGC_HL_MC.ordinal(), GC, nGC, MC, HL);
		fv = addCode(TSecondOrder, code, 1.0, fv);
		fv = addCode(TSecondOrder, code | dirFlag, 1.0, fv);
		
		code = genCodePPPWF(SecondOrder.GC_pHC_HL_MC.ordinal(), GC, pHC, MC, HL);
		fv = addCode(TSecondOrder, code, 1.0, fv);
		fv = addCode(TSecondOrder, code | dirFlag, 1.0, fv);
		
		code = genCodePPPWF(SecondOrder.GC_HL_nHC_MC.ordinal(), GC, nHC, MC, HL);
		fv = addCode(TSecondOrder, code, 1.0, fv);
		fv = addCode(TSecondOrder, code | dirFlag, 1.0, fv);
		
		code = genCodePPPWF(SecondOrder.GC_HL_pMC_MC.ordinal(), GC, pMC, MC, HL);
		fv = addCode(TSecondOrder, code, 1.0, fv);
		fv = addCode(TSecondOrder, code | dirFlag, 1.0, fv);
		
		code = genCodePPPWF(SecondOrder.GC_HL_MC_nMC.ordinal(), GC, MC, nMC, HL);
		fv = addCode(TSecondOrder, code, 1.0, fv);
		fv = addCode(TSecondOrder, code | dirFlag, 1.0, fv);
		
		// CCL
		code = genCodePPPWF(SecondOrder.pGC_GC_HC_ML.ordinal(), pGC, GC, HC, ML);
		fv = addCode(TSecondOrder, code, 1.0, fv);
		fv = addCode(TSecondOrder, code | dirFlag, 1.0, fv);
		
		code = genCodePPPWF(SecondOrder.GC_nGC_HC_ML.ordinal(), GC, nGC, HC, ML);
		fv = addCode(TSecondOrder, code, 1.0, fv);
		fv = addCode(TSecondOrder, code | dirFlag, 1.0, fv);
		
		code = genCodePPPWF(SecondOrder.GC_pHC_HC_ML.ordinal(), GC, pHC, HC, ML);
		fv = addCode(TSecondOrder, code, 1.0, fv);
		fv = addCode(TSecondOrder, code | dirFlag, 1.0, fv);
		
		code = genCodePPPWF(SecondOrder.GC_HC_nHC_ML.ordinal(), GC, HC, nHC, ML);
		fv = addCode(TSecondOrder, code, 1.0, fv);
		fv = addCode(TSecondOrder, code | dirFlag, 1.0, fv);
		
		code = genCodePPPWF(SecondOrder.GC_HC_pMC_ML.ordinal(), GC, HC, pMC, ML);
		fv = addCode(TSecondOrder, code, 1.0, fv);
		fv = addCode(TSecondOrder, code | dirFlag, 1.0, fv);
		
		code = genCodePPPWF(SecondOrder.GC_HC_ML_nMC.ordinal(), GC, HC, nMC, ML);
		fv = addCode(TSecondOrder, code, 1.0, fv);
		fv = addCode(TSecondOrder, code | dirFlag, 1.0, fv);

		if (DependencyParser.useLargeTurbo) {
			
			code = genCodePPPPPF(SecondOrder.GC_HC_MC_pGC_pHC.ordinal(), GC, HC, MC, pGC, pHC);
			fv = addCode(TSecondOrder, code, 1.0, fv);
			fv = addCode(TSecondOrder, code | dirFlag, 1.0, fv);

			code = genCodePPPPPF(SecondOrder.GC_HC_MC_pGC_pMC.ordinal(), GC, HC, MC , pGC, pMC);
			fv = addCode(TSecondOrder, code, 1.0, fv);
			fv = addCode(TSecondOrder, code | dirFlag, 1.0, fv);

			code = genCodePPPPPF(SecondOrder.GC_HC_MC_pHC_pMC.ordinal(), GC, HC, MC, pHC, pMC);
			fv = addCode(TSecondOrder, code, 1.0, fv);
			fv = addCode(TSecondOrder, code | dirFlag, 1.0, fv);

			code = genCodePPPPPF(SecondOrder.GC_HC_MC_nGC_nHC.ordinal(), GC, HC, MC, nGC, nHC);
			fv = addCode(TSecondOrder, code, 1.0, fv);
			fv = addCode(TSecondOrder, code | dirFlag, 1.0, fv);

			code = genCodePPPPPF(SecondOrder.GC_HC_MC_nGC_nMC.ordinal(), GC, HC, MC, nGC, nMC);
			fv = addCode(TSecondOrder, code, 1.0, fv);
			fv = addCode(TSecondOrder, code | dirFlag, 1.0, fv);

			code = genCodePPPPPF(SecondOrder.GC_HC_MC_nHC_nMC.ordinal(), GC, HC, MC, nHC, nMC);
			fv = addCode(TSecondOrder, code, 1.0, fv);
			fv = addCode(TSecondOrder, code | dirFlag, 1.0, fv);

			code = genCodePPPPPF(SecondOrder.GC_HC_MC_pGC_nHC.ordinal(), GC, HC, MC, pGC, nHC);
			fv = addCode(TSecondOrder, code, 1.0, fv);
			fv = addCode(TSecondOrder, code | dirFlag, 1.0, fv);

			code = genCodePPPPPF(SecondOrder.GC_HC_MC_pGC_nMC.ordinal(), GC, HC, MC, pGC, nMC);
			fv = addCode(TSecondOrder, code, 1.0, fv);
			fv = addCode(TSecondOrder, code | dirFlag, 1.0, fv);

			code = genCodePPPPPF(SecondOrder.GC_HC_MC_pHC_nMC.ordinal(), GC, HC, MC, pHC, nMC);
			fv = addCode(TSecondOrder, code, 1.0, fv);
			fv = addCode(TSecondOrder, code | dirFlag, 1.0, fv);

			code = genCodePPPPPF(SecondOrder.GC_HC_MC_nGC_pHC.ordinal(), GC, HC, MC, nGC, pHC);
			fv = addCode(TSecondOrder, code, 1.0, fv);
			fv = addCode(TSecondOrder, code | dirFlag, 1.0, fv);

			code = genCodePPPPPF(SecondOrder.GC_HC_MC_nGC_pMC.ordinal(), GC, HC, MC, nGC, pMC);
			fv = addCode(TSecondOrder, code, 1.0, fv);
			fv = addCode(TSecondOrder, code | dirFlag, 1.0, fv);

			code = genCodePPPPPF(SecondOrder.GC_HC_MC_nHC_pMC.ordinal(), GC, HC, MC, nHC, pMC);
			fv = addCode(TSecondOrder, code, 1.0, fv);
			fv = addCode(TSecondOrder, code | dirFlag, 1.0, fv);

			// lexical feature/bilex feature
			code = genCodePPPWF(ThirdOrder.GL_HC_MC_pGC.ordinal(), HC, MC, pGC, GL);
			fv = addCode(TThirdOrder, code, 1.0, fv);
			fv = addCode(TThirdOrder, code | dirFlag, 1.0, fv);

			code = genCodePPPWF(ThirdOrder.GL_HC_MC_nGC.ordinal(), HC, MC, nGC, GL);
			fv = addCode(TThirdOrder, code, 1.0, fv);
			fv = addCode(TThirdOrder, code | dirFlag, 1.0, fv);

			code = genCodePPPWF(ThirdOrder.GL_HC_MC_pHC.ordinal(), HC, MC, pHC, GL);
			fv = addCode(TThirdOrder, code, 1.0, fv);
			fv = addCode(TThirdOrder, code | dirFlag, 1.0, fv);

			code = genCodePPPWF(ThirdOrder.GL_HC_MC_nHC.ordinal(), HC, MC, nHC, GL);
			fv = addCode(TThirdOrder, code, 1.0, fv);
			fv = addCode(TThirdOrder, code | dirFlag, 1.0, fv);

			code = genCodePPPWF(ThirdOrder.GL_HC_MC_pMC.ordinal(), HC, MC, pMC, GL);
			fv = addCode(TThirdOrder, code, 1.0, fv);
			fv = addCode(TThirdOrder, code | dirFlag, 1.0, fv);

			code = genCodePPPWF(ThirdOrder.GL_HC_MC_nMC.ordinal(), HC, MC, nMC, GL);
			fv = addCode(TThirdOrder, code, 1.0, fv);
			fv = addCode(TThirdOrder, code | dirFlag, 1.0, fv);

			code = genCodePPPWF(ThirdOrder.GC_HL_MC_pGC.ordinal(), GC, MC, pGC, HL);
			fv = addCode(TThirdOrder, code, 1.0, fv);
			fv = addCode(TThirdOrder, code | dirFlag, 1.0, fv);

			code = genCodePPPWF(ThirdOrder.GC_HL_MC_nGC.ordinal(), GC, MC, nGC, HL);
			fv = addCode(TThirdOrder, code, 1.0, fv);
			fv = addCode(TThirdOrder, code | dirFlag, 1.0, fv);

			code = genCodePPPWF(ThirdOrder.GC_HL_MC_pHC.ordinal(), GC, MC, pHC, HL);
			fv = addCode(TThirdOrder, code, 1.0, fv);
			fv = addCode(TThirdOrder, code | dirFlag, 1.0, fv);

			code = genCodePPPWF(ThirdOrder.GC_HL_MC_nHC.ordinal(), GC, MC, nHC, HL);
			fv = addCode(TThirdOrder, code, 1.0, fv);
			fv = addCode(TThirdOrder, code | dirFlag, 1.0, fv);

			code = genCodePPPWF(ThirdOrder.GC_HL_MC_pMC.ordinal(), GC, MC, pMC, HL);
			fv = addCode(TThirdOrder, code, 1.0, fv);
			fv = addCode(TThirdOrder, code | dirFlag, 1.0, fv);

			code = genCodePPPWF(ThirdOrder.GC_HL_MC_nMC.ordinal(), GC, MC, nMC, HL);
			fv = addCode(TThirdOrder, code, 1.0, fv);
			fv = addCode(TThirdOrder, code | dirFlag, 1.0, fv);

			code = genCodePPPWF(ThirdOrder.GC_HC_ML_pGC.ordinal(), GC, HC, pGC, ML);
			fv = addCode(TThirdOrder, code, 1.0, fv);
			fv = addCode(TThirdOrder, code | dirFlag, 1.0, fv);

			code = genCodePPPWF(ThirdOrder.GC_HC_ML_nGC.ordinal(), GC, HC, nGC, ML);
			fv = addCode(TThirdOrder, code, 1.0, fv);
			fv = addCode(TThirdOrder, code | dirFlag, 1.0, fv);

			code = genCodePPPWF(ThirdOrder.GC_HC_ML_pHC.ordinal(), GC, HC, pHC, ML);
			fv = addCode(TThirdOrder, code, 1.0, fv);
			fv = addCode(TThirdOrder, code | dirFlag, 1.0, fv);

			code = genCodePPPWF(ThirdOrder.GC_HC_ML_nHC.ordinal(), GC, HC, nHC, ML);
			fv = addCode(TThirdOrder, code, 1.0, fv);
			fv = addCode(TThirdOrder, code | dirFlag, 1.0, fv);

			code = genCodePPPWF(ThirdOrder.GC_HC_ML_pMC.ordinal(), GC, HC, pMC, ML);
			fv = addCode(TThirdOrder, code, 1.0, fv);
			fv = addCode(TThirdOrder, code | dirFlag, 1.0, fv);

			code = genCodePPPWF(ThirdOrder.GC_HC_ML_nMC.ordinal(), GC, HC, nMC, ML);
			fv = addCode(TThirdOrder, code, 1.0, fv);
			fv = addCode(TThirdOrder, code | dirFlag, 1.0, fv);

			code = genCodePWWF(ThirdOrder.GL_HL_MC.ordinal(), MC, GL, HL);
			fv = addCode(TThirdOrder, code, 1.0, fv);
			fv = addCode(TThirdOrder, code | dirFlag, 1.0, fv);

			code = genCodePWWF(ThirdOrder.GL_HC_ML.ordinal(), HC, GL, ML);
			fv = addCode(TThirdOrder, code, 1.0, fv);
			fv = addCode(TThirdOrder, code | dirFlag, 1.0, fv);

			code = genCodePWWF(ThirdOrder.GC_HL_ML.ordinal(), GC, HL, ML);
			fv = addCode(TThirdOrder, code, 1.0, fv);
			fv = addCode(TThirdOrder, code | dirFlag, 1.0, fv);

			code = genCodeWWW(ThirdOrder.GL_HL_ML.ordinal(), GL, HL, ML);
			fv = addCode(TThirdOrder, code, 1.0, fv);

			code = genCodePPF(ThirdOrder.GC_HC.ordinal(), GC, HC);
			fv = addCode(TThirdOrder, code, 1.0, fv);
			fv = addCode(TThirdOrder, code | dirFlag, 1.0, fv);

			code = genCodePWF(ThirdOrder.GL_HC.ordinal(), HC, GL);
			fv = addCode(TThirdOrder, code, 1.0, fv);
			fv = addCode(TThirdOrder, code | dirFlag, 1.0, fv);

			code = genCodePWF(ThirdOrder.GC_HL.ordinal(), GC, HL);
			fv = addCode(TThirdOrder, code, 1.0, fv);
			fv = addCode(TThirdOrder, code | dirFlag, 1.0, fv);

			code = genCodeWWF(ThirdOrder.GL_HL.ordinal(), GL, HL);
			fv = addCode(TThirdOrder, code, 1.0, fv);
			fv = addCode(TThirdOrder, code | dirFlag, 1.0, fv);

			code = genCodePPF(ThirdOrder.GC_MC.ordinal(), GC, MC);
			fv = addCode(TThirdOrder, code | dirFlag, 1.0, fv);

			code = genCodePWF(ThirdOrder.GL_MC.ordinal(), MC, GL);
			fv = addCode(TThirdOrder, code | dirFlag, 1.0, fv);

			code = genCodePWF(ThirdOrder.GC_ML.ordinal(), GC, ML);
			fv = addCode(TThirdOrder, code | dirFlag, 1.0, fv);

			code = genCodeWWF(ThirdOrder.GL_ML.ordinal(), GL, ML);
			fv = addCode(TThirdOrder, code | dirFlag, 1.0, fv);

			code = genCodePPF(ThirdOrder.HC_MC.ordinal(), HC, MC);
			fv = addCode(TThirdOrder, code | dirFlag, 1.0, fv);

			code = genCodePWF(ThirdOrder.HL_MC.ordinal(), MC, HL);
			fv = addCode(TThirdOrder, code | dirFlag, 1.0, fv);

			code = genCodePWF(ThirdOrder.HC_ML.ordinal(), HC, ML);
			fv = addCode(TThirdOrder, code | dirFlag, 1.0, fv);

			code = genCodeWWF(ThirdOrder.HL_ML.ordinal(), HL, ML);
			fv = addCode(TThirdOrder, code | dirFlag, 1.0, fv);

		}
		return fv;
	}
	
	public FeatureVector createFeatureVectorGGPC(DependencyInstance inst,
			int ggp, int gp, int par, int c, FeatureVector fv) {
		Morphology[] m = inst.m;
		int[] posA = inst.posA;

		int flag = ggp < gp ? 0 : 1;
		flag = (flag << 1) | (gp < par ? 0 : 1);
		flag = (flag << 1) | (par < c ? 0 : 1);
		flag = ((flag << 1) | 1) << tempOff;
		
		int GGC = posA[ggp];
		int GC = posA[gp];
		int HC = posA[par];
		int MC = posA[c];
		int GGL = m[ggp].lemma;
		int GL = m[gp].lemma;
		int HL = m[par].lemma;
		int ML = m[c].lemma;
		
		long code = 0;
		
		code = genCodePPPPF(ThirdOrder.GGC_GC_HC_MC.ordinal(), GGC, GC, HC, MC);
		fv = addCode(TThirdOrder, code, 1.0, fv);
		fv = addCode(TThirdOrder, code | flag, 1.0, fv);
		
		code = genCodePPPWF(ThirdOrder.GGL_GC_HC_MC.ordinal(), GC, HC, MC, GGL);
		fv = addCode(TThirdOrder, code, 1.0, fv);
		fv = addCode(TThirdOrder, code | flag, 1.0, fv);
		
		code = genCodePPPWF(ThirdOrder.GGC_GL_HC_MC.ordinal(), GGC, HC, MC, GL);
		fv = addCode(TThirdOrder, code, 1.0, fv);
		fv = addCode(TThirdOrder, code | flag, 1.0, fv);
		
		code = genCodePPPWF(ThirdOrder.GGC_GC_HL_MC.ordinal(), GGC, GC, MC, HL);
		fv = addCode(TThirdOrder, code, 1.0, fv);
		fv = addCode(TThirdOrder, code | flag, 1.0, fv);
		
		code = genCodePPPWF(ThirdOrder.GGC_GC_HC_ML.ordinal(), GGC, GC, HC, ML);
		fv = addCode(TThirdOrder, code, 1.0, fv);
		fv = addCode(TThirdOrder, code | flag, 1.0, fv);
		
		code = genCodePPPF(ThirdOrder.GGC_HC_MC.ordinal(), GGC, HC, MC);
		fv = addCode(TThirdOrder, code, 1.0, fv);
		fv = addCode(TThirdOrder, code | flag, 1.0, fv);
		
		code = genCodePPWF(ThirdOrder.GGL_HC_MC.ordinal(), HC, MC, GGL);
		fv = addCode(TThirdOrder, code, 1.0, fv);
		fv = addCode(TThirdOrder, code | flag, 1.0, fv);
		
		code = genCodePPWF(ThirdOrder.GGC_HL_MC.ordinal(), GGC, MC, HL);
		fv = addCode(TThirdOrder, code, 1.0, fv);
		fv = addCode(TThirdOrder, code | flag, 1.0, fv);
		
		code = genCodePPWF(ThirdOrder.GGC_HC_ML.ordinal(), GGC, HC, ML);
		fv = addCode(TThirdOrder, code, 1.0, fv);
		fv = addCode(TThirdOrder, code | flag, 1.0, fv);
		
		code = genCodePPPF(ThirdOrder.GGC_GC_MC.ordinal(), GGC, GC, MC);
		fv = addCode(TThirdOrder, code, 1.0, fv);
		fv = addCode(TThirdOrder, code | flag, 1.0, fv);
		
		code = genCodePPWF(ThirdOrder.GGL_GC_MC.ordinal(), GC, MC, GGL);
		fv = addCode(TThirdOrder, code, 1.0, fv);
		fv = addCode(TThirdOrder, code | flag, 1.0, fv);
		
		code = genCodePPWF(ThirdOrder.GGC_GL_MC.ordinal(), GGC, MC, GL);
		fv = addCode(TThirdOrder, code, 1.0, fv);
		fv = addCode(TThirdOrder, code | flag, 1.0, fv);
		
		code = genCodePPWF(ThirdOrder.GGC_GC_ML.ordinal(), GGC, GC, ML);
		fv = addCode(TThirdOrder, code, 1.0, fv);
		fv = addCode(TThirdOrder, code | flag, 1.0, fv);
		
		code = genCodePPF(ThirdOrder.GGC_MC.ordinal(), GGC, MC);
		fv = addCode(TThirdOrder, code, 1.0, fv);
		fv = addCode(TThirdOrder, code | flag, 1.0, fv);
		
		code = genCodePWF(ThirdOrder.GGL_MC.ordinal(), MC, GGL);
		fv = addCode(TThirdOrder, code, 1.0, fv);
		fv = addCode(TThirdOrder, code | flag, 1.0, fv);
		
		code = genCodePWF(ThirdOrder.GGC_ML.ordinal(), GGC, ML);
		fv = addCode(TThirdOrder, code, 1.0, fv);
		fv = addCode(TThirdOrder, code | flag, 1.0, fv);
		
		code = genCodeWWF(ThirdOrder.GGL_ML.ordinal(), GGL, ML);
		fv = addCode(TThirdOrder, code, 1.0, fv);
		fv = addCode(TThirdOrder, code | flag, 1.0, fv);

		return fv;
	}

	public FeatureVector createFeatureVectorPSC(DependencyInstance inst,
			int par, int mod, int ch, int sib, FeatureVector fv) {
		
		int[] posA = inst.posA;
		Morphology[] m = inst.m;
		//int type = (mod - par) * (mod - sib) > 0 ? 0x1 : 0x2;
		int type = (mod - par) * (mod - sib) > 0 ? 0x0 : 0x1;
		int dir = ((par < mod ? 0x0 : 0x1) << 1) | (mod < ch ? 0x0 : 0x1);
		
		int typeFlag = type;
		//int dirFlag = (2 + (dir + 1) + 4 * (type - 1));
		int dirFlag = ((dir + 1) << 1) | type;
		
		typeFlag = typeFlag << tempOff;
		dirFlag = dirFlag << tempOff;
		
		int HC = posA[par];
		int MC = posA[mod];
		int CC = posA[ch];
		int SC = posA[sib];
		int HL = m[par].lemma;
		int ML = m[mod].lemma;
		int CL = m[ch].lemma;
		int SL = m[sib].lemma;
		
		long code = 0;
		
		code = genCodePPPPF(ThirdOrder.HC_MC_CC_SC.ordinal(), HC, MC, CC, SC);
		fv = addCode(TThirdOrder, code | typeFlag, 1.0, fv);
		fv = addCode(TThirdOrder, code | dirFlag, 1.0, fv);
		
		code = genCodePPPWF(ThirdOrder.HL_MC_CC_SC.ordinal(), MC, CC, SC, HL);
		fv = addCode(TThirdOrder, code | typeFlag, 1.0, fv);
		fv = addCode(TThirdOrder, code | dirFlag, 1.0, fv);
		
		code = genCodePPPWF(ThirdOrder.HC_ML_CC_SC.ordinal(), HC, CC, SC, ML);
		fv = addCode(TThirdOrder, code | typeFlag, 1.0, fv);
		fv = addCode(TThirdOrder, code | dirFlag, 1.0, fv);
		
		code = genCodePPPWF(ThirdOrder.HC_MC_CL_SC.ordinal(), HC, MC, SC, CL);
		fv = addCode(TThirdOrder, code | typeFlag, 1.0, fv);
		fv = addCode(TThirdOrder, code | dirFlag, 1.0, fv);
		
		code = genCodePPPWF(ThirdOrder.HC_MC_CC_SL.ordinal(), HC, MC, CC, SL);
		fv = addCode(TThirdOrder, code | typeFlag, 1.0, fv);
		fv = addCode(TThirdOrder, code | dirFlag, 1.0, fv);
		
		code = genCodePPPF(ThirdOrder.HC_CC_SC.ordinal(), HC, CC, SC);
		fv = addCode(TThirdOrder, code | typeFlag, 1.0, fv);
		fv = addCode(TThirdOrder, code | dirFlag, 1.0, fv);
		
		code = genCodePPWF(ThirdOrder.HL_CC_SC.ordinal(), CC, SC, HL);
		fv = addCode(TThirdOrder, code | typeFlag, 1.0, fv);
		fv = addCode(TThirdOrder, code | dirFlag, 1.0, fv);
		
		code = genCodePPWF(ThirdOrder.HC_CL_SC.ordinal(), HC, SC, CL);
		fv = addCode(TThirdOrder, code | typeFlag, 1.0, fv);
		fv = addCode(TThirdOrder, code | dirFlag, 1.0, fv);
		
		code = genCodePPWF(ThirdOrder.HC_CC_SL.ordinal(), HC, CC, SL);
		fv = addCode(TThirdOrder, code | typeFlag, 1.0, fv);
		fv = addCode(TThirdOrder, code | dirFlag, 1.0, fv);
		
		return fv;
	}

	public FeatureVector createFeatureVectorPP(DependencyInstance inst, int gp,
			int par, int c, FeatureVector fv) {
		
		
		int[] posA = inst.posA;
		Morphology[] m = inst.m;

		int HC = posA[gp];
		int MC = posA[c];
		int HL = m[gp].lemma;
		int ML = m[c].lemma;
		int PL = m[par].lemma;
		
		long code = 0;
		
		code = genCodePPF(HighOrder.PP_HC_MC.ordinal(), HC, MC);
		fv = addCode(THighOrder, code, 1.0, fv);
		
		code = genCodePWF(HighOrder.PP_HL_MC.ordinal(), MC, HL);
		fv = addCode(THighOrder, code, 1.0, fv);
		
		code = genCodePWF(HighOrder.PP_HC_ML.ordinal(), HC, ML);
		fv = addCode(THighOrder, code, 1.0, fv);
		
		code = genCodeWWF(HighOrder.PP_HL_ML.ordinal(), HL, ML);
		fv = addCode(THighOrder, code, 1.0, fv);
		
		code = genCodePPWF(HighOrder.PP_PL_HC_MC.ordinal(), HC, MC, PL);
		fv = addCode(THighOrder, code, 1.0, fv);
		
		code = genCodePWWF(HighOrder.PP_PL_HL_MC.ordinal(), MC, HL, PL);
		fv = addCode(THighOrder, code, 1.0, fv);
		
		code = genCodePWWF(HighOrder.PP_PL_HC_ML.ordinal(), HC, ML, PL);
		fv = addCode(THighOrder, code, 1.0, fv);
		
		code = genCodeWWW(HighOrder.PP_PL_HL_ML.ordinal(), HL, ML, PL);
		fv = addCode(THighOrder, code, 1.0, fv);
		
		return fv;
	}

	public FeatureVector createFeatureVectorCC1(DependencyInstance inst,
			int left, int arg, int right, FeatureVector fv) {
		
		
		int[] pos = inst.pos;
		int[] word = inst.sentence;
		int[] posA = inst.posA;
		Morphology[] m = inst.m;

		int CP = pos[arg];
		int CW = word[arg];
		int LP = pos[left];
		int RP = pos[right];
		int LC = posA[left];
		int RC = posA[right];
		
		long code = 0;
		
		code = genCodePPPF(HighOrder.CC_CP_LP_RP.ordinal(), CP, LP, RP);
		fv = addCode(THighOrder, code, 1.0, fv);
		
		code = genCodePPPF(HighOrder.CC_CP_LC_RC.ordinal(), CP, LC, RC);
		fv = addCode(THighOrder, code, 1.0, fv);
		
		code = genCodePPWF(HighOrder.CC_CW_LP_RP.ordinal(), LP, RP, CW);
		fv = addCode(THighOrder, code, 1.0, fv);
		
		code = genCodePPWF(HighOrder.CC_CW_LC_RC.ordinal(), LC, RC, CW);
		fv = addCode(THighOrder, code, 1.0, fv);
		
		//if (posA[left] == posA[right]) {
			for (int i = 0; i < m[left].id.length; i++) {
				if (m[left].id[i] <= 0)
					continue;
				for (int j = 0; j < m[right].id.length; j++) {
					if (m[right].id[j] <= 0)
						continue;
					if (m[left].id[i] == m[right].id[j]) {
						int flag = m[left].val[i] == m[right].val[j] ? 1 : 2;
						flag = flag << tempOff;
						code = genCodeIPPF(HighOrder.CC_LC_RC_FID.ordinal(), m[left].id[i], LC, RC);
						fv = addCode(THighOrder, code | flag, 1.0, fv);
						break;
					}
				}
			}
		//}

		return fv;
	}

	public FeatureVector createFeatureVectorCC2(DependencyInstance inst,
			int arg, int head, int child, FeatureVector fv) {
		
		
		int[] pos = inst.pos;
		int[] word = inst.sentence;
		int[] posA = inst.posA;
		Morphology[] m = inst.m;
		
		int CP = pos[arg];
		int CW = word[arg];
		int HC = posA[head];
		int HL = m[head].lemma;
		int AC = posA[child];
		int AL = m[child].lemma;
		
		long code = 0;
		
		code = genCodePPPF(HighOrder.CC_CP_HC_AC.ordinal(), CP, HC, AC);
		fv = addCode(THighOrder, code, 1.0, fv);
		
		code = genCodePWWF(HighOrder.CC_CP_HL_AL.ordinal(), CP, HL, AL);
		fv = addCode(THighOrder, code, 1.0, fv);
		
		code = genCodePPWF(HighOrder.CC_CW_HC_AC.ordinal(), HC, AC, CW);
		fv = addCode(THighOrder, code, 1.0, fv);
		
		code = genCodeWWW(HighOrder.CC_CW_HL_AL.ordinal(), CW, HL, AL);
		fv = addCode(THighOrder, code, 1.0, fv);

		code = genCodePPF(ArcMST.HP_MP.ordinal(), pos[head], pos[child]);
		fv = addCode(TArcMST, code, 1.0, fv);
		
		code = genCodePWF(ArcMST.HL_MP.ordinal(), pos[child], m[head].lemma);
		fv = addCode(TArcMST, code, 1.0, fv);
		
		code = genCodePWF(ArcMST.HP_ML.ordinal(), pos[head], m[child].lemma);
		fv = addCode(TArcMST, code, 1.0, fv);
		
		code = genCodeWWF(ArcMST.HL_ML.ordinal(), m[head].lemma, m[child].lemma);
		fv = addCode(TArcMST, code, 1.0, fv);

		if (DependencyParser.useMorphology) {
			for (int fh = 0; fh < m[head].id.length; fh++) {
				if (m[head].id[fh] <= 0)
					continue;
				for (int fc = 0; fc < m[child].id.length; fc++) {
					if (m[child].id[fc] <= 0)
						continue;
					
					int IDH = m[head].id[fh];
					int IDM = m[child].id[fc];
					int HV = m[head].val[fh];
					int MV = m[child].val[fc];
					int HP = pos[head];
					int MP = pos[child];

					code = genCodeIIVF(ArcMST.FF_IDH_IDM_HV.ordinal(), IDH, IDM, HV);
					fv = addCode(TArcMST, code, 1.0, fv);
					
					code = genCodeIIVF(ArcMST.FF_IDH_IDM_MV.ordinal(), IDH, IDM, MV);
					fv = addCode(TArcMST, code, 1.0, fv);
					
					code = genCodeIIVPF(ArcMST.FF_IDH_IDM_HP_MV.ordinal(), IDH, IDM, MV, HP);
					fv = addCode(TArcMST, code, 1.0, fv);
					
					code = genCodeIIVPF(ArcMST.FF_IDH_IDM_HV_MP.ordinal(), IDH, IDM, HV, MP);
					fv = addCode(TArcMST, code, 1.0, fv);
					
					code = genCodeIIVVF(ArcMST.FF_IDH_IDM_HV_MV.ordinal(), IDH, IDM, HV, MV);
					fv = addCode(TArcMST, code, 1.0, fv);
					
					code = genCodeIIVPF(ArcMST.FF_IDH_IDM_HV_HP.ordinal(), IDH, IDM, HV, HP);
					fv = addCode(TArcMST, code, 1.0, fv);
					
					code = genCodeIIVPF(ArcMST.FF_IDH_IDM_MV_MP.ordinal(), IDH, IDM, MV, MP);
					fv = addCode(TArcMST, code, 1.0, fv);
					
					code = genCodeIIVPPF(ArcMST.FF_IDH_IDM_HP_MP_MV.ordinal(), IDH, IDM, MV, HP, MP);
					fv = addCode(TArcMST, code, 1.0, fv);
					
					code = genCodeIIVPPF(ArcMST.FF_IDH_IDM_HP_HV_MP.ordinal(), IDH, IDM, HV, HP, MP);
					fv = addCode(TArcMST, code, 1.0, fv);
					
					code = genCodeIIVVPF(ArcMST.FF_IDH_IDM_HV_MP_MV.ordinal(), IDH, IDM, HV, MV, MP);
					fv = addCode(TArcMST, code, 1.0, fv);
					
					code = genCodeIIVVPF(ArcMST.FF_IDH_IDM_HP_HV_MV.ordinal(), IDH, IDM, HV, MV, HP);
					fv = addCode(TArcMST, code, 1.0, fv);
					
					code = genCodeIIVVPPF(ArcMST.FF_IDH_IDM_HP_HV_MP_MV.ordinal(), IDH, IDM, HV, MV, HP, MP);
					fv = addCode(TArcMST, code, 1.0, fv);
				}
			}
		}

		return fv;
	}

	public FeatureVector createFeatureVectorPNX(DependencyInstance inst,
			int head, int arg, int pair, FeatureVector fv) {
		
		int[] pos = inst.pos;
		int[] word = inst.sentence;
		int flag = (head - arg) * (head - pair) < 0 ? 0 : 1;
		flag = (flag + 1) << tempOff;
		
		long code = 0;
		
		code = genCodeWF(HighOrder.PNX_MW.ordinal(), word[arg]);
		fv = addCode(THighOrder, code | flag, 1.0, fv);
		
		code = genCodePWF(HighOrder.PNX_HP_MW.ordinal(), pos[head], word[arg]);
		fv = addCode(THighOrder, code | flag, 1.0, fv);
		
		return fv;
	}

	public FeatureVector createFeatureVectorGPSib(DependencyInstance inst,
			int par, int arg, int prev, int curr, FeatureVector fv) {
		Morphology[] m = inst.m;
		int[] posA = inst.posA;

		int flag = par < arg ? 0 : 1;
		flag = (flag << 1) | (arg < curr ? 0 : 1);
		flag = (flag + 1) << tempOff;

		int GC = posA[par];
		int HC = posA[arg];
		int SC = prev == arg ? START : posA[prev];
		int MC = posA[curr];
		int GL = m[par].lemma;
		int HL = m[arg].lemma;
		int SL = prev == arg ? START : m[prev].lemma;
		int ML = m[curr].lemma;
		
		long code = 0;
		
		code = genCodePPPPF(ThirdOrder.GC_HC_MC_SC.ordinal(), GC, HC, SC, MC);
		fv = addCode(TThirdOrder, code, 1.0, fv);
		fv = addCode(TThirdOrder, code | flag, 1.0, fv);
		
		code = genCodePPPWF(ThirdOrder.GL_HC_MC_SC.ordinal(), HC, SC, MC, GL);
		fv = addCode(TThirdOrder, code, 1.0, fv);
		fv = addCode(TThirdOrder, code | flag, 1.0, fv);
		
		code = genCodePPPWF(ThirdOrder.GC_HL_MC_SC.ordinal(), GC, SC, MC, HL);
		fv = addCode(TThirdOrder, code, 1.0, fv);
		fv = addCode(TThirdOrder, code | flag, 1.0, fv);
		
		code = genCodePPPWF(ThirdOrder.GC_HC_ML_SC.ordinal(), GC, HC, SC, ML);
		fv = addCode(TThirdOrder, code, 1.0, fv);
		fv = addCode(TThirdOrder, code | flag, 1.0, fv);

		code = genCodePPPWF(ThirdOrder.GC_HC_MC_SL.ordinal(), GC, HC, MC, SL);
		fv = addCode(TThirdOrder, code, 1.0, fv);
		fv = addCode(TThirdOrder, code | flag, 1.0, fv);
		
		return fv;
	}

	public FeatureVector createFeatureVectorTriSib(DependencyInstance inst,
			int arg, int prev, int curr, int next, FeatureVector fv) {
		Morphology[] m = inst.m;
		int[] posA = inst.posA;

		int flag = arg < curr ? 0 : 1;
		flag = (flag + 1) << tempOff;

		int HC = posA[arg];
		int PC = prev == arg ? START : posA[prev];
		int MC = posA[curr];
		int NC = posA[next];
		int HL = m[arg].lemma;
		int PL = prev == arg ? START : m[prev].lemma;
		int ML = m[curr].lemma;
		int NL = m[next].lemma;
		
		long code = 0;
		
		code = genCodePPPPF(ThirdOrder.HC_PC_MC_NC.ordinal(), HC, PC, MC, NC);
		fv = addCode(TThirdOrder, code, 1.0, fv);
		fv = addCode(TThirdOrder, code | flag, 1.0, fv);
		
		code = genCodePPPWF(ThirdOrder.HL_PC_MC_NC.ordinal(), PC, MC, NC, HL);
		fv = addCode(TThirdOrder, code, 1.0, fv);
		fv = addCode(TThirdOrder, code | flag, 1.0, fv);
		
		code = genCodePPPWF(ThirdOrder.HC_PL_MC_NC.ordinal(), HC, MC, NC, PL);
		fv = addCode(TThirdOrder, code, 1.0, fv);
		fv = addCode(TThirdOrder, code | flag, 1.0, fv);
		
		code = genCodePPPWF(ThirdOrder.HC_PC_ML_NC.ordinal(), HC, PC, NC, ML);
		fv = addCode(TThirdOrder, code, 1.0, fv);
		fv = addCode(TThirdOrder, code | flag, 1.0, fv);

		code = genCodePPPWF(ThirdOrder.HC_PC_MC_NL.ordinal(), HC, PC, MC, NL);
		fv = addCode(TThirdOrder, code, 1.0, fv);
		fv = addCode(TThirdOrder, code | flag, 1.0, fv);

		code = genCodePPPF(ThirdOrder.HC_PC_NC.ordinal(), HC, PC, NC);
		fv = addCode(TThirdOrder, code, 1.0, fv);
		fv = addCode(TThirdOrder, code | flag, 1.0, fv);
		
		code = genCodePPPF(ThirdOrder.PC_MC_NC.ordinal(), PC, MC, NC);
		fv = addCode(TThirdOrder, code, 1.0, fv);
		fv = addCode(TThirdOrder, code | flag, 1.0, fv);
		
		code = genCodePPWF(ThirdOrder.HL_PC_NC.ordinal(), PC, NC, HL);
		fv = addCode(TThirdOrder, code, 1.0, fv);
		fv = addCode(TThirdOrder, code | flag, 1.0, fv);
		
		code = genCodePPWF(ThirdOrder.HC_PL_NC.ordinal(), HC, NC, PL);
		fv = addCode(TThirdOrder, code, 1.0, fv);
		fv = addCode(TThirdOrder, code | flag, 1.0, fv);
		
		code = genCodePPWF(ThirdOrder.HC_PC_NL.ordinal(), HC, PC, NL);
		fv = addCode(TThirdOrder, code, 1.0, fv);
		fv = addCode(TThirdOrder, code | flag, 1.0, fv);
		
		code = genCodePPWF(ThirdOrder.PL_MC_NC.ordinal(), MC, NC, PL);
		fv = addCode(TThirdOrder, code, 1.0, fv);
		fv = addCode(TThirdOrder, code | flag, 1.0, fv);
		
		code = genCodePPWF(ThirdOrder.PC_ML_NC.ordinal(), PC, NC, ML);
		fv = addCode(TThirdOrder, code, 1.0, fv);
		fv = addCode(TThirdOrder, code | flag, 1.0, fv);
		
		code = genCodePPWF(ThirdOrder.PC_MC_NL.ordinal(), PC, MC, NL);
		fv = addCode(TThirdOrder, code, 1.0, fv);
		fv = addCode(TThirdOrder, code | flag, 1.0, fv);
		
		code = genCodePPF(ThirdOrder.PC_NC.ordinal(), PC, NC);
		fv = addCode(TThirdOrder, code, 1.0, fv);
		fv = addCode(TThirdOrder, code | flag, 1.0, fv);
		
		code = genCodePWF(ThirdOrder.PL_NC.ordinal(), NC, PL);
		fv = addCode(TThirdOrder, code, 1.0, fv);
		fv = addCode(TThirdOrder, code | flag, 1.0, fv);
		
		code = genCodePWF(ThirdOrder.PC_NL.ordinal(), PC, NL);
		fv = addCode(TThirdOrder, code, 1.0, fv);
		fv = addCode(TThirdOrder, code | flag, 1.0, fv);
		
		return fv;
	}

	public FeatureVector createGlobalFeatureVector(DependencyInstance inst,
			FeatureVector fv) {
		if (!DependencyParser.useGlobal)
			return fv;
		
		// have to concate string efficiently, so use StringBuilder

		int[] toks = inst.sentence;
		int[] deps = inst.dep;
		int[] pos = inst.pos;
		SpecialPos[] specialPos = inst.specialPos;
		int[][] child = inst.child;
		int[] en = inst.en;
		int[] st = inst.st;
		
		long code = 0;

		int rb = getMSTRightBranch(inst, 0);
		//fv = add("RB=", (double) rb / toks.length, fv);
		code = HighOrder.RB.ordinal();
		fv = addCode(THighOrder, code, (double)rb / toks.length, fv);
		
		if (Math.abs(inst.parserRerankScore) > 1e-4) {
			fv = add("Score", inst.parserRerankScore, fv);
		}

		for (int i = 1; i < toks.length; i++) {
			if (deps[i] == -1)
				continue;

			// parent-arg-child-(child) in local

			// child num & parent-child
			int leftNum = 0;
			int rightNum = 0;
			StringBuilder childStr = new StringBuilder();
			for (int j = 0; j < child[i].length; j++) {
				if (specialPos[child[i][j]] != SpecialPos.PNX) {
					if (child[i][j] < i
							&& leftNum < GlobalFeatureData.maxChildNum)
						leftNum++;
					else if (child[i][j] > i
							&& rightNum < GlobalFeatureData.maxChildNum)
						rightNum++;
					childStr.append(" ").append(pos[child[i][j]]);
				}
			}
			fv = add("Gl0=" + pos[i] + " " + childStr.toString(), 1.0, fv);

			fv = createChildNumFeatureVector(inst, i, leftNum, rightNum, fv);

			// heavy
			int end = en[i] == toks.length ? 1 : 0;
			int punc = (en[i] < toks.length && specialPos[en[i]] == SpecialPos.PNX) ? 1
					: 0;
			int bin = Math.min(GlobalFeatureData.maxBinLength, (en[i] - st[i]));
			fv = createHeavyFeatureVector(inst, i, end, punc, bin, fv);

			// neighbors
			int leftID = st[i] > 0 ? inst.posA[st[i] - 1] : START;
			int rightID = en[i] < toks.length ? inst.posA[en[i]] : END;
			if (leftID > 0 && rightID > 0) {
				fv = createNeighborFeatureVector(inst, i, leftID, rightID, fv);
			}

			// conjunction
			if (specialPos[i] == SpecialPos.C) {
				int[] arg = findConjArg(inst, i);
				//int head = arg[0];
				int left = arg[1];
				int right = arg[2];
				if (left != -1 && right != -1 && left < right) {
					int lid = findRightNearestChild(child[left], i);
					int leftSpan = lid == -1 ? 1
							: (en[lid] - st[child[left][0]]);

					int rid = findLeftNearestChild(child[right], i);
					int rightSpan = rid == -1 ? 1
							: (en[child[right][child[right].length - 1]] - st[rid]);

					int lenDiff = Math.abs(rightSpan - leftSpan);
					if (lenDiff >= GlobalFeatureData.maxLenDiff)
						lenDiff = GlobalFeatureData.maxLenDiff;
					fv = createFeatureVectorCC3(inst, left, right, lenDiff, fv);

					boolean sameChild = (lid == -1 && rid == -1)
							|| (lid != -1
									&& rid != -1
									&& inst.posA[child[left][0]] == inst.posA[rid] && inst.posA[lid] == inst.posA[child[right][child[right].length - 1]]);
					fv = createFeatureVectorCCChild(inst, left, right,
							sameChild ? 1 : 0, fv);

				}
			}

			// pp attachment in local feature

			// punc head in local feature
		}

		if (DependencyParser.isSpmrl) {
			// noun-adj
			//fv = adjectiveDetAgreement(inst, fv);
			
			// idf feature
			//fv = idfDep(inst, fv);
		}
		return fv;
	}

	public FeatureVector createGlobalFeatureVector(DependencyInstance inst,
			GlobalFeatureData gfd, FeatureVector fv) {

		if (!DependencyParser.useGlobal)
			return fv;
		
		
		int[] toks = inst.sentence;
		int[] deps = inst.dep;
		int[] pos = inst.pos;
		SpecialPos[] specialPos = inst.specialPos;
		int[][] child = inst.child;
		int[] en = inst.en;
		int[] st = inst.st;

		int rb = getMSTRightBranch(inst, 0);
		if (gfd.RB == -1) {
			//gfd.RB = dataAlphabet.lookupIndex("RB=");
			gfd.RB = HighOrder.RB.ordinal();
		}
		//fv = add(gfd.RB, (double) rb / toks.length, fv);
		fv = addCode(THighOrder, gfd.RB, (double)rb / toks.length, fv);

		if (Math.abs(inst.parserRerankScore) > 1e-4) {
			fv = add("Score", inst.parserRerankScore, fv);
		}

		for (int i = 1; i < toks.length; i++) {
			if (deps[i] == -1)
				continue;

			// child num & parent-child
			int leftNum = 0;
			int rightNum = 0;
			StringBuilder childStr = new StringBuilder();
			for (int j = 0; j < child[i].length; j++) {
				if (specialPos[child[i][j]] != SpecialPos.PNX) {
					if (child[i][j] < i
							&& leftNum < GlobalFeatureData.maxChildNum)
						leftNum++;
					else if (child[i][j] > i
							&& rightNum < GlobalFeatureData.maxChildNum)
						rightNum++;
					childStr.append(" ").append(pos[child[i][j]]);
				}
			}
			fv = add("Gl0=" + pos[i] + " " + childStr.toString(), 1.0, fv);

			if (gfd.fvs_cn[i][leftNum][rightNum] == null) {
				gfd.fvs_cn[i][leftNum][rightNum] = 
						createChildNumFeatureVector(inst, i, leftNum, rightNum, end);
			}
			fv = fv.concat(gfd.fvs_cn[i][leftNum][rightNum]);

			// heavy
			int endflag = en[i] == toks.length ? 1 : 0;
			int punc = (en[i] < toks.length && specialPos[en[i]] == SpecialPos.PNX) ? 1
					: 0;
			int bin = Math.min(GlobalFeatureData.maxBinLength, (en[i] - st[i]));
			if (gfd.fvs_heavy[i][endflag][punc][bin] == null) {
				gfd.fvs_heavy[i][endflag][punc][bin] = 
						createHeavyFeatureVector(inst, i, endflag, punc, bin, end);
			}
			fv = fv.concat(gfd.fvs_heavy[i][endflag][punc][bin]);

			// neighbors
			int leftID = st[i] > 0 ? inst.posA[st[i] - 1] : START;
			int rightID = en[i] < toks.length ? inst.posA[en[i]] : END;
			if (leftID > 0 && rightID > 0) {
				if (gfd.fvs_nb[i][leftID][rightID] == null) {
					gfd.fvs_nb[i][leftID][rightID] = 
							createNeighborFeatureVector(inst, i, leftID, rightID, end);
				}
				fv = fv.concat(gfd.fvs_nb[i][leftID][rightID]);
			}

			// conjunction
			if (specialPos[i] == SpecialPos.C) {
				int[] arg = findConjArg(inst, i);
				//int head = arg[0];
				int left = arg[1];
				int right = arg[2];
				if (left != -1 && right != -1 && left < right) {
					int lid = findRightNearestChild(child[left], i);
					int leftSpan = lid == -1 ? 1
							: (en[lid] - st[child[left][0]]);

					int rid = findLeftNearestChild(child[right], i);
					int rightSpan = rid == -1 ? 1
							: (en[child[right][child[right].length - 1]] - st[rid]);

					int lenDiff = Math.abs(rightSpan - leftSpan);
					if (lenDiff >= GlobalFeatureData.maxLenDiff)
						lenDiff = GlobalFeatureData.maxLenDiff;
					if (gfd.fvs_cc3[left][right][lenDiff] == null) {
						gfd.fvs_cc3[left][right][lenDiff] = 
								createFeatureVectorCC3(inst, left, right, lenDiff, end);
					}
					fv = fv.concat(gfd.fvs_cc3[left][right][lenDiff]);

					boolean sameChild = (lid == -1 && rid == -1)
							|| (lid != -1
									&& rid != -1
									&& inst.posA[child[left][0]] == inst.posA[rid] && inst.posA[lid] == inst.posA[child[right][child[right].length - 1]]);
					if (gfd.fvs_ccChild[left][right][sameChild ? 1 : 0] == null) {
						gfd.fvs_ccChild[left][right][sameChild ? 1 : 0] = 
								createFeatureVectorCCChild(inst, left, right, sameChild ? 1 : 0, end);
					}
					fv = fv.concat(gfd.fvs_ccChild[left][right][sameChild ? 1 : 0]);
				}
			}

		}

		if (DependencyParser.isSpmrl) {
			// noun-adj
			//fv = adjectiveDetAgreement(inst, fv);
			
			// idf feature
			//fv = idfDep(inst, fv);
		}

		return fv;
	}

	public double getGlobalScore(DependencyInstance inst,
			GlobalFeatureData gfd, Parameters params) {

		if (!DependencyParser.useGlobal)
			return 0.0;

		
		
		int[] toks = inst.sentence;
		int[] deps = inst.dep;
		int[] pos = inst.pos;
		SpecialPos[] specialPos = inst.specialPos;
		int[][] child = inst.child;
		int[] en = inst.en;
		int[] st = inst.st;

		double score = 0.0;
		FeatureVector fv = end;

		int rb = getMSTRightBranch(inst, 0);
		if (gfd.RB == -1) {
			//gfd.RB = dataAlphabet.lookupIndex("RB=");
			gfd.RB = HighOrder.RB.ordinal();
		}
		//fv = add(gfd.RB, (double) rb / toks.length, fv);
		fv = addCode(THighOrder, gfd.RB, (double)rb / toks.length, fv);

		if (Math.abs(inst.parserRerankScore) > 1e-4) {
			fv = add("Score", inst.parserRerankScore, fv);
		}

		for (int i = 1; i < toks.length; i++) {
			if (deps[i] == -1)
				continue;

			// child num & parent-child
			int leftNum = 0;
			int rightNum = 0;
			StringBuilder childStr = new StringBuilder();
			for (int j = 0; j < child[i].length; j++) {
				if (specialPos[child[i][j]] != SpecialPos.PNX) {
					if (child[i][j] < i
							&& leftNum < GlobalFeatureData.maxChildNum)
						leftNum++;
					else if (child[i][j] > i
							&& rightNum < GlobalFeatureData.maxChildNum)
						rightNum++;
					childStr.append(" ").append(pos[child[i][j]]);
				}
			}
			fv = add("Gl0=" + pos[i] + " " + childStr.toString(), 1.0, fv);

			if (gfd.fvs_cn[i][leftNum][rightNum] == null) {
				gfd.fvs_cn[i][leftNum][rightNum] = 
						createChildNumFeatureVector(inst, i, leftNum, rightNum, end);
				gfd.probs_cn[i][leftNum][rightNum] = params.getScore(gfd.fvs_cn[i][leftNum][rightNum]);
			}
			score += gfd.probs_cn[i][leftNum][rightNum];

			// heavy
			int endflag = en[i] == toks.length ? 1 : 0;
			int punc = (en[i] < toks.length && specialPos[en[i]] == SpecialPos.PNX) ? 1
					: 0;
			int bin = Math.min(GlobalFeatureData.maxBinLength, (en[i] - st[i]));
			if (gfd.fvs_heavy[i][endflag][punc][bin] == null) {
				gfd.fvs_heavy[i][endflag][punc][bin] = 
						createHeavyFeatureVector(inst, i, endflag, punc, bin, end);
				gfd.probs_heavy[i][endflag][punc][bin] = params.getScore(gfd.fvs_heavy[i][endflag][punc][bin]);
			}
			score += gfd.probs_heavy[i][endflag][punc][bin];

			// neighbors
			int leftID = st[i] > 0 ? inst.posA[st[i] - 1] : START;
			int rightID = en[i] < toks.length ? inst.posA[en[i]] : END;
			if (leftID > 0 && rightID > 0) {
				if (gfd.fvs_nb[i][leftID][rightID] == null) {
					gfd.fvs_nb[i][leftID][rightID] = 
							createNeighborFeatureVector(inst, i, leftID, rightID, end);
					gfd.probs_nb[i][leftID][rightID] = params.getScore(gfd.fvs_nb[i][leftID][rightID]);
				}
				score += gfd.probs_nb[i][leftID][rightID];
			}

			// conjunction
			if (specialPos[i] == SpecialPos.C) {
				int[] arg = findConjArg(inst, i);
				//int head = arg[0];
				int left = arg[1];
				int right = arg[2];
				if (left != -1 && right != -1 && left < right) {
					int lid = findRightNearestChild(child[left], i);
					int leftSpan = lid == -1 ? 1
							: (en[lid] - st[child[left][0]]);

					int rid = findLeftNearestChild(child[right], i);
					int rightSpan = rid == -1 ? 1
							: (en[child[right][child[right].length - 1]] - st[rid]);

					int lenDiff = Math.abs(rightSpan - leftSpan);
					if (lenDiff >= GlobalFeatureData.maxLenDiff)
						lenDiff = GlobalFeatureData.maxLenDiff;
					if (gfd.fvs_cc3[left][right][lenDiff] == null) {
						gfd.fvs_cc3[left][right][lenDiff] = 
								createFeatureVectorCC3(inst, left, right, lenDiff, end);
						gfd.probs_cc3[left][right][lenDiff] = params.getScore(gfd.fvs_cc3[left][right][lenDiff]);
					}
					score += gfd.probs_cc3[left][right][lenDiff];

					boolean sameChild = (lid == -1 && rid == -1)
							|| (lid != -1
									&& rid != -1
									&& inst.posA[child[left][0]] == inst.posA[rid] && inst.posA[lid] == inst.posA[child[right][child[right].length - 1]]);
					if (gfd.fvs_ccChild[left][right][sameChild ? 1 : 0] == null) {
						gfd.fvs_ccChild[left][right][sameChild ? 1 : 0] = 
								createFeatureVectorCCChild(inst, left, right, sameChild ? 1 : 0, end);
						gfd.probs_ccChild[left][right][sameChild ? 1 : 0] = params.getScore(gfd.fvs_ccChild[left][right][sameChild ? 1 : 0]);
					}
					
					score += gfd.probs_ccChild[left][right][sameChild ? 1 : 0];
				}
			}

		}

		if (DependencyParser.isSpmrl) {
			// noun-adj
			//fv = adjectiveDetAgreement(inst, fv);
			
			// idf feature
			//fv = idfDep(inst, fv);
		}

		return score + params.getScore(fv);
	}

	public FeatureVector adjectiveDetAgreement(DependencyInstance s, FeatureVector fv) {
		for(int m = 1; m < s.pos.length; m++) {
			int h = s.dep[m];
			if (s.specialPos[h] == SpecialPos.N && s.specialPos[m] == SpecialPos.AJ) {
				boolean det = s.cm[h].dt;
				boolean findIDF = true;
				int tmph = h;
				while (findIDF) {
					findIDF = false;
					for (int i = s.child[tmph].length - 1; i >= 0; i--) {
						if (s.labs[s.child[tmph][i]] == IDF) {
							findIDF = true;
							tmph = s.child[tmph][i];
							break;
						}
					}
				}
				det = s.cm[tmph].dt;
				if (det == s.cm[m].dt)
					fv = addCode(TCatib, CatibFeature.ADJ_DT_AGREE.ordinal(), 1.0, fv);
			}
		}

		return fv;
	}

	public FeatureVector idfDep(DependencyInstance s, FeatureVector fv) {
		for (int i = 0; i < s.sentence.length; i++) {
			if (s.labs[i] == IDF) {
				int h = s.dep[i];
				if (s.cm[h].dt)
					fv = addCode(TCatib, CatibFeature.IDF_AGREE.ordinal(), 1.0, fv);
				if (h > i)
					fv = addCode(TCatib, CatibFeature.IDF_DIR.ordinal(), 1.0, fv);
			}
			int idfNum = 0;
			for (int j = 0; j < s.child[i].length; j++)
				if (s.labs[s.child[i][j]] == IDF)
					idfNum++;
			if (idfNum > 1)
				fv = addCode(TCatib, CatibFeature.IDF_NUM.ordinal(), 1.0, fv);
		}
		return fv;
	}

	public FeatureVector createChildNumFeatureVector(DependencyInstance s,
			int id, int leftNum, int rightNum, FeatureVector fv) {
		
		
		int childNum = Math.min(GlobalFeatureData.maxChildNum, leftNum + rightNum);
		childNum = childNum << tempOff;
		int HP = s.pos[id];
		long code = 0;
		
		code = genCodePF(HighOrder.CN_HP_NUM.ordinal(), HP);
		fv = addCode(THighOrder, code | childNum, 1.0, fv);
		
		code = genCodePFF(HighOrder.CN_HP_LNUM_RNUM.ordinal(), HP, leftNum, rightNum);
		fv = addCode(THighOrder, code, 1.0, fv);
		
		return fv;
	}

	public FeatureVector createHeavyFeatureVector(DependencyInstance s, int id,
			int end, int punc, int bin, FeatureVector fv) {
		
		
		int HP = s.pos[id];
		int HC = s.posA[id];
		int epFlag = (end << 1) | punc;
		
		long code = 0;
		
		code = genCodePFF(HighOrder.HV_HP.ordinal(), HP, epFlag, bin);
		fv = addCode(THighOrder, code, 1.0, fv);
		
		code = genCodePFF(HighOrder.HV_HC.ordinal(), HC, epFlag, bin);
		fv = addCode(THighOrder, code, 1.0, fv);
		
		return fv;
	}

	public FeatureVector createNeighborFeatureVector(DependencyInstance s,
			int id, int left, int right, FeatureVector fv) {
		
		
		int HP = s.pos[id];
		int HC = s.posA[id];
		int HL = s.m[id].lemma;
		
		long code = 0;
		
		code = genCodePPP(HighOrder.NB_HP_LC_RC.ordinal(), HP, left, right);
		fv = addCode(THighOrder, code, 1.0, fv);

		code = genCodePPP(HighOrder.NB_HC_LC_RC.ordinal(), HC, left, right);
		fv = addCode(THighOrder, code, 1.0, fv);

		code = genCodePPW(HighOrder.NB_HL_LC_RC.ordinal(), left, right, HL);
		fv = addCode(THighOrder, code, 1.0, fv);

		return fv;
	}

	public int findPuncCounterpart(int[] word, int arg) {
		int quoteID = QUOTE;
		int lrbID = LRB;
		int rrbID = RRB;
		if (word[arg] == quoteID) {
			boolean left = false;
			int prev = -1;
			int curr = -1;
			for (int i = 1; i < word.length; i++) {
				if (word[i] == quoteID) {
					left = !left;
					prev = curr;
					curr = i;
				}
				if (i == arg) {
					break;
				}
			}
			if (left) {
				// left quote
				curr = -1;
				for (int i = arg + 1; i < word.length; i++) {
					if (word[i] == quoteID) {
						curr = i;
						break;
					}
				}
			} else {
				// right quote
				curr = prev;
			}
			return curr;
		} else if (word[arg] == lrbID) {
			// left bracket
			int curr = -1;
			for (int i = arg + 1; i < word.length; i++) {
				if (word[i] == rrbID) {
					curr = i;
					break;
				}
			}
			return curr;
		} else if (word[arg] == rrbID) {
			int curr = -1;
			for (int i = arg - 1; i >= 0; i--) {
				if (word[i] == lrbID) {
					curr = i;
					break;
				}
			}
			return curr;
		}

		return -1;
	}

	public void getTreeString(DependencyInstance s, int id, int dep,
			StringBuilder str) {
		str.append(s.pos[id]);
		if (dep > 0) {
			for (int i = 0; i < s.child[id].length; i++) {
				str.append("|"); // go to next level
				getTreeString(s, s.child[id][i], dep - 1, str);
				str.append("#"); // go back
			}
		}
	}

	public FeatureVector createFeatureVectorCC3(DependencyInstance s, int left,
			int right, int lenDiff, FeatureVector fv) {
		// dep 1
		// StringBuilder leftRightTree = new StringBuilder();
		// getTreeString(s, left, 1, leftRightTree);
		// leftRightTree.append(" ");
		// getTreeString(s, right, 1, leftRightTree);

		// StringBuilder feat = new StringBuilder(prefix + "1=" +
		// leftRightTree.toString());
		// fv = add(feat.toString(), 1.0, fv);
		// fv = add(feat.append(wstr).toString(), 1.0, fv);
		
		
		lenDiff = lenDiff << tempOff;
		int LP = s.pos[left];
		int RP = s.pos[right];
		int LC = s.posA[left];
		int RC = s.posA[right];
		
		long code = 0;
		
		code = genCodePPF(HighOrder.CC_LP_RP_LENDIFF.ordinal(), LP, RP);
		fv = addCode(THighOrder, code | lenDiff, 1.0, fv);

		code = genCodePPF(HighOrder.CC_LC_RC_LENDIFF.ordinal(), LC, RC);
		fv = addCode(THighOrder, code | lenDiff, 1.0, fv);

		code = lenDiff | HighOrder.CC_LENDIFF.ordinal();
		fv = addCode(THighOrder, code | lenDiff, 1.0, fv);

		return fv;
	}

	public FeatureVector createFeatureVectorCCChild(DependencyInstance s,
			int left, int right, int flag, FeatureVector fv) {
		
		
		int LP = s.pos[left];
		int RP = s.pos[right];
		int LC = s.posA[left];
		int RC = s.posA[right];
		flag = flag << tempOff;
		
		long code = 0;
		
		code = genCodePPF(HighOrder.CC_LP_RP_CHILDF.ordinal(), LP, RP);
		fv = addCode(THighOrder, code | flag, 1.0, fv);

		code = genCodePPF(HighOrder.CC_LC_RC_CHILDF.ordinal(), LC, RC);
		fv = addCode(THighOrder, code | flag, 1.0, fv);

		return fv;
	}

	public int findLeftNearestChild(int[] child, int id) {
		int ret = -1;
		for (int i = child.length - 1; i >= 0; i--)
			if (child[i] < id) {
				ret = child[i];
				break;
			}
		return ret;
	}

	public int findRightNearestChild(int[] child, int id) {
		int ret = -1;
		for (int i = 0; i < child.length; i++)
			if (child[i] > id) {
				ret = child[i];
				break;
			}
		return ret;
	}

	public int findRightNearestChildID(int[] child, int id) {
		int ret = 0;
		for (; ret < child.length; ret++)
			if (child[ret] > id) {
				break;
			}
		return ret;
	}

	public int getMSTRightBranch(DependencyInstance s, int id) {
		int node = 1;
		for (int i = s.child[id].length - 1; i >= 0; i--) {
			//if (s.pos[s.child[id][i]].equals("PNX"))
			if (s.specialPos[s.child[id][i]] == SpecialPos.PNX)
				continue;
			node += getMSTRightBranch(s, s.child[id][i]);
			break;
		}
		return node;
	}

	public boolean isProj(int h, int a, int[] dep) {
		int large = h > a ? h : a;
		int small = h > a ? a : h;
		for (int i = 0; i < dep.length; i++) {
			if (i == h || i == a)
				continue;
			if ((i < small || i > large) && (dep[i] > small && dep[i] < large))
				return false;
			if ((i > small && i < large) && (dep[i] < small || dep[i] > large))
				return false;
		}
		return true;
	}

    public HashMap<Integer, String> processChangeStr(String str) {
    	String[] data = str.split("\t");
    	HashMap<Integer, String> change = new HashMap<Integer, String>();
    	for (int i = 0; i < data.length; i++) {
    		if (data[i].isEmpty())
    			continue;
    		int id = Integer.parseInt(data[i].substring(data[i].lastIndexOf(":") + 1));
    		change.put(id, data[i].substring(0, data[i].lastIndexOf(":")));
    	}
    	return change;
    }
    
	public DependencyInstance[] createInstances(String goldFile, String predFile) throws IOException {

		createAlphabet(goldFile);

		System.out.println("Num Features: " + (dataAlphabet.size() - 1));
	    System.out.println("Num Edge Labels: " + (typeAlphabet.size() - 1));

	    // don't need to build forest anymore
	    
		BufferedReader in = new BufferedReader(new InputStreamReader(new FileInputStream(goldFile), "UTF8"));
		BufferedReader in_p = null;
		if (DependencyParser.learningMode == 0 || DependencyParser.learningMode == 1 
				|| DependencyParser.learningMode == 4 || DependencyParser.learningMode == 5)
			in_p = new BufferedReader(new InputStreamReader(new FileInputStream(predFile),"UTF8"));
		else if (DependencyParser.learningMode == 2 || DependencyParser.learningMode == 3)
			in_p = new BufferedReader(new InputStreamReader(new GZIPInputStream(new FileInputStream(predFile + ".gz")),"UTF8"));

		BufferedReader in_c = null;
		if (DependencyParser.pruningMode == 4) {
			// candidate in a separate file
			String candfile = predFile.substring(0, predFile.lastIndexOf(".")) + "." + DependencyParser.candExt;
			in_c = new BufferedReader(new InputStreamReader(new FileInputStream(candfile),"UTF8"));
		}

		//ObjectOutputStream out = null;
		//if (DependencyParser.learningMode == 2 || DependencyParser.learningMode == 3) 
		//		out = new ObjectOutputStream(new GZIPOutputStream(new FileOutputStream(goldFile + ".forest")));

		LinkedList<DependencyInstance> lt = new LinkedList<DependencyInstance>();

		System.out.print("Creating Instances ... ");
		
		DependencyInstance gold = readConllSent(in, labeled);
		
		int pruningRecallTotal = 0;
		int pruningRecallCorr = 0;
		int pruningTotalCandidate = 0;
		int candidatePos = 0; 

		int num1 = 0;
		while (gold != null) {
			if (num1 % 1000 == 0)
				System.out.print("  " + num1);

			int len = gold.sentence.length;

			DependencyInstance pred = readConllSent(in_p, labeled);
			pred.m = gold.m;
			pred.posA = gold.posA;
    		pred.cm = gold.cm;
    		pred.origWord = gold.origWord;
    		if (DependencyParser.correctPos) {
	    		for (int z = 1; z < pred.dep.length; z++) {
	    			if (pred.sentence[z] != gold.sentence[z]
	    					|| pred.goldPos[z] != gold.pos[z]) {
	    				System.out.println("train bug 1");
	    			}
	    		}
    		}
    		else {
	    		for (int z = 1; z < pred.dep.length; z++) {
	    			if (pred.sentence[z] != gold.sentence[z]
	    					|| pred.pos[z] != gold.pos[z]) {
	    				System.out.println("train bug 1");
	    			}
	    		}
    		}

			if (DependencyParser.learningMode == 0 || DependencyParser.learningMode == 1 || DependencyParser.learningMode == 4
					|| DependencyParser.learningMode == 5) {
				LocalFeatureData lfd = new LocalFeatureData(len, types.length, false);
	    		if (DependencyParser.pruningMode == 0) {
	    			// currently the pred-file should still exist, lazy to change the code...
	    			lfd.loadCandidate(null, this, pred, gold, false);
	    			in_p.readLine();
	    			for (int z = 1; z < pred.sentence.length; z++)
	    				in_p.readLine();
	    			in_p.readLine();
	    		}
	    		else if (DependencyParser.pruningMode == 1) {
					// generate candidate
					lfd.loadCandidate(null, this, pred, gold, false);
				} 
	    		else if (DependencyParser.pruningMode == 2 || DependencyParser.pruningMode == 3) {
					// load candidate
					in_p.readLine(); // score
					lfd.loadCandidate(in_p, this, pred, gold, false);
					in_p.readLine(); // empty line
				}
	    		else if (DependencyParser.pruningMode == 4) {
	    			if (DependencyParser.predHasCand) {
		    			in_p.readLine();
		    			for (int z = 1; z < gold.dep.length; z++)
		    				in_p.readLine();
		    			in_p.readLine();
	    			}
	    			// load real candidate
	    			readConllSent(in_c, labeled);
	    			in_c.readLine();
	    			lfd.loadCandidate(in_c, this, pred, gold, false);
	    			in_c.readLine();
	    		}
				// in.readLine(); // empty line
	    		
				if (len > DependencyParser.maxLength) {
					System.out.println("too long: " + len);
				}
				else if ((double)lfd.maxid / len > DependencyParser.maxAvgHead) {
					System.out.println("too many heads: " + ((double)lfd.maxid / len));
				}
				else {
		    		// pruning quality
					for (int i = 1; i < gold.dep.length; i++) {
	    				pruningRecallTotal++;
	    				for (int h = 0; h < gold.dep.length; h++) {
	    					if (!lfd.cand[h][i])
	    						continue;
	        				pruningTotalCandidate++;
	    					if (h == gold.dep[i])
	    						pruningRecallCorr++;
	    				}
	    				if (DependencyParser.correctPos) {
	    					candidatePos += pred.candPos[i].length;
	    				}
	    				else {
	    					candidatePos++;
	    				}
	    			}
				}
			}
			else if (DependencyParser.learningMode == 2 || DependencyParser.learningMode == 3) {
    			// compute feature vector for top-k list
				String tmp = in_p.readLine();
				//System.out.println(tmp);
    			HashMap<Integer, String> labChange = processChangeStr(tmp);
    			HashMap<Integer, String> depChange = processChangeStr(in_p.readLine());
    			
    			int k = 1;
    			String str = in_p.readLine();
    			int bestAcc = 0;
				for (int z = 1; z < pred.dep.length; z++)
					if (pred.dep[z] == gold.dep[z])
						bestAcc++;
				copyPredDepLab(pred, pred.dep, pred.labs);
				
				//FeatureVector fv = createFeatureVector(pred);
				//outputFeatureVector(out, fv);

    			while (!str.isEmpty()) {
    				if (k < DependencyParser.maxRerankK) {
	    				String[] data = str.split("\t");
	    				pred.parserRerankScore = Double.parseDouble(data[0]);
	    				for (int z = 1; z < data.length; z++) {
	    					int changeID = Integer.parseInt(data[z]);
	    					if (depChange.containsKey(changeID)) {
	    						String s = depChange.get(changeID);
	    						int p = s.indexOf(":");
	    						int mod = Integer.parseInt(s.substring(0, p));
	    						int head = Integer.parseInt(s.substring(p + 1));
	    						pred.dep[mod] = head;
	    					}
	    					else if (labeled && labChange.containsKey(z)) {
	    						String s = depChange.get(changeID);
	    						int p = s.indexOf(":");
	    						int mod = Integer.parseInt(s.substring(0, p));
	    						int lab = typeAlphabet.lookupIndex(s.substring(p + 1));
	    						pred.labs[mod] = lab;
	    					}
	    				}
	    				
	    				int acc = 0;
	    				for (int z = 1; z < pred.dep.length; z++)
	    					if (pred.dep[z] == gold.dep[z])
	    						acc++;
	    				if (acc > bestAcc)
	    					bestAcc = acc;
	    				
	    				//fv = createFeatureVector(pred);
	    				//outputFeatureVector(out, fv);

	    				// recover
	    				for (int z = 0; z < pred.dep.length; z++) {
	    					pred.dep[z] = pred.parserPredDep[z];
	    					pred.labs[z] = pred.parserPredLab[z];
	    				}
	    				buildChild(pred);
	    				
    					k++;
    				}
    				str = in_p.readLine();
    			}
    			//out.writeInt(-1);
    			
				if (len > DependencyParser.maxLength) {
					System.out.println("too long: " + len);
				}
				else {
		    		// pruning quality
    				pruningRecallTotal += gold.dep.length - 1;
    				pruningTotalCandidate += Math.min(DependencyParser.maxRerankK, k) * (gold.dep.length - 1);
    				pruningRecallCorr += bestAcc;
				}
			}

			lt.add(new DependencyInstance(len));
			gold = readConllSent(in, labeled);
			num1++;
			if (num1 >= DependencyParser.trainSentences)
				break;
		}

		System.out.println("purning recall: "
					+ (double)pruningRecallCorr
					/ pruningRecallTotal);
		System.out.println("avg head: "
					+ (double) pruningTotalCandidate
					/ pruningRecallTotal);
		System.out.println("avg pos candidates: " + (double)candidatePos / pruningRecallTotal);

		DependencyInstance[] pti = new DependencyInstance[lt.size()];
		for (int i = 0; i < pti.length; i++) {
			pti[i] = (DependencyInstance) lt.get(i);
		}

		in.close();
		in_p.close();

		if (DependencyParser.pruningMode == 4)
			in_c.close();
		
		//if (DependencyParser.learningMode == 2 || DependencyParser.learningMode == 3)
		//	out.close();

		System.out.println("Done.");
		return pti;
	}

	public DependencyInstance createInstance(BufferedReader in)
			throws IOException {
		DependencyInstance pti = readConllSent(in, labeled);

		if (pti != null) {
			String spans = "";
			for (int i = 1; i < pti.dep.length; i++) {
				spans += pti.dep[i] + "|" + i + ":"
						+ pti.labs[i] + " ";
			}
			pti.actParseTree = spans.trim();
		}

		return pti;
	}

	public void outputFeatureVector(ObjectOutputStream out, FeatureVector fv)
			throws IOException {
		if (fv == null) {
			System.out.println("output bug");
		}
		FeatureVector sfv = new FeatureVector(-1, -1.0, null);
		for (FeatureVector curr = fv; curr != null; curr = curr.next) {
			if (curr.index >= 0) {
				if (Math.abs(curr.value - 1.0) < 1e-4) {
					out.writeInt(curr.index);
				} else {
					sfv = new FeatureVector(curr.index, curr.value, sfv);
				}
			}
		}
		out.writeInt(-2);

		for (FeatureVector curr = sfv; curr != null; curr = curr.next) {
			if (curr.index >= 0) {
				out.writeInt(curr.index);
				out.writeDouble(curr.value);
			}
		}
		out.writeInt(-2);
	}

	public FeatureVector readFeatureVector(ObjectInputStream in)
			throws IOException {
		FeatureVector prodFV = new FeatureVector(-1, -1.0, null);
		int indx = in.readInt();
		while (indx != -2) {
			prodFV = new FeatureVector(indx, 1.0, prodFV);
			indx = in.readInt();
		}
		indx = in.readInt();
		while (indx != -2) {
			double v = in.readDouble();
			prodFV = new FeatureVector(indx, v, prodFV);
			indx = in.readInt();
		}
		return prodFV;
	}

}
